// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/CORBA.h>

#include <Simple.h>

#include <stdlib.h>

#ifdef WIN32
#   include <sys/timeb.h>
#endif

namespace OB
{

timeval
TimeNow()
{
    timeval tv;
#ifdef WIN32
    struct _timeb timebuffer;
    _ftime(&timebuffer);
    tv.tv_sec = timebuffer.time;
    tv.tv_usec = timebuffer.millitm * 1000;
#else
    gettimeofday(&tv, 0);
#endif
    return tv;
}

timeval
TimeSub(const timeval& a, const timeval& b)
{
    timeval tv;

    tv.tv_sec = a.tv_sec - b.tv_sec;
    tv.tv_usec = a.tv_usec - b.tv_usec;

    tv.tv_sec += tv.tv_usec / 1000000;
    tv.tv_usec = tv.tv_usec % 1000000;

    if(tv.tv_usec < 0)
    {
        tv.tv_usec += 1000000;
        tv.tv_sec -= 1;
    }

    return tv;
}

} // End of namespace OB

void
usage(const char* prog)
{
    fprintf(stderr, "Usage: %s [--oneway] repetitions\n", prog);
}

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    bool oneway = false;
    int reps;

    //
    // Get number of repetitions
    //
    if(argc < 2)
    {
	usage(argv[0]);
        return EXIT_FAILURE;
    }

    if(argc == 3)
    {
	if(strcmp(argv[1], "--oneway") != 0)
	{
	    usage(argv[0]);
	    return EXIT_FAILURE;
	}
	oneway = true;
	reps = atoi(argv[2]);
    }
    else
	reps = atoi(argv[1]);

    if(reps <= 0)
    {
        fprintf(stderr, "%s: number of repetitions must be larger than 0\n",
		argv[0]);
        return EXIT_FAILURE;
    }
	
    //
    // Get "I" object
    //
    CORBA::Object_var obj = orb -> string_to_object("relfile:/I.ref");
    if(CORBA::is_nil(obj))
    {
        fprintf(stderr, "%s: cannot read IOR from I.ref\n", argv[0]);
        return EXIT_FAILURE;
    }
	
    I_var i = I::_narrow(obj);
    assert(!CORBA::is_nil(i));
    
    //
    // Force a connection to the server.
    //
    timeval tv = OB::TimeNow();
    i -> f();
    tv = OB::TimeSub(OB::TimeNow(), tv);
    printf("Time needed for initialization: %f ms\n",
           ((double)tv.tv_sec + (double)tv.tv_usec / 1000000.0) * 1000.0);

    //
    // Main loop
    //
    tv = OB::TimeNow();
    if(oneway)
    {
	for(int j = 0; j < reps; ++j)
	    i -> o();
    }
    else
    {
	for(int j = 0; j < reps; ++j)
	    i -> f();
    }
    tv = OB::TimeSub(OB::TimeNow(), tv);
    printf("Total time needed: %f ms\n",
          ((double)tv.tv_sec + (double)tv.tv_usec / 1000000.0) * 1000.0);
    printf("Time per %s method call: %f ms\n",
           ((oneway) ? "oneway" : "twoway"),
           (((double)tv.tv_sec + (double)tv.tv_usec / 1000000.0)
    	   / (double)reps) * 1000.0);
    
    return EXIT_SUCCESS;
}


int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	fprintf(stderr, "%s\n", ex._rep_id());;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    fprintf(stderr, "%s\n", ex._rep_id());;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
