// **********************************************************************
//
// Copyright (c) 1999
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/CORBA.h>

#include <I.h>

#include <stdlib.h>

#ifdef WIN32
#   include <sys/timeb.h>
#endif

namespace OB
{

timeval
TimeNow()
{
    timeval tv;
#ifdef WIN32
    struct _timeb timebuffer;
    _ftime(&timebuffer);
    tv.tv_sec = timebuffer.time;
    tv.tv_usec = timebuffer.millitm * 1000;
#else
    gettimeofday(&tv, 0);
#endif
    return tv;
}

timeval
TimeSub(const timeval& a, const timeval& b)
{
    timeval tv;

    tv.tv_sec = a.tv_sec - b.tv_sec;
    tv.tv_usec = a.tv_usec - b.tv_usec;

    tv.tv_sec += tv.tv_usec / 1000000;
    tv.tv_usec = tv.tv_usec % 1000000;

    if(tv.tv_usec < 0)
    {
        tv.tv_usec += 1000000;
        tv.tv_sec -= 1;
    }

    return tv;
}

} // End of namespace OB

int
run(CORBA::ORB_ptr orb, int argc, char* argv[])
{
    //
    // Get number of repetitions
    //
    if(argc < 2)
    {
        fprintf(stderr, "Usage: %s repetitions\n", argv[0]);
        return EXIT_FAILURE;
    }

    CORBA::ULong reps = atoi(argv[1]);
	
    //
    // Get "I" object
    //
    CORBA::Object_var obj = orb -> string_to_object("relfile:/I.ref");
    if(CORBA::is_nil(obj))
    {
        fprintf(stderr, "%s: cannot read IOR from I.ref\n", argv[0]);
        return EXIT_FAILURE;
    }
	
    I_var i = I::_narrow(obj);
    assert(!CORBA::is_nil(i));
    
    timeval tv = OB::TimeNow();
    i -> _non_existent();

    //
    // Construct the sequences
    //
    CORBA::ULong j;
#define LENGTH	(1024*10)

    I::ShortSeq h(LENGTH);
    h.length(LENGTH);
    for(j = 0 ; j < LENGTH ; j++)
	h[j] = (CORBA::Short)j;

    I::LongSeq l;
    l.length(LENGTH);
    for(j = 0 ; j < LENGTH ; j++)
	l[j] = j;

    I::DoubleSeq d;
    d.length(LENGTH);
    for(j = 0 ; j < LENGTH ; j++)
	d[j] = j/LENGTH;

    I::CharSeq c;
    c.length(LENGTH);
    for(j = 0 ; j < LENGTH ; j++)
	c[j] = (CORBA::Char)j - '0';

    I::OctetSeq o;
    o.length(LENGTH);
    for(j = 0 ; j < LENGTH ; j++)
	o[j] = (CORBA::Octet)(j % 256);

    I::EnumSeq e;
    e.length(LENGTH);
    for(j = 0 ; j < LENGTH ; j++)
	e[j] = (I::EnumType)(j % 2);

    CORBA::String_var str = CORBA::string_alloc(LENGTH);
    for(j = 0 ; j < LENGTH ; j++)
	str[j] = (CORBA::Char)(j - '0');
    str[LENGTH] = '\0';

    I::ShortArr sarr;
    CORBA::ULong j1;
    for(j = 0 ; j < 64 ; j++)
	for(j1 = 0 ; j1 < 64 ; j1++)
	    sarr[j][j1] = (CORBA::Short)((j * 64) + j1);

    I::LongArr larr;
    for(j = 0 ; j < 64 ; j++)
	for(j1 = 0 ; j1 < 64 ; j1++)
	    larr[j][j1] = (CORBA::Long)((j * 64) + j1);

    I::DoubleArr darr;
    for(j = 0 ; j < 64 ; j++)
	for(j1 = 0 ; j1 < 64 ; j1++)
	    darr[j][j1] = (CORBA::Double)(((j * 64) + j1) / LENGTH);

    I::CharArr carr;
    for(j = 0 ; j < 64 ; j++)
	for(j1 = 0 ; j1 < 64 ; j1++)
	    carr[j][j1] = (CORBA::Char)(((j * 64) + j1) - '0');

    I::OctetArr oarr;
    for(j = 0 ; j < 64 ; j++)
	for(j1 = 0 ; j1 < 64 ; j1++)
	    oarr[j][j1] = (CORBA::Octet)(((j * 64) + j1) % 256);

    I::EnumArr earr;
    for(j = 0 ; j < 64 ; j++)
	for(j1 = 0 ; j1 < 64 ; j1++)
	    earr[j][j1] = (I::EnumType)(((j * 64) + j1) % 2);
    
    //
    // Main loop
    //
    for(j = 0 ; j < reps; j++)
    {
	i -> in_string(str);
	i -> in_shortseq(h);
	i -> in_longseq(l);
	i -> in_doubleseq(d);
	i -> in_charseq(c);
	i -> in_octetseq(o);
	i -> in_enumseq(e);
	i -> in_shortarr(sarr);
	i -> in_longarr(larr);
	i -> in_doublearr(darr);
	i -> in_chararr(carr);
	i -> in_octetarr(oarr);
	i -> in_enumarr(earr);
    }

    tv = OB::TimeSub(OB::TimeNow(), tv);
    printf("Total time needed: %f ms\n",
          ((double)tv.tv_sec + (double)tv.tv_usec / 1000000.0) * 1000.0);
    printf("Time per iteration: %f ms\n",
           (((double)tv.tv_sec + (double)tv.tv_usec / 1000000.0)
    	   / (double)reps) * 1000.0);
    
    return EXIT_SUCCESS;
}


int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;
    CORBA::ORB_var orb;

    try
    {
	orb = CORBA::ORB_init(argc, argv);
	status = run(orb, argc, argv);
    }
    catch(const CORBA::Exception& ex)
    {
	fprintf(stderr, "%s\n", ex._rep_id());;
	status = EXIT_FAILURE;
    }

    if(!CORBA::is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const CORBA::Exception& ex)
	{
	    fprintf(stderr, "%s\n", ex._rep_id());;
	    status = EXIT_FAILURE;
	}
    }
    
    return status;
}
