// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_CLIENT_H
#define OB_CLIENT_H

#include <OBE/Client_fwd.h>
#include <OBE/Downcall_fwd.h>

#include <OBE/Object.h>
#include <OBE/OCITypes.h>
#include <OBE/list.h>

namespace OB
{

//
// The Client class
//
class Client : virtual public RefCount
{
    //
    // Hide copy-constructor and assignment operator
    //
    Client(const Client&);
    void operator=(const Client&);

    int usage_; // The usage counter
    
protected:

    Client();
    virtual ~Client();

public:

    static inline Client_ptr _duplicate(Client_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline Client_ptr _nil()
    { return 0; }

    //
    // Destroy the client
    //
    virtual void destroy() = 0;

    //
    // Increment usage (not mutex protected)
    //
    void incUsage() { assert(usage_ >= 0); usage_++; }

    //
    // Decrement usage (not mutex protected)
    //
    // Returns true if after the decrement the usage counter is larger
    // than 0, and false otherwise.
    //
    bool decUsage() { assert(usage_ > 0); usage_--; return usage_ > 0; }

    //
    // Get a new request ID
    //
    virtual CORBA::ULong requestId() = 0;

    //
    // Get all profiles that are usable with this client
    //
    virtual OCI::ProfileInfoSeq* getUsableProfiles(const OCI::IOR&,
						 const CORBA::PolicyList&) = 0;
    
    //
    // Start a downcall, returning a downcall emitter
    //
    virtual DowncallEmitter_ptr startDowncall(DowncallImpl*)
	throw() = 0;

    //
    // Checks whether this client is equal to another client
    //
    virtual bool equal(Client_ptr) = 0;
};

//
// A sequence of clients
// 
typedef ObjSeq< Client, int > ClientSeq;
typedef SeqVar< ClientSeq > ClientSeq_var;

} // End of namespace OB

namespace CORBA
{

inline void
release(OB::Client_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::Client_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

namespace OB
{

//
// A client/profile pair
//
struct ClientProfilePair
{
    Client_var client;
    RefCountProfileInfo_var profile;
};

//
// The ClientManager class
//
class ClientManager : public OB::RefCount, public JTCMutex
{
    //
    // Hide copy-constructor and assignment operator
    //
    ClientManager(const ClientManager&);
    void operator=(const ClientManager&);

    bool destroy_; // True if destroy() was called

protected:

    //
    // The ORB Instance
    //
    OB::ORBInstance_var orbInstance_;

    //
    // All clients
    //
    OB_STL::list<Client_var> allClients_;

    //
    // All reusable clients
    //
    OB_STL::list<Client_var> reusableClients_;

    virtual void getCollocatedPairs(const IOP::IOR&, const CORBA::PolicyList&,
				    OB_STL::list<ClientProfilePair_var>*) = 0;

    ClientManager(ORBInstance_ptr);
    virtual ~ClientManager();

public:

    static inline ClientManager_ptr _duplicate(ClientManager_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline ClientManager_ptr _nil()
    { return 0; }

    //
    // Destroy the client manager, including all clients it contains
    //
    void destroy();

    //
    // Get a list of ClientProfilePairs for an IOR and a list of policies
    //
    OB_STL::list<ClientProfilePair_var>*
    getClientProfilePairs(const IOP::IOR&, const CORBA::PolicyList&);

    //
    // Release a client
    //
    void releaseClient(Client_ptr);
};

} // End of namespace OB

namespace CORBA
{

inline void
release(OB::ClientManager_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::ClientManager_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

#endif
