// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_DOWNCALL_H
#define OB_DOWNCALL_H

#include <OBE/Client_fwd.h>
#include <OBE/RefCounted_fwd.h>
#include <OBE/ORBInstance_fwd.h>
#include <OBE/Downcall_fwd.h>

#include <OBE/OCITypes.h>
#include <OBE/Stream.h>

namespace OB
{

//
// The DowncallImpl class (not reference counted)
//
class DowncallImpl
{
    //
    // Hide copy-constructor and assignment operator
    //
    DowncallImpl(const DowncallImpl&);
    void operator=(const DowncallImpl&);

protected:

    //
    // The ORBInstance object
    //
    ORBInstance_ptr orbInstance_;

    //
    // The client
    //
    Client_ptr client_;

    //
    // The downcall emitter
    //
    DowncallEmitter_var emitter_;

    //
    // Information about the IOR profile
    //
    RefCountProfileInfo_ptr profileInfo_;

    //
    // The list of policies
    //
    RefCountPolicyList_ptr policies_;

    //
    // The unique request ID
    //
    CORBA::ULong reqId_;

    //
    // The name of the operation
    //
    // Note for OBE a "const char*" can be used since the operation
    // comes from statically allocated memory.
    //
    const char* op_;
    CORBA::ULong opLen_;

    //
    // The marshalled headers and parameters
    //
    OutputStreamImpl out_;

    //
    // Holds the results of the operation
    //
    InputStreamImpl in_;

    //
    // The state of this invocation
    //
    enum DowncallState
    {
        DowncallStateUnsent,
        DowncallStatePending,
        DowncallStateNoException,
        DowncallStateUserException,
        DowncallStateSystemException,
        DowncallStateFailureException,
        DowncallStateForward
    };
    DowncallState state_;
    
    //
    // Holds the exception if state_ is DowncallStateUserException,
    // DowncallStateSystemException, or DowncallStateFailureException
    //
    CORBA::Exception* ex_;

    //
    // Holds the forward IOR if state_ is DowncallStateLocationForward
    //
    RefCountIOR_var forwardIOR_;

    //
    // Whether a response is expected
    //
    bool responseExpected_;

    //
    // Raise an exception if necessary
    //
    void checkForException()
	throw(OB::ExceptionBase);

public:

    DowncallImpl(ORBInstance_ptr, Client_ptr, RefCountProfileInfo_ptr,
		 RefCountPolicyList_ptr, const char*, CORBA::ULong, bool);
    ~DowncallImpl();

    //
    // Accessors
    //
    ORBInstance_ptr orbInstance_nodup() const
    { return orbInstance_; }
    Client_ptr client_nodup() const
    { return client_; }
    RefCountProfileInfo_ptr profileInfo_nodup() const
    { return profileInfo_; }
    RefCountPolicyList_ptr policies_nodup() const
    { return policies_; }
    CORBA::ULong requestId() const { return reqId_; }
    const char* operation() const { return op_; }
    CORBA::ULong operationLength() const { return opLen_; }
    bool responseExpected() const { return responseExpected_; }

    //
    // Obtain the input and output stream
    //
    OutputStreamImpl* output() { return &out_; }
    InputStreamImpl* input() { return &in_; }

    //
    // Marshalling interception points
    //
    OutputStreamImpl* preMarshal()
	throw(OB::ExceptionBase);
    void marshalEx(const CORBA::SystemException&)
	throw(OB::ExceptionBase);
    void postMarshal()
	throw(OB::ExceptionBase);

    //
    // Request/response operations
    //
    void request()
	throw(OB::ExceptionBase);
    void oneway()
	throw(OB::ExceptionBase);

    //
    // Unmarshalling interception points
    //
    InputStreamImpl* preUnmarshal()
	throw(OB::ExceptionBase);
    void unmarshalEx(const CORBA::SystemException&)
	throw(OB::ExceptionBase);
    void postUnmarshal()
	throw(OB::ExceptionBase);

    //
    // For DowncallStateUserException
    //
    char* unmarshalExceptionId();

    //
    // Check for a specific state
    //
    bool unsent() { return state_ == DowncallStateUnsent; }
    bool pending() { return state_ == DowncallStatePending; }
    bool userException() { return state_ == DowncallStateUserException; }
    bool failureException() { return state_ == DowncallStateFailureException; }

    //
    // Set the state of the downcall
    //
    void setPending()
	throw();
    void setNoException()
	throw();
    void setUserException()
	throw();
    void setUserException(CORBA::UserException*)
	throw();
    void setUserException(const CORBA::UserException&)
	throw();
    void setSystemException(const CORBA::SystemException&)
	throw();
    void setFailureException(const CORBA::SystemException&,
			     CORBA::CompletionStatus)
	throw();
    void setLocationForward(RefCountIOR_ptr)
	throw();
};

} // End of namespace OB

namespace OB
{

//
// The DowncallEmitter class
//
class DowncallEmitter : virtual public RefCount
{
public:

    virtual ~DowncallEmitter() { };

    static inline DowncallEmitter_ptr _duplicate(DowncallEmitter_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline DowncallEmitter_ptr _nil()
    { return 0; }

    //
    // Send and send/receive downcalls
    //
    virtual void send(DowncallImpl*)
        throw() = 0;
    virtual void sendReceive(DowncallImpl*)
        throw() = 0;
};

} // End of namespace OB

namespace CORBA
{

inline void
release(OB::DowncallEmitter_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::DowncallEmitter_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

#endif
