// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_EXCEPT_H
#define OB_EXCEPT_H

#include <OBE/RefCounted_fwd.h>

#include <OBE/Basic.h>
#include <OBE/AnyDeclaration.h>
#include <OBE/TypeCodeConst.h>
#include <OBE/CompletionStatus.h>

namespace CORBA
{

//
// Forward declaration for SystemException
//
class SystemException;

} // End of namespace CORBA

namespace OB
{

//
// The OB::ExceptionBase super class
//
class ExceptionBase
{
public:
    virtual ~ExceptionBase() { }
    virtual void _raise() const = 0;
};

//
// Location forward exception
//
class LocationForward : public ExceptionBase
{
public:

    RefCountIOR_var ior; // Forwarded IOR

    LocationForward(RefCountIOR_ptr);

    virtual void _raise() const { throw *this; }
};

//
// Location forward exception
//
class FailureException : public ExceptionBase
{
public:

    //
    // The system exception causing the failure
    //
    CORBA::SystemException* exception;

    FailureException() : exception(0) { }
    FailureException(const CORBA::SystemException&);
    FailureException(const FailureException&);
    virtual ~FailureException();

    FailureException& operator=(const FailureException&);

    virtual void _raise() const { throw *this; }
};

} // End of namespace OB


namespace CORBA
{

//
// The Exception classes
//
class Exception : public OB::ExceptionBase
{
protected:

    Exception() { }

public:

    //
    // Standard IDL to C++ Mapping
    //
    Exception(const Exception&) { }
    virtual ~Exception() { }
    Exception& operator=(const Exception&) { return *this; }

    virtual void _raise() const = 0;
    virtual const char* _rep_id() const;

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    virtual Exception* _OB_clone() const = 0;
    virtual void _OB_marshal(OB::OutputStreamImpl*) const = 0;
};

} // End of namespace CORBA

namespace CORBA
{

//
// When adding a new system exception, don't forget to add (or to add
// to) the following stuff:
//
// Except.h:
//   OB_MAKE_STANDARD_EXCEPTION
//
// Except.cpp:
//   OB_MAKE_STANDARD_EXCEPTION_I
//   SystemException::_OB_unmarshal
//
// Any.h:
//   OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION
//
// Any.cpp:
//   OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I
//
// Typecode.h:
//   OB_IMPORT extern const OB::TypeCodeSysExConst _tc_*;
//
// Typecode.cpp:
//   const OB::TypeCodeSysExConst _tc_*("*");
//   bool CORBA::TypeCode::_OB_isSystemException() const
//
// Also update the test suite:
//
// ob/test/main/TestIntf.idl:
//   Add void op_*_Ex()
//
// ob/test/main/TestDSI_impl.cpp:
//   Add new case for operation
//
// ob/test/main/TestStubException.cpp:
//   Add new call for exception
//
// ob/test/main/Test_impl.cpp:
//   Add new method
//
class SystemException : public Exception
{
    //
    // ORBacus specific. Not transmitted in remote method calls.
    //
    String_var reason_; 

    //
    // Standard CORBA
    //
    ULong minor_;
    CompletionStatus status_;

public:

    //
    // Standard IDL to C++ Mapping
    //
    SystemException(const char* = "");
    SystemException(ULong, CompletionStatus);
    SystemException(const char*, ULong, CompletionStatus);
    SystemException(const SystemException&);
    SystemException& operator=(const SystemException&);

    ULong minor() const { return minor_; }
    void minor(ULong minor) { minor_ = minor; }
 
    CompletionStatus completed() const { return status_; }
    void completed(CompletionStatus status) { status_ = status; }

    //
    // Additional ORBacus specific functions
    //
    const char* reason() const { return reason_; }
    void reason(const char* reason) { reason_ = reason; }

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    char* _OB_defaultToString(const char*) const;
    virtual void _OB_marshal(OB::OutputStreamImpl*) const;
    static SystemException* _OB_unmarshal(OB::InputStreamImpl*);
};

class UserException : public Exception
{
public:

    //
    // Standard IDL to C++ Mapping
    //
    UserException() { }
#ifdef UNDEFINED
    UserException(const UserException&) { }
    UserException& operator=(const UserException&) { return *this; }
#endif

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    char* _OB_defaultToString(const char*) const;
    virtual void _OB_marshal(OB::OutputStreamImpl*) const;
};

#define OB_MAKE_STANDARD_EXCEPTION(T) \
\
class T : public SystemException \
{ \
public: \
\
    T(const char* r = ""); \
    T(ULong, CompletionStatus); \
    T(const char*, ULong, CompletionStatus); \
    T(const T&); \
\
    virtual void _raise() const { throw *this; } \
    virtual const char* _rep_id() const; \
    virtual Exception* _OB_clone() const; \
};

OB_MAKE_STANDARD_EXCEPTION(UNKNOWN)
OB_MAKE_STANDARD_EXCEPTION(BAD_PARAM)
OB_MAKE_STANDARD_EXCEPTION(NO_MEMORY)
OB_MAKE_STANDARD_EXCEPTION(IMP_LIMIT)
OB_MAKE_STANDARD_EXCEPTION(COMM_FAILURE)
OB_MAKE_STANDARD_EXCEPTION(INV_OBJREF)
OB_MAKE_STANDARD_EXCEPTION(NO_PERMISSION)
OB_MAKE_STANDARD_EXCEPTION(INTERNAL)
OB_MAKE_STANDARD_EXCEPTION(MARSHAL)
OB_MAKE_STANDARD_EXCEPTION(INITIALIZE)
OB_MAKE_STANDARD_EXCEPTION(NO_IMPLEMENT)
OB_MAKE_STANDARD_EXCEPTION(BAD_TYPECODE)
OB_MAKE_STANDARD_EXCEPTION(BAD_OPERATION)
OB_MAKE_STANDARD_EXCEPTION(NO_RESOURCES)
OB_MAKE_STANDARD_EXCEPTION(NO_RESPONSE)
OB_MAKE_STANDARD_EXCEPTION(BAD_INV_ORDER)
OB_MAKE_STANDARD_EXCEPTION(TRANSIENT)
OB_MAKE_STANDARD_EXCEPTION(OBJ_ADAPTER)
OB_MAKE_STANDARD_EXCEPTION(OBJECT_NOT_EXIST)
OB_MAKE_STANDARD_EXCEPTION(INV_POLICY)

} // End of namespace CORBA

namespace OB
{

enum VMCID
{
    OMGVMCID = 0x4f4d0000,
    OOCVMCID = 0x4f4f0000
};

//
// Minor error codes for INITIALIZE
//
// Note: If something is changed here, it must also be changed in
// other ORBacus implementations, like ORBacus for Java. 
//
enum MinorInitialize
{
    //
    // ORBacus specific minor codes
    //
    MinorORBDestroyed =  1 | OOCVMCID
};

//
// Minor error codes for NO_IMPLEMENT
//
// Note: If something is changed here, it must also be changed in
// other ORBacus implementations, like ORBacus for Java.
//
enum MinorNoImplement
{
    //
    // Standard minor codes
    //
    MinorMissingLocalValueImplementation = 1 | OMGVMCID,
    MinorIncompatibleValueImplementationVersion = 2 | OMGVMCID
};

//
// Minor error codes for OBJECT_NOT_EXIST
//
// Note: If something is changed here, it must also be changed in
// other ORBacus implementations, like ORBacus for Java.
//
enum MinorObjectNotExist
{
    //
    // Standard minor codes
    //
    MinorUnregisteredValue = 1 | OMGVMCID
};

//
// Minor error codes for BAD_PARAM
//
// Note: If something is changed here, it must also be changed in
// other ORBacus implementations, like ORBacus for Java.
//
enum MinorBadParam
{
    //
    // Standard minor codes
    //
    MinorBadSchemeName = 7 | OMGVMCID,
    MinorBadAddress = 8 | OMGVMCID,
    MinorBadSchemeSpecificPart = 9 | OMGVMCID,
    MinorOther = 10 | OMGVMCID
};

//
// Minor error codes for BAD_INV_ORDER
//
// Note: If something is changed here, it must also be changed in
// other ORBacus implementations, like ORBacus for Java.
//
enum MinorBadInvOrder
{
    //
    // Standard minor codes
    //
    MinorDestroyWouldBlock = 3 | OMGVMCID,
    MinorShutdownCalled = 4 | OMGVMCID,

    //
    // ORBacus specific minor codes
    //
    MinorORBRunning = 2 | OOCVMCID,
    MinorRequestAlreadySent = 3 | OOCVMCID,
    MinorRequestNotSent = 4 | OOCVMCID,
    MinorResponseAlreadyReceived = 5 | OOCVMCID
};

//
// Minor error codes for COMM_FAILURE
//
// Note: If something is changed here, it must also be changed in
// other ORBacus implementations, like ORBacus for Java.
//
enum MinorCommFailure
{
    //
    // Standard minor codes
    //
    /* None yet */

    //
    // ORBacus specific minor codes
    //
    MinorRecv = 1 | OOCVMCID,
    MinorSend = 2 | OOCVMCID,
    MinorRecvZero = 3 | OOCVMCID,
    MinorSendZero = 4 | OOCVMCID,
    MinorSocket = 5 | OOCVMCID,
    MinorSetsockopt = 6 | OOCVMCID,
    MinorGetsockopt = 7 | OOCVMCID,
    MinorBind = 8 | OOCVMCID,
    MinorListen = 9 | OOCVMCID,
    MinorConnect = 10 | OOCVMCID,
    MinorAccept = 11 | OOCVMCID,
    MinorSelect = 12 | OOCVMCID,
    MinorSetSoTimeout = 26 | OOCVMCID,

    MinorGethostname = 13 | OOCVMCID,
    MinorGethostbyname = 14 | OOCVMCID,

    MinorWSAStartup = 15 | OOCVMCID,
    MinorWSACleanup = 16 | OOCVMCID,

    MinorNoGIOP = 17 | OOCVMCID,
    MinorUnknownMessage = 18 | OOCVMCID,
    MinorWrongMessage = 19 | OOCVMCID,
    MinorMessageError = 21 | OOCVMCID,
    MinorFragment = 22 | OOCVMCID,
    MinorUnknownReqId = 24 | OOCVMCID,
    MinorVersion = 25 | OOCVMCID
};

//
// Minor error codes for MARSHAL
//
// Note: If something is changed here, it must also be changed in
// other ORBacus implementations, like ORBacus for Java.
//
enum MinorMarshal
{
    //
    // Standard minor codes
    //
    MinorNoValueFactory = 1 | OMGVMCID, // No value factory

    //
    // ORBacus specific minor codes
    //
    MinorReadOverflow = 1 | OOCVMCID,
    MinorReadBooleanOverflow = 2 | OOCVMCID,
    MinorReadCharOverflow = 3 | OOCVMCID,
    MinorReadOctetOverflow = 5 | OOCVMCID,
    MinorReadShortOverflow = 6 | OOCVMCID,
    MinorReadUShortOverflow = 7 | OOCVMCID,
    MinorReadLongOverflow = 8 | OOCVMCID,
    MinorReadULongOverflow = 9 | OOCVMCID,
    MinorReadLongLongOverflow = 10 | OOCVMCID,
    MinorReadULongLongOverflow = 11 | OOCVMCID,
    MinorReadFloatOverflow = 12 | OOCVMCID,
    MinorReadDoubleOverflow = 13 | OOCVMCID,
    MinorReadLongDoubleOverflow = 14 | OOCVMCID,
    MinorReadStringOverflow = 15 | OOCVMCID,
    MinorReadStringZeroLength = 16 | OOCVMCID,
    MinorReadStringNullChar = 17 | OOCVMCID,
    MinorReadStringNoTerminator = 18 | OOCVMCID,
    MinorReadBooleanArrayOverflow = 25 | OOCVMCID,
    MinorReadCharArrayOverflow = 26 | OOCVMCID,
    MinorReadOctetArrayOverflow = 28 | OOCVMCID,
    MinorReadShortArrayOverflow = 29 | OOCVMCID,
    MinorReadUShortArrayOverflow = 30 | OOCVMCID,
    MinorReadLongArrayOverflow = 31 | OOCVMCID,
    MinorReadULongArrayOverflow = 32 | OOCVMCID,
    MinorReadLongLongArrayOverflow = 33 | OOCVMCID,
    MinorReadULongLongArrayOverflow = 34 | OOCVMCID,
    MinorReadFloatArrayOverflow = 35 | OOCVMCID,
    MinorReadDoubleArrayOverflow = 36 | OOCVMCID,
    MinorReadLongDoubleArrayOverflow = 37 | OOCVMCID,
    MinorReadInvTypeCodeIndirection = 38 | OOCVMCID,
    MinorWriteObjectLocal = 39 | OOCVMCID,
    MinorLongDoubleNotSupported = 40 | OOCVMCID
};

//
// Minor error codes for IMP_LIMIT
//
// Note: If something is changed here, it must also be changed in
// other ORBacus implementations, like ORBacus for Java.
//
enum MinorImpLimit
{
    //
    // ORBacus specific minor codes
    //
    MinorMessageSizeLimit = 1 | OOCVMCID,
    MinorThreadLimit = 2 | OOCVMCID
};

//
// Minor error codes for NO_MEMORY
//
// Note: If something is changed here, it must also be changed in
// other ORBacus implementations, like ORBacus for Java.
//
enum MinorNoMemory
{
    //
    // ORBacus specific minor codes
    //
    MinorAllocationFailure = 1 | OOCVMCID
};

//
// Minor error codes for TRANSIENT
//
// Note: If something is changed here, it must also be changed in
// other ORBacus implementations, like ORBacus for Java.
//
enum MinorTransient
{
    //
    // PortableInterceptor minor codes
    //
    MinorRequestCancelled = 1 | OMGVMCID,

    //
    // ORBacus specific minor codes
    //
    MinorConnectFailed = 1 | OOCVMCID,
    MinorCloseConnection = 2 | OOCVMCID,
    MinorActiveConnectionManagement = 3 | OOCVMCID,
    MinorForcedShutdown = 4 | OOCVMCID
};

//
// Minor error codes for UNKNOWN
//
// Note: If something is changed here, it must also be changed in
// other ORBacus implementations, like ORBacus for Java.
//
enum MinorUnknown
{
    //
    // PortableInterceptor minor codes
    //
    MinorUnknownUserException = 1 | OMGVMCID
};

} // End of namespace OB

#endif
