// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_HASHTABLE_I_H
#define OB_HASHTABLE_I_H

#include <OBE/Hashtable.h>

namespace OB
{

// ----------------------------------------------------------------------
// Hashtable constructor and destructor
// ----------------------------------------------------------------------

template<class K, class T, class H>
Hashtable<K, T, H>::Hashtable(CORBA::ULong len, CORBA::Float load)
    : length_(len), load_(load), count_(0)
{
    threshold_ = (CORBA::ULong)(length_ * load_);
    table_ = new node*[length_];
    for(CORBA::ULong i = 0 ; i < length_ ; i++)
        table_[i] = 0;
}

template<class K, class T, class H>
Hashtable<K, T, H>::~Hashtable()
{
    removeAll();
    delete []table_;
}

// ----------------------------------------------------------------------
// Hashtable public member implementation
// ----------------------------------------------------------------------

template<class K, class T, class H>
void
Hashtable<K, T, H>::clear()
{
    removeAll();

    for(CORBA::ULong i = 0 ; i < length_ ; i++)
        table_[i] = 0;
}

template<class K, class T, class H>
bool
Hashtable<K, T, H>::containsKey(const K& key) const
{
    CORBA::ULong fullHash = H::hash(key);
    CORBA::ULong h = fullHash % length_;

    node* n = table_[h];
    while(n != 0)
    {
        if(n -> fullHash == fullHash && H::comp(n -> key, key))
            return true;
        n = n -> next;
    }

    return false;
}

template<class K, class T, class H>
void
Hashtable<K, T, H>::put(const K& key, const T& value)
{
    CORBA::ULong fullHash = H::hash(key);
    CORBA::ULong h = fullHash % length_;

    node* n = table_[h];
    while(n != 0)
    {
        if(n -> fullHash == fullHash && H::comp(n -> key, key))
        {
            n -> value = value;
            return;
        }

        n = n -> next;
    }

    //
    // Insert at head of bucket
    //
    n = new node(key, value, table_[h]);
    table_[h] = n;

    count_++;
    if(count_ > threshold_)
        resize();
}

template<class K, class T, class H>
void
Hashtable<K, T, H>::remove(const K& key)
{
    CORBA::ULong fullHash = H::hash(key);
    CORBA::ULong h = fullHash % length_;

    node* prev = 0;
    node* n = table_[h];

    while(n != 0)
    {
        if(n -> fullHash == fullHash && H::comp(n -> key, key))
            break;

        prev = n;
        n = n -> next;
    }

    if(n != 0)
    {
        if(prev != 0)
            prev -> next = n -> next;
        else
            table_[h] = n -> next;

        count_--;
        delete n;
    }
}

template<class K, class T, class H>
bool
Hashtable<K, T, H>::get(const K& key, T& value) const
{
    CORBA::ULong fullHash = H::hash(key);
    CORBA::ULong h = fullHash % length_;

    node* n = table_[h];
    while(n != 0)
    {
        if(n -> fullHash == fullHash && H::comp(n -> key, key))
        {
            value = n -> value;
            return true;
        }

        n = n -> next;
    }

    return false;
}

#ifdef HAVE_NO_TYPENAME
template<class K, class T, class H>
Hashtable<K, T, H>::Enumerator
Hashtable<K, T, H>::keys() const
#else
template<class K, class T, class H>
typename Hashtable<K, T, H>::Enumerator
Hashtable<K, T, H>::keys() const
#endif
{
    CORBA::ULong index = 0;
    K* keys = new K[count_];

    for(CORBA::ULong i = 0 ; i < length_ ; i++)
    {
        for(node* n = table_[i] ; n != 0 ; n = n -> next)
            keys[index++] = n -> key;
    }

    Enumerator e(keys, count_);

    return e;
}

template<class K, class T, class H>
CORBA::ULong
Hashtable<K, T, H>::size() const
{
    return count_;
}

template<class K, class T, class H>
char*
Hashtable<K, T, H>::displayStats()
{
    CORBA::ULong most = 0;
    CORBA::ULong occupied = 0;
    for(CORBA::ULong i = 0 ; i < length_ ; i++)
    {
        CORBA::ULong bucketLen = 0;
        for(node* n = table_[i] ; n != 0 ; n = n -> next, bucketLen++)
            ;
        if(bucketLen > most)
            most = bucketLen;
        if(bucketLen > 0)
            occupied++;
    }
    CORBA::String_var ret;
    ret += CORBA::string_dup("Hash table: ");
    ret += length_;
    ret += CORBA::string_dup(" buckets.\nKeys      : ");
    ret += count_;
    ret += CORBA::string_dup("\nOccupied  : ");
    ret += occupied;
    ret += CORBA::string_dup("\nLargest   : ");
    ret += most;
    ret += CORBA::string_dup("\n");

    return ret._retn();
}

// ----------------------------------------------------------------------
// Hashtable private member implementation
// ----------------------------------------------------------------------

template<class K, class T, class H>
void
Hashtable<K, T, H>::removeAll()
{
    if(table_ != 0)
    {
	CORBA::ULong i;

	for(i = 0; i < length_; i++)
	{
	    node* n = table_[i];

	    while(n)
	    {
		node *n1 = n;
		n = n -> next;
		delete n1;
	    }

	    table_[i] = 0;
	}

	count_ = 0;
    }
}

template<class K, class T, class H>
void
Hashtable<K, T, H>::resize()
{
    CORBA::ULong i;
    CORBA::ULong size = length_ * 2 + 1;

    //
    // Create new table
    //
    node** table = new node*[size];
    for(i = 0 ; i < size ; i++)
        table[i] = 0;

    //
    // Migrate nodes from existing table to new one
    //
    for(i = 0 ; i < length_ ; i++)
    {
        node* n = table_[i];
        while(n != 0)
        {
            node* next = n -> next;
            CORBA::ULong h = n -> fullHash % size;
            n -> next = table[h];
            table[h] = n;
            n = next;
        }
    }

    //
    // Update state
    //
    length_ = size;
    threshold_ = (CORBA::ULong)(length_ * load_);
    delete []table_;
    table_ = table;
}

} // End of namespace OB

#endif
