// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_ORB_INSTANCE_H
#define OB_ORB_INSTANCE_H

#include <OBE/ORBInstance_fwd.h>
#include <OBE/Object_fwd.h>
#include <OBE/Client_fwd.h>
#include <OBE/InitialServiceManager_fwd.h>
#include <OBE/Properties_fwd.h>
#include <OBE/Logger_fwd.h>
#ifndef HAVE_OBE_NO_TRACE
#  include <OBE/CoreTraceLevels_fwd.h>
#endif // !HAVE_OBE_NO_TRACE
#include <OBE/OCIClient_fwd.h>
#include <OBE/Except.h>

namespace CORBA
{

class ORB; // For the friend declaration

} // End of namespace OBCORBA

namespace OB
{

class GlobalStateMutexDestroyer;

//
// The ORBInstance class. This class only holds references to ORB
// internal class instances. It should not have any behaviour other
// than managing the lifecycle of these objects, and setting/cleaning
// up global state objects.
//
class ORBInstance : public RefCount
{
    //
    // Hide copy-constructor and assignment operator
    //
    ORBInstance(const ORBInstance&);
    void operator=(const ORBInstance&);

    //
    // ORBInstance members
    //
    // NOTE: For each new member "X_var x_", add the following to
    // the destroy() operation:
    //
    // x_ -> destroy(); // Only if X has an operation destroy()
    // x_ = X::_nil(); // To resolve potential cyclic dependencies
    //
    ObjectFactory_var objectFactory_;
    ClientManager_var clientManager_;
    InitialServiceManager_var initServiceManager_;
    Properties_var properties_;
    Logger_var logger_;
#ifndef HAVE_OBE_NO_TRACE
    CoreTraceLevels_var coreTraceLevels_;
#endif // !HAVE_OBE_NO_TRACE
    OCI::ConFactoryRegistry_var conFactoryRegistry_;

    //
    // True if destroy() was called
    //
    bool destroy_;

    //
    // Create and destroy global state
    //
    void createGlobalState();
    void destroyGlobalState();

    //
    // Initialize the ORBInstance
    //
    void initialize(Properties_ptr, Logger_ptr
#ifndef HAVE_OBE_NO_TRACE
		    , CoreTraceLevels_ptr
#endif // !HAVE_OBE_NO_TRACE
		    );

    //
    // Destroy the ORBInstance
    //
    void destroy();

    friend class CORBA::ORB; // only CORBA::ORB can call initialize()
			     // and destroy
protected:

    //
    // Global state management
    //
    friend class ::OB::GlobalStateMutexDestroyer;
    static JTCMutex* globalStateMutex_;
    static int globalStateCounter_;

    //
    // Initialize and destroy hooks used by derived classes (called by
    // initialize() and destroy() respectively)
    //
    virtual void initializeHook() = 0;
    virtual void destroyHook() = 0;

    //
    // Create the ClientManager
    //
    virtual ClientManager_ptr createClientManager(ORBInstance_ptr) = 0;

    ORBInstance();
    virtual ~ORBInstance();

public:

    static inline ORBInstance_ptr _duplicate(ORBInstance_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline ORBInstance_ptr _nil()
    { return 0; }

    //
    // Get the ObjectFactory
    //
    ObjectFactory_ptr getObjectFactory();

    //
    // Get the ClientManager
    //
    ClientManager_ptr getClientManager();

    //
    // Get the InitialServiceManager
    //
    InitialServiceManager_ptr getInitialServiceManager();

    //
    // Get the Properties
    //
    Properties_ptr getProperties();

    //
    // Get the message viewer
    //
    Logger_ptr getLogger();

#ifndef HAVE_OBE_NO_TRACE
    //
    // Get the core trace levels
    //
    CoreTraceLevels_ptr getCoreTraceLevels();
#endif // !HAVE_OBE_NO_TRACE

    //
    // Get the OCI connector factory registry
    //
    OCI::ConFactoryRegistry_ptr getConFactoryRegistry();
};

} // End of namespace OB

namespace CORBA
{

inline void
release(OB::ORBInstance_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::ORBInstance_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

#endif
