// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OBE_POA_H
#define OBE_POA_H

#include <OBE/POA_fwd.h>
#include <OBE/POAManager_fwd.h>

#include <OBE/OCITypes.h>

#include <OBE/NativeTypes.h>
#include <OBE/POAServantBase.h>
#include <OBE/POACommon.h>

#include <OBE/Defs.h>

namespace OB
{

//
// Forward declaration
//
struct ObjectKeyData;

};

namespace PortableServer
{

struct ActiveObjectEntry
{
    //
    // The servant
    //
    PortableServer::ServantBase_var servant;
    
    //
    // Slot iteration count
    //
    CORBA::ULong iter;

    //
    // The Mutex
    //
    JTCMutex mutex;

    ActiveObjectEntry();
};

class POA : public CORBA::Object
{
public:

    //
    // Exceptions are needed first
    //

    //
    // IDL:omg.org/PortableServer/POA/ObjectAlreadyActive:1.0
    //
    struct ObjectAlreadyActive : public CORBA::UserException
    {
        ObjectAlreadyActive() { }

        virtual void _raise() const { throw *this; }
        virtual const char* _rep_id() const;

        virtual CORBA::Exception* _OB_clone() const;
    };

    //
    // IDL:omg.org/PortableServer/POA/ObjectNotActive:1.0
    //
    struct ObjectNotActive : public CORBA::UserException
    {
        ObjectNotActive() { }

        virtual void _raise() const { throw *this; }
        virtual const char* _rep_id() const;

        virtual CORBA::Exception* _OB_clone() const;
    };

    //
    // IDL:omg.org/PortableServer/POA/ServantNotActive:1.0
    //
    struct ServantNotActive : public CORBA::UserException
    {
        ServantNotActive() { }

        virtual void _raise() const { throw *this; }
        virtual const char* _rep_id() const;

        virtual CORBA::Exception* _OB_clone() const;
    };

    //
    // IDL:omg.org/PortableServer/POA/WrongAdapter:1.0
    //
    struct WrongAdapter : public CORBA::UserException
    {
        WrongAdapter() { }

        virtual void _raise() const { throw *this; }
        virtual const char* _rep_id() const;

        virtual CORBA::Exception* _OB_clone() const;
    };

private:

    POA(const POA&);
    void operator=(const POA&);

    //
    // The ORBInstance object
    //
    OB::ORBInstance_var orbInstance_;

    //
    // My POA manager
    //
    PortableServer::POAManager_var manager_;

    //
    // The AOM
    //
    PortableServer::ActiveObjectEntry activeObjectArray_[MAX_ACTIVE_OBJECTS];
    CORBA::ULong nextSlot_;
    CORBA::ULong freeSlots_;
    JTCMutex AOMMutex_;

    //
    // Handle unknown exceptions
    //
    void handleUnknownException(OB::UpcallImpl*);

    //
    // Return the Servant associated with the ObjectId
    //
    PortableServer::ServantBase* idToServant(const PortableServer::ObjectId&);

    //
    // Locate the servant with the ObjectId
    //
    PortableServer::ServantBase* locate(const OB::ObjectKeyData&)
	throw(CORBA::SystemException, OB::LocationForward);

    //
    // Create IOR
    //
    OB::RefCountIOR_ptr createIOR(const OCI::ObjectKey&, const char* = "");

    //
    // Convert ObjectId to ObjectKeyData
    //
    bool ObjectIdToObjectKey(const PortableServer::ObjectId&,
			     OB::ObjectKeyData&);

    //
    // Convert from ObjectKeyData to ObjectId
    //
    PortableServer::ObjectId* ObjectKeyToObjectId(const OB::ObjectKeyData&);

    //
    // Activate an object
    //
    PortableServer::ObjectId* activate(PortableServer::Servant servant);
    
public:

    //
    // Public constructor for the POA
    //
    POA(OB::ORBInstance_ptr, PortableServer::POAManager_ptr);
    ~POA();

    typedef POA_ptr _ptr_type;
    typedef POA_var _var_type;

    static inline POA_ptr
    _duplicate(POA_ptr p) { if(p) p -> _OB_incRef(); return p; }

    static inline POA_ptr _nil() { return 0; }

    static POA_ptr _narrow(CORBA::Object_ptr);
    virtual bool _OB_isLC() const { return true; };

    //
    // IDL to C++ mapping
    //
    POAManager_ptr the_POAManager()
	throw(CORBA::SystemException);

    ObjectId* activate_object(Servant)
	throw(CORBA::SystemException);
    
    void deactivate_object(const ObjectId&)
	throw(ObjectNotActive, CORBA::SystemException);

    ObjectId* servant_to_id(Servant)
	throw(ServantNotActive, CORBA::SystemException);

    CORBA::Object_ptr servant_to_reference(Servant)
	throw(ServantNotActive, CORBA::SystemException);

    Servant reference_to_servant(CORBA::Object_ptr)
	throw(ObjectNotActive, WrongAdapter,
	      CORBA::SystemException);

    ObjectId* reference_to_id(CORBA::Object_ptr)
	throw(WrongAdapter, CORBA::SystemException);

    Servant id_to_servant(const ObjectId&)
	throw(ObjectNotActive, CORBA::SystemException);

    CORBA::Object_ptr id_to_reference(const ObjectId&)
	throw(ObjectNotActive, CORBA::SystemException);

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    
    //
    // Called when POA being destroyed
    //
    void _OB_destroy();

    //
    // Determine if a servant is present
    //
    bool _OB_locateServant(const OB::ObjectKeyData&)
	throw(CORBA::SystemException, OB::LocationForward);

    //
    // Dispatch a method invocation
    //
    void _OB_dispatch(const OB::ObjectKeyData&, OB::UpcallImpl*)
	throw();

    //
    // Create an object reference given an object key and interface name
    //
    CORBA::Object_ptr _OB_createReference(const PortableServer::ObjectId&,
					  const char*);
};

} // End of namespace PortableServer

namespace CORBA
{

inline void
release(PortableServer::POA_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

inline Boolean
is_nil(PortableServer::POA_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

#endif
