// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_STREAM_H
#define OB_STREAM_H

#include <OBE/Principal_fwd.h>
#include <OBE/Object_fwd.h>
#include <OBE/TypeCode_fwd.h>
#include <OBE/ORBInstance_fwd.h>
#include <OBE/OCIBuffer.h>

#include <OBE/Any.h>
#include <OBE/Except.h>
#include <OBE/Hashtable.h>

namespace OB
{

class TypeCodeReader;
class TypeCodeWriter;

//
// The InputStreamImpl class (not reference counted)
//
class InputStreamImpl
{
    //
    // Hide copy-constructor and assignment operator
    //
    InputStreamImpl(const InputStreamImpl&);
    InputStreamImpl& operator=(const InputStreamImpl&);

    //
    // Read/write primitives
    //
    void readBooleanArray(CORBA::Boolean*& value, CORBA::ULong length,
			  bool copy)
    {
        if(length)
        {
            assert_nca(!copy || (copy && value != 0), NCANullValue);

            register CORBA::Octet* nextCur = buf_.cur_ + length;
            if(nextCur > buf_.last_)
                throw CORBA::MARSHAL(MinorReadBooleanArrayOverflow,
                                     CORBA::COMPLETED_NO);

	    if(copy)
		memcpy(value, buf_.cur_, length);
	    else
		value = (CORBA::Boolean*)buf_.cur_;
            buf_.cur_ = nextCur;
        }
    }

    void readOctetArray(CORBA::Octet*& value, CORBA::ULong length,
			bool copy)
    {
        if(length)
        {
            assert_nca(!copy || (copy && value != 0), NCANullValue);

            register CORBA::Octet* nextCur = buf_.cur_ + length;
            if(nextCur > buf_.last_)
                throw CORBA::MARSHAL(MinorReadOctetArrayOverflow,
                                     CORBA::COMPLETED_NO);
	    if(copy)
		memcpy(value, buf_.cur_, length);
	    else
		value = buf_.cur_;
            buf_.cur_ = nextCur;
        }
    }

    void
    readCharArray(CORBA::Char*& value, CORBA::ULong length, bool copy)
    {
	if(length)
	{
            assert_nca(!copy || (copy && value != 0), NCANullValue);
	    
	    if(buf_.cur_ + length > buf_.last_)
		throw CORBA::MARSHAL(MinorReadCharArrayOverflow,
				     CORBA::COMPLETED_NO);
	    
	    if(copy)
		memcpy(value, buf_.cur_, length);
	    else
		value = (CORBA::Char*)buf_.cur_;
	    
	    buf_.cur_ += length;
	}
    }

    void readShortArray(CORBA::Short*&, CORBA::ULong, bool);
    void readUShortArray(CORBA::UShort*&, CORBA::ULong, bool);
    void readLongArray(CORBA::Long*&, CORBA::ULong, bool);
    void readULongArray(CORBA::ULong*&, CORBA::ULong, bool);
    void readLongLongArray(CORBA::LongLong*&, CORBA::ULong, bool);
    void readULongLongArray(CORBA::ULongLong*&, CORBA::ULong, bool);
    void readFloatArray(CORBA::Float*&, CORBA::ULong, bool);
    void readDoubleArray(CORBA::Double*&, CORBA::ULong, bool);
    void readLongDoubleArray(CORBA::LongDouble*&, CORBA::ULong, bool);

protected:

    ORBInstance* orbInstance_;
    OCI::BufferImpl buf_;
    CORBA::Octet* origCur_;
    bool origSwap_;
    bool swap_;

    friend class OutputStreamImpl;
    friend class TypeCodeReader;

public:

    InputStreamImpl(CORBA::ULong, CORBA::ULong, bool);
    InputStreamImpl(CORBA::Octet*, CORBA::ULong);

    OCI::BufferImpl* _OB_bufferImpl() { return &buf_; }

    CORBA::ULong _OB_pos() { return buf_.pos(); }
    void _OB_pos(CORBA::ULong pos) { buf_.pos(pos); }

    bool _OB_swap() const { return swap_; }
    void _OB_swap(bool s) { swap_ = s; }

    InputStreamImpl* _OB_clone();

    void _OB_reset();

    void _OB_skip(CORBA::ULong length)
    {
        register CORBA::Octet* newCur = buf_.cur_ + length;
        if(newCur > buf_.last_)
            throw CORBA::MARSHAL(MinorReadOverflow, CORBA::COMPLETED_NO);
        buf_.cur_ = newCur;
    }

    void _OB_skipAlign(CORBA::ULong align)
    {
        register CORBA::Octet* newCur = buf_.cur_ + align - 1;
        newCur -= (unsigned long)newCur % align;
        if(newCur > buf_.last_)
            throw CORBA::MARSHAL(MinorReadOverflow, CORBA::COMPLETED_NO);
        buf_.cur_ = newCur;
    }

    void _OB_ORBInstance(ORBInstance_ptr);
    ORBInstance_ptr _OB_ORBInstance_nodup();

    void _OB_readEndian();

    CORBA::Long _OB_readLongUnchecked();

    CORBA::Boolean read_boolean()
    {
        if(buf_.cur_ > buf_.last_)
            throw CORBA::MARSHAL(MinorReadBooleanOverflow,
                                 CORBA::COMPLETED_NO);

        CORBA::Boolean val = (*(buf_.cur_)++) ? true : false;
        return val;
    }

    CORBA::Octet read_octet()
    {
        if(buf_.cur_ > buf_.last_)
            throw CORBA::MARSHAL(MinorReadOctetOverflow, CORBA::COMPLETED_NO);

        CORBA::Octet val = *(buf_.cur_)++;
        return val;
    }

    CORBA::Short read_short();
    CORBA::UShort read_ushort();
    CORBA::Long read_long();
    CORBA::ULong read_ulong();
    CORBA::LongLong read_longlong();
    CORBA::ULongLong read_ulonglong();
    CORBA::Float read_float();
    CORBA::Double read_double();
    CORBA::LongDouble read_longdouble();
    CORBA::Char read_char();

    void read_char_array(CORBA::Char* v, CORBA::ULong l)
    {
	readCharArray(v, l, true);
    }

    void read_char_array_nodup(CORBA::Char*& v, CORBA::ULong l)
    {
	readCharArray((CORBA::Char*&)v, l, false);
    }

    char* read_string();
    const char* read_string_nodup();
    CORBA::Object_ptr read_Object();
    CORBA::TypeCode_ptr read_TypeCode();
    CORBA::Principal_ptr read_Principal();
    CORBA::Any* read_any();
    void read_any(CORBA::Any&);

    void read_boolean_array(CORBA::Boolean* value, CORBA::ULong length)
    {
	readBooleanArray(value, length, true);
    }

    void read_octet_array(CORBA::Octet* value, CORBA::ULong length)
    {
	readOctetArray(value, length, true);
    }

    void read_short_array(CORBA::Short* v, CORBA::ULong l)
    {
	readShortArray(v, l, true);
    }

    void read_ushort_array(CORBA::UShort* v, CORBA::ULong l)
    {
	readUShortArray(v, l, true);
    }

    void read_long_array(CORBA::Long* v, CORBA::ULong l)
    {
	readLongArray(v, l, true);
    }

    void read_ulong_array(CORBA::ULong* v, CORBA::ULong l)
    {
	readULongArray(v, l, true);
    }

    void read_longlong_array(CORBA::LongLong* v, CORBA::ULong l)
    {
	readLongLongArray(v, l, true);
    }

    void read_ulonglong_array(CORBA::ULongLong* v, CORBA::ULong l)
    {
	readULongLongArray(v, l, true);
    }

    void read_float_array(CORBA::Float* v, CORBA::ULong l)
    {
	readFloatArray(v, l, true);
    }

    void read_double_array(CORBA::Double* v, CORBA::ULong l)
    {
	readDoubleArray(v, l, true);
    }

    void read_longdouble_array(CORBA::LongDouble* v, CORBA::ULong l)
    {
	readLongDoubleArray(v, l, true);
    }

    void read_boolean_array_nodup(CORBA::Boolean*& value, CORBA::ULong l)
    {
	readBooleanArray((CORBA::Boolean*&)value, l, false);
    }

    void read_octet_array_nodup(CORBA::Octet*& value, CORBA::ULong l)
    {
	readOctetArray((CORBA::Octet*&)value, l, false);
    }

    void read_short_array_nodup(CORBA::Short*& v, CORBA::ULong l)
    {
	readShortArray((CORBA::Short*&)v, l, false);
    }

    void read_ushort_array_nodup(CORBA::UShort*& v, CORBA::ULong l)
    {
	readUShortArray((CORBA::UShort*&)v, l, false);
    }

    void read_long_array_nodup(CORBA::Long*& v, CORBA::ULong l)
    {
	readLongArray((CORBA::Long*&)v, l, false);
    }

    void read_ulong_array_nodup(CORBA::ULong*& v, CORBA::ULong l)
    {
	readULongArray((CORBA::ULong*&)v, l, false);
    }

    void read_longlong_array_nodup(CORBA::LongLong*& v, CORBA::ULong l)
    {
	readLongLongArray((CORBA::LongLong*&)v, l, false);
    }

    void read_ulonglong_array_nodup(CORBA::ULongLong*& v, CORBA::ULong l)
    {
	readULongLongArray((CORBA::ULongLong*&)v, l, false);
    }

    void read_float_array_nodup(CORBA::Float*& v, CORBA::ULong l)
    {
	readFloatArray((CORBA::Float*&)v, l, false);
    }

    void read_double_array_nodup(CORBA::Double*& v, CORBA::ULong l)
    {
	readDoubleArray((CORBA::Double*&)v, l, false);
    }

    void read_longdouble_array_nodup(CORBA::LongDouble*& v, CORBA::ULong l)
    {
	readLongDoubleArray((CORBA::LongDouble*&)v, l, false);
    }

    void read_any_NoTypeCode(CORBA::Any&, CORBA::TypeCode_ptr);
};

//
// The OutputStreamImpl class (not reference counted)
//
class OutputStreamImpl
{
    //
    // Hide copy-constructor and assignment operator
    //
    OutputStreamImpl(const OutputStreamImpl&);
    OutputStreamImpl& operator=(const OutputStreamImpl&);

protected:

    OCI::BufferImpl buf_;

    //
    // If alignNext_ > 0, the next write should be aligned on this
    // boundary
    //
    CORBA::ULong alignNext_;

    CORBA::ULong writeGap();
    void writeLength(CORBA::ULong);

    void addCapacity(CORBA::ULong size)
    {
        //
        // Expand buffer to hold requested size
        //
        // Note: OutputStreamImpls are not always written to in a
        //       linear fashion, i.e., sometimes the position is reset
        //       to an earlier point and data is patched
        //       in. Therefore, do NOT do this:
        //
        //buf_.realloc(buf_.len_ + size);
        //
        // 
        if(alignNext_ > 0)
        {
            CORBA::ULong align = alignNext_;
            alignNext_ = 0;
            addCapacity(size, align);
        }
        else
        {
            //
            // If there isn't enough room, then reallocate the buffer
            //
            CORBA::ULong len = buf_.cur_ - buf_.data_ + size;
            if(len > buf_.len_)
                buf_.realloc(len);
        }
    }

    void addCapacity(CORBA::ULong size, CORBA::ULong align)
    {
        assert(align > 0); // use addCapacity(ULong) if align == 0

        //
        // If alignNext_ is set, then use the larger of alignNext_ and align
        //
        if(alignNext_ > 0)
        {
            align = (alignNext_ > align ? alignNext_ : align);
            alignNext_ = 0;
        }

        //
        // Align to the requested boundary
        //
        register CORBA::Octet* newCur = buf_.cur_ + align - 1;
        newCur -= CORBA::ULong((long)newCur) % align;
        buf_.cur_ = newCur;

        //
        // If there isn't enough room, then reallocate the buffer
        //
        if(newCur + size > buf_.last_)
            buf_.realloc(newCur - buf_.data_ + size);
    }

    friend class InputStreamImpl;
    friend class TypeCodeWriter;

public:

    OutputStreamImpl(CORBA::ULong);
    OutputStreamImpl();

    InputStreamImpl* create_input_stream();

    OCI::BufferImpl* _OB_bufferImpl() { return &buf_; }

    CORBA::ULong _OB_pos() { return buf_.pos(); }
    void _OB_pos(CORBA::ULong pos) { buf_.pos(pos); }

    void _OB_align(CORBA::ULong n)
    {
        if(CORBA::ULong((long)(buf_.cur_)) % n != 0)
            addCapacity(0, n);
    }

    void _OB_alignNext(CORBA::ULong n)
    {
        alignNext_ = n;
    }

    void _OB_writeEndian() { write_boolean(Endian); }

    void write_boolean(CORBA::Boolean value)
    {
        addCapacity(1);
        *(buf_.cur_) = value;
        buf_.cur_ += 1;
    }

    void write_octet(CORBA::Octet value)
    {
        addCapacity(1);
        *(buf_.cur_) = value;
        buf_.cur_ += 1;
    }

    void write_short(CORBA::Short value)
    {
        addCapacity(2, 2);
        *(CORBA::Short*)(buf_.cur_) = value;
        buf_.cur_ += 2;
    }

    void write_ushort(CORBA::UShort value)
    {
        addCapacity(2, 2);
        *(CORBA::UShort*)(buf_.cur_) = value;
        buf_.cur_ += 2;
    }

    void write_long(CORBA::Long value)
    {
        addCapacity(4, 4);
        *(CORBA::Long*)(buf_.cur_) = value;
        buf_.cur_ += 4;
    }

    void write_ulong(CORBA::ULong value)
    {
        addCapacity(4, 4);
        *(CORBA::ULong*)(buf_.cur_) = value;
        buf_.cur_ += 4;
    }

    void write_longlong(CORBA::LongLong value)
    {
        addCapacity(8, 8);
        *(CORBA::LongLong*)(buf_.cur_) = value;
        buf_.cur_ += 8;
    }

    void write_ulonglong(CORBA::ULongLong value)
    {
        addCapacity(8, 8);
        *(CORBA::ULongLong*)(buf_.cur_) = value;
        buf_.cur_ += 8;
    }

    void write_float(CORBA::Float value)
    {
        addCapacity(4, 4);
        *(CORBA::Float*)(buf_.cur_) = value;
        buf_.cur_ += 4;
    }

    void write_double(CORBA::Double value)
    {
        addCapacity(8, 8);
        *(CORBA::Double*)(buf_.cur_) = value;
        buf_.cur_ += 8;
    }

    void write_longdouble(CORBA::LongDouble);
    void write_char(CORBA::Char);
    void write_char_array(const CORBA::Char*, CORBA::ULong);
    void write_string(const char*);
    void write_Object(CORBA::Object_ptr);
    void write_TypeCode(CORBA::TypeCode_ptr);
    void write_Principal(CORBA::Principal_ptr);
    void write_any(const CORBA::Any&);

    void write_boolean_array(const CORBA::Boolean* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value != 0, NCANullValue);
            addCapacity(length);
            memcpy(buf_.cur_, value, length);
            buf_.cur_ += length;
        }
    }

    void write_octet_array(const CORBA::Octet* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value != 0, NCANullValue);
            addCapacity(length);
            memcpy(buf_.cur_, value, length);
            buf_.cur_ += length;
        }
    }

    void write_short_array(const CORBA::Short* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value != 0, NCANullValue);
            CORBA::ULong arrsize = length * 2;
            addCapacity(arrsize, 2);
            memcpy(buf_.cur_, value, arrsize);
            buf_.cur_ += arrsize;
        }
    }

    void write_ushort_array(const CORBA::UShort* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value != 0, NCANullValue);
            CORBA::ULong arrsize = length * 2;
            addCapacity(arrsize, 2);
            memcpy(buf_.cur_, value, arrsize);
            buf_.cur_ += arrsize;
        }
    }

    void write_long_array(const CORBA::Long* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value != 0, NCANullValue);
            CORBA::ULong arrsize = length * 4;
            addCapacity(arrsize, 4);
            memcpy(buf_.cur_, value, arrsize);
            buf_.cur_ += arrsize;
        }
    }

    void write_ulong_array(const CORBA::ULong* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value != 0, NCANullValue);
            CORBA::ULong arrsize = length * 4;
            addCapacity(arrsize, 4);
            memcpy(buf_.cur_, value, arrsize);
            buf_.cur_ += arrsize;
        }
    }

    void write_longlong_array(const CORBA::LongLong* value,
			      CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value != 0, NCANullValue);
            CORBA::ULong arrsize = length * 8;
            addCapacity(arrsize, 8);
            memcpy(buf_.cur_, value, arrsize);
            buf_.cur_ += arrsize;
        }
    }

    void write_ulonglong_array(const CORBA::ULongLong* value,
                               CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value != 0, NCANullValue);
            CORBA::ULong arrsize = length * 8;
            addCapacity(arrsize, 8);
            memcpy(buf_.cur_, value, arrsize);
            buf_.cur_ += arrsize;
        }
    }

    void write_float_array(const CORBA::Float* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value != 0, NCANullValue);
            CORBA::ULong arrsize = length * 4;
            addCapacity(arrsize, 4);
            memcpy(buf_.cur_, value, arrsize);
            buf_.cur_ += arrsize;
        }
    }

    void write_double_array(const CORBA::Double* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value != 0, NCANullValue);
            CORBA::ULong arrsize = length * 8;
            addCapacity(arrsize, 8);
            memcpy(buf_.cur_, value, arrsize);
            buf_.cur_ += arrsize;
        }
    }

    void write_longdouble_array(const CORBA::LongDouble*, CORBA::ULong);

    void write_any_NoTypeCode(const CORBA::Any&);
    void write_InputStreamImpl(InputStreamImpl*, CORBA::TypeCode_ptr);
};

} // End of namespace OB

#endif
