// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_TYPE_CODE_H
#define OB_TYPE_CODE_H

#include <OBE/Any.h>
#include <OBE/TypeCodeConst.h>
#include <OBE/TCKind.h>

namespace OB
{

//
// For friend declarations
//
// We don't need to inherit from OB::RefCount since the type code
// reference count is protected by the global garbage collection
// mutex.
//
//
class TypeCodeConst;
class TypeCodeFactory;
class TypeCodeReader;
class TypeCodeWriter;

} // End of namespace OB

namespace CORBA
{

//
// The TypeCode class
//
class TypeCode : public OB::SimpleRefCount
{
    //
    // TypeCode kind
    //
    TCKind kind_;

    //
    // tk_objref, tk_struct, tk_union, tk_enum, tk_alias, tk_native,
    // tk_except
    //
    String_var id_;
    String_var name_;

    //
    // tk_struct, tk_union, tk_enum, tk_except
    //
    OB::StrSeq< int > memberNames_;

    //
    // tk_struct, tk_union, tk_except
    //
    OB::ObjSeq< TypeCode, int > memberTypes_;

    //
    // tk_union
    //
    OB::VarSeq< Any, int > labels_;

    //
    // tk_union
    //
    TypeCode_var discriminatorType_;

    //
    // tk_string, tk_sequence, tk_array
    //
    ULong length_;

    //
    // tk_sequence, tk_array, tk_alias
    //
    TypeCode_var contentType_;

    //
    // If recId_ is set, this is a placeholder recursive TypeCode that
    // was generated with create_recursive_tc(). If the placeholder
    // recursive TypeCode is already embedded, recType_ points to the
    // recursive TypeCode this placeholder delegates to.
    //
    String_var recId_;
    TypeCode_var recType_;

    //
    // Helper functions for equivalent()
    //
    bool equivalentRecHelper(TypeCode_ptr, OB::FixSeq< TypeCode_ptr, int >&,
                             OB::FixSeq< TypeCode_ptr, int >&) const;
    bool equivalentRec(TypeCode_ptr, OB::FixSeq< TypeCode_ptr, int >&,
                       OB::FixSeq< TypeCode_ptr, int >&) const;
    
    //
    // Helper function for get_compact_typecode()
    //
    TypeCode_ptr getCompactTypeCodeRec(OB::FixSeq< TypeCode_ptr, int >&,
				       OB::FixSeq< TypeCode_ptr, int >&) const;

    //
    // Helper functions for releasing recursive TypeCodes
    //
    ULong tryRefCount_;
    bool releaseVisit_;
    bool tryReleaseVisit_;
    bool reachableVisit_;
    void resetTryRelease();
    void tryRelease();
    bool checkTryRelease(TypeCode_ptr);
    bool isReachable(TypeCode_ptr);

    //
    // Hide copy-constructor and assignment operator
    //
    TypeCode(const TypeCode&);
    void operator=(const TypeCode&);

protected:

    //
    // Only these classes may create TypeCodes
    //
    friend class ::OB::TypeCodeConst;
    friend class ::OB::TypeCodeFactory;

    //
    // CDR
    //
    friend class ::OB::InputStreamImpl;
    friend class ::OB::OutputStreamImpl;
    friend class ::OB::TypeCodeReader;
    friend class ::OB::TypeCodeWriter;

    TypeCode();
    virtual ~TypeCode();

public:

    //
    // Standard IDL to C++ Mapping
    //
    static inline TypeCode_ptr _duplicate(TypeCode_ptr p)
    { 
	if(p)
	{
	    //
	    // Must call _OB_duplicateWithGC() due to special cyclic object
	    // dependency handling
	    //
	    p -> _OB_duplicateWithGC();
	}

	return p;
    }

    static inline TypeCode_ptr _nil()
    { return 0; }

    struct BadKind : public UserException
    {
	//
	// Standard IDL to C++ Mapping
	//
	BadKind() { }
	BadKind(const BadKind& ex) : UserException(ex) { }
	BadKind& operator=(const BadKind&) { return *this; }
	
	static BadKind* _downcast(Exception* p)
	{ return dynamic_cast<BadKind*>(p); }
	static const BadKind* _downcast(const Exception* p)
	{ return dynamic_cast<const BadKind*>(p); }
	virtual void _raise() const { throw *this; }
        virtual const char* _name() const { return "BadKind"; }
        virtual const char* _rep_id() const
        { return "IDL:omg.org/CORBA/TypeCode/BadKind:1.0"; }
	
	//
	// Additional ORBacus specific functions
	//
	virtual char* _to_string() const
        { return _OB_defaultToString("CORBA::TypeCode::BadKind"); }

	//
	// ORBacus internal functions
	// Application programs must not use these functions directly
	//
	virtual Exception* _OB_clone() const { return new BadKind(*this); }
        virtual void _OB_insertInAny(Any&);
	static void _OB_unmarshal(BadKind&, OB::InputStreamImpl*);
    };

    class TypeCodeConstForBadKind : public JTCMutex
    {
	TypeCodeConstForBadKind(const TypeCodeConstForBadKind&);
	void operator=(const TypeCodeConstForBadKind&);
	
	CORBA::TypeCode_var typeCode_;
	
    public:
	
	TypeCodeConstForBadKind() { }

	operator CORBA::TypeCode_ptr();
	CORBA::TypeCode_ptr operator->()
	{ return operator CORBA::TypeCode_ptr(); }
    };
    static OBE_IMPORT TypeCodeConstForBadKind _tc_BadKind;

    Boolean equal(TypeCode_ptr) const;
    Boolean equivalent(TypeCode_ptr) const;
    TypeCode_ptr get_compact_typecode() const;
    TCKind kind() const;

    const char* id() const;
    const char* name() const;
    
    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    bool _OB_isSystemException() const;
    bool _OB_isRecursive() const { return recId_.in() ? true : false; }
    void _OB_duplicateWithGC();
    void _OB_releaseWithGC();

    //
    // Embed recursive placeholder TypeCodes
    //
    static void _OB_embedRecTC(TypeCode_ptr);
    static void _OB_embedRecTC(TypeCode_ptr, TypeCode_ptr);

    //
    // Accessor methods
    //
    TypeCode_ptr _OB_contentType_nodup() const;
    ULong _OB_length() const;
    Long _OB_defaultIndex() const;

#ifndef NDEBUG
    ULong _OB_memberCount() const;
    const char* _OB_memberName(ULong index) const;
    TypeCode_ptr _OB_memberType_nodup(ULong index) const;
    Any* _OB_memberLabel(ULong index) const;
    TypeCode_ptr _OB_discriminatorType_nodup() const;
#endif
};

inline Boolean
is_nil(TypeCode_ptr p)
{
    return p == 0;
}

inline void
release(TypeCode_ptr p)
{
    if(!is_nil(p))
    {
        //
	// Must call _OB_releaseWithGC() due to special cyclic object
	// dependency handling
	//
	p -> _OB_releaseWithGC();
    }
}

} // End of namespace CORBA

//
// Insertion and extraction for TypeCode::BadKind
//
void operator<<=(CORBA::Any&, CORBA::TypeCode::BadKind*);
void operator<<=(CORBA::Any&, const CORBA::TypeCode::BadKind&);
CORBA::Boolean operator>>=(const CORBA::Any&, CORBA::TypeCode::BadKind*&);

inline void
operator<<=(CORBA::Any_var& any, CORBA::TypeCode::BadKind* val)
{
    any.inout() <<= val;
}

inline void
operator<<=(CORBA::Any_var& any, const CORBA::TypeCode::BadKind& val)
{
    any.inout() <<= val;
}

inline CORBA::Boolean
operator>>=(const CORBA::Any_var& any, CORBA::TypeCode::BadKind*& val)
{
    return any.in() >>= val;
}

namespace OB
{

//
// Classes for _tc_<type> constants
//
class TypeCodePrimitiveConst : public JTCMutex
{
    TypeCodePrimitiveConst(const TypeCodePrimitiveConst&);
    void operator=(const TypeCodePrimitiveConst&);

    CORBA::TypeCode_var typeCode_;
    CORBA::TCKind kind_;

public:
    
    TypeCodePrimitiveConst(CORBA::TCKind kind) : kind_(kind) { }

    operator CORBA::TypeCode_ptr();
    CORBA::TypeCode_ptr operator->()
    { return operator CORBA::TypeCode_ptr(); }
};

class TypeCodeStringConst : public JTCMutex
{
    TypeCodeStringConst(const TypeCodeStringConst&);
    void operator=(const TypeCodeStringConst&);

    CORBA::TypeCode_var typeCode_;
    bool wide_;

public:
    
    TypeCodeStringConst(bool wide) : wide_(wide) { }

    operator CORBA::TypeCode_ptr();
    CORBA::TypeCode_ptr operator->()
    { return operator CORBA::TypeCode_ptr(); }

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    CORBA::TypeCode_ptr _OB_bounded(CORBA::ULong);
};

class TypeCodeObjectConst : public JTCMutex
{
    TypeCodeObjectConst(const TypeCodeObjectConst&);
    void operator=(const TypeCodeObjectConst&);

    CORBA::TypeCode_var typeCode_;

public:
    
    TypeCodeObjectConst() { }

    operator CORBA::TypeCode_ptr();
    CORBA::TypeCode_ptr operator->()
    { return operator CORBA::TypeCode_ptr(); }
};

class TypeCodeSysExConst : public JTCMutex
{
    TypeCodeSysExConst(const TypeCodeSysExConst&);
    void operator=(const TypeCodeSysExConst&);

    CORBA::TypeCode_var typeCode_;
    const char* name_;

public:
    
    TypeCodeSysExConst(const char* name) : name_(name) { }

    operator CORBA::TypeCode_ptr();
    CORBA::TypeCode_ptr operator->()
    { return operator CORBA::TypeCode_ptr(); }
};

//
// Get original type, i.e., remove aliases
//
CORBA::TypeCode_ptr GetOrigType_nodup(CORBA::TypeCode_ptr);

} // End of namespace OB

namespace CORBA
{

//
// TypeCode constants
//
OBE_IMPORT extern OB::TypeCodePrimitiveConst _tc_null;
OBE_IMPORT extern OB::TypeCodePrimitiveConst _tc_void;
OBE_IMPORT extern OB::TypeCodePrimitiveConst _tc_short;
OBE_IMPORT extern OB::TypeCodePrimitiveConst _tc_long;
OBE_IMPORT extern OB::TypeCodePrimitiveConst _tc_longlong;
OBE_IMPORT extern OB::TypeCodePrimitiveConst _tc_ushort;
OBE_IMPORT extern OB::TypeCodePrimitiveConst _tc_ulong;
OBE_IMPORT extern OB::TypeCodePrimitiveConst _tc_ulonglong;
OBE_IMPORT extern OB::TypeCodePrimitiveConst _tc_float;
OBE_IMPORT extern OB::TypeCodePrimitiveConst _tc_double;
OBE_IMPORT extern OB::TypeCodePrimitiveConst _tc_longdouble;
OBE_IMPORT extern OB::TypeCodePrimitiveConst _tc_boolean;
OBE_IMPORT extern OB::TypeCodePrimitiveConst _tc_char;
OBE_IMPORT extern OB::TypeCodePrimitiveConst _tc_octet;
OBE_IMPORT extern OB::TypeCodePrimitiveConst _tc_any;
OBE_IMPORT extern OB::TypeCodePrimitiveConst _tc_TypeCode;
OBE_IMPORT extern OB::TypeCodePrimitiveConst _tc_Principal;

OBE_IMPORT extern OB::TypeCodeStringConst _tc_string;

OBE_IMPORT extern OB::TypeCodeObjectConst _tc_Object;

OBE_IMPORT extern OB::TypeCodeSysExConst _tc_UNKNOWN;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_BAD_PARAM;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_NO_MEMORY;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_IMP_LIMIT;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_COMM_FAILURE;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_INV_OBJREF;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_NO_PERMISSION;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_INTERNAL;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_MARSHAL;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_INITIALIZE;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_NO_IMPLEMENT;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_BAD_TYPECODE;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_BAD_OPERATION;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_NO_RESOURCES;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_NO_RESPONSE;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_BAD_INV_ORDER;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_TRANSIENT;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_OBJ_ADAPTER;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_OBJECT_NOT_EXIST;
OBE_IMPORT extern OB::TypeCodeSysExConst _tc_INV_POLICY;

} // End of namespace CORBA

#endif
