// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_UPCALL_H
#define OB_UPCALL_H

#include <OBE/ORBInstance_fwd.h>
#include <OBE/RefCounted_fwd.h>
#include <OBE/POA_fwd.h>

#include <OBE/Except.h>
#include <OBE/OCITypes.h>
#include <OBE/ObjectKey.h>
#include <OBE/Stream.h>

namespace OB
{

//
// The UpcallImpl class (not reference counted)
//
class UpcallReturn;
class UpcallImpl
{
    //
    // Hide copy-constructor and assignment operator
    //
    UpcallImpl(const UpcallImpl&);
    void operator=(const UpcallImpl&);

protected:

    //
    // The ORBInstance object
    //
    ORBInstance* orbInstance_;

    //
    // Upcall delegates to UpcallReturn upon return from the
    // upcall. If this is nil, then no response is expected (i.e.,
    // this is a oneway call).
    //
    UpcallReturn* upcallReturn_;

    //
    // Information about the IOR profile
    //
    const OCI::ObjectKey* ocikey_;

    //
    // The unique request ID
    //
    CORBA::ULong reqId_;

    //
    // The name of the operation
    //
    const char* op_;

    //
    // The marshalled headers and parameters
    //
    OutputStreamImpl out_;

    //
    // Holds the results of the operation
    //
    InputStreamImpl in_;

    //
    // The POA and the object key
    //
    PortableServer::POA* poa_;
    ObjectKeyData key_;

public:

    UpcallImpl(ORBInstance_ptr);
    ~UpcallImpl();

    //
    // Setup the Upcall
    //
    void setup(UpcallReturn*, const OCI::ObjectKey*, CORBA::ULong,
	       const char*);

    //
    // Accessors
    //
    ORBInstance_ptr orbInstance_nodup() const
    { return orbInstance_; }
    UpcallReturn* upcallReturn_nodup() const
    { return upcallReturn_; }
    const OCI::ObjectKey& key() const
    { return *ocikey_; }
    CORBA::ULong requestId() const { return reqId_; }
    const char* operation() const { return op_; }
    CORBA::Boolean responseExpected() const;

    //
    // Obtain the input and output stream
    //
    OutputStreamImpl* output() { return &out_; }
    InputStreamImpl* input() { return &in_; }

    //
    // Unmarshalling interception points
    //
    InputStreamImpl* preUnmarshal()
	throw(CORBA::SystemException, OB::LocationForward);
    void unmarshalEx(const CORBA::SystemException&)
	throw(CORBA::SystemException, OB::LocationForward);
    void postUnmarshal()
	throw(CORBA::SystemException, OB::LocationForward);

    //
    // Marshalling interception points
    //
    OutputStreamImpl* preMarshal()
	throw(CORBA::SystemException, OB::LocationForward);
    void marshalEx(const CORBA::SystemException&)
	throw(CORBA::SystemException, OB::LocationForward);
    void postMarshal()
	throw(CORBA::SystemException, OB::LocationForward);

    //
    // Set the state of the Upcall
    //
    void setUserException(const CORBA::UserException&)
	throw();
    void setSystemException(const CORBA::SystemException&)
	throw();
    void setLocationForward(RefCountIOR_ptr)
	throw();

    //
    // Set the dispatch info
    //
    void setDispatchInfo(PortableServer::POA_ptr, const ObjectKeyData& key)
	throw();

    //
    // Do the invocation
    //
    void invoke()
	throw();
};

} // End of namespace OB

namespace OB
{

//
// The UpcallReturn class
//
class UpcallReturn
{
public:

    virtual ~UpcallReturn() { };

    //
    // Called upon return from an upcall
    //
    virtual void upcallBeginReply(UpcallImpl*) = 0;
    virtual void upcallEndReply(UpcallImpl*) = 0;
    virtual void upcallUserException(UpcallImpl*,
				     const CORBA::UserException&) = 0;
    virtual void upcallSystemException(UpcallImpl*,
				       const CORBA::SystemException&) = 0;
    virtual void upcallForward(UpcallImpl*, RefCountIOR_ptr) = 0;
};

} // End of namespace OB

#endif
