// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/TypeCode.h>
#include <OBE/Principal.h>
#include <OBE/Object.h>
#include <OBE/Stream.h>
#include <OBE/ORBInstance.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Constructor and destructor
// ----------------------------------------------------------------------

Any::Any()
    : type_(TypeCode::_duplicate(_tc_null)), value_(0), rel_(false), info_(0)
{
}

Any::Any(const Any& r)
    : type_(TypeCode::_duplicate(_tc_null)), value_(0),
      rel_(false), info_(0)
{
    *this = r;
}

Any::Any(TypeCode_ptr tc, void* val, Boolean rel)
    : type_(TypeCode::_duplicate(tc)), value_(val),
      rel_(rel), info_(0)
{
}

Any::~Any()
{
    replace(TypeCode::_nil(), 0, false);
}

// ----------------------------------------------------------------------
// Assignment
// ----------------------------------------------------------------------

Any&
Any::operator=(const Any& any)
{
    if(this != &any)
    {
        orbInstance_ = ORBInstance::_duplicate(any.orbInstance_);

	if(any.value_)
	{
	    void* val = 0;
	    
	    TypeCode_ptr p = GetOrigType_nodup(any.type_);
	    
	    switch(p -> kind())
	    {
	    case tk_null:
	    case tk_void:
		break;
		
	    case tk_short:
		val = new Short(*(const Short*)any.value_);
		break;
		
	    case tk_long:
		val = new Long(*(const Long*)any.value_);
		break;
		
	    case tk_longlong:
		val = new LongLong(*(const LongLong*)any.value_);
		break;
		
	    case tk_ushort:
		val = new UShort(*(const UShort*)any.value_);
		break;
		
	    case tk_ulong:
	    case tk_enum:
		val = new ULong(*(const ULong*)any.value_);
		break;
		
	    case tk_ulonglong:
		val = new ULongLong(*(const ULongLong*)any.value_);
		break;
		
	    case tk_float:
		val = new Float(*(const Float*)any.value_);
		break;
		
	    case tk_double:
		val = new Double(*(const Double*)any.value_);
		break;
		
	    case tk_longdouble:
#if SIZEOF_DOUBLE != 0
		val = new LongDouble(*(const LongDouble*)any.value_);
#endif
		break;
		
	    case tk_boolean:
		val = new Boolean(*(const Boolean*)any.value_);
		break;
		
	    case tk_char:
		val = new Char(*(const Char*)any.value_);
		break;
		
	    case tk_octet:
		val = new Octet(*(const Octet*)any.value_);
		break;
		
	    case tk_any:
		val = new Any(*(const Any*)any.value_);
		break;
		
	    case tk_TypeCode:
		val = TypeCode::_duplicate((TypeCode_ptr)any.value_);
		break;
		
	    case tk_Principal:
		val = Principal::_duplicate((Principal_ptr)any.value_);
		break;
		
	    case tk_objref:
		val = Object::_duplicate((Object_ptr)any.value_);
		break;
		
	    case tk_struct:
	    case tk_sequence:
	    case tk_union:
	    case tk_array:
	    case tk_except:
		
		if(p -> _OB_isSystemException())
		{
                    SystemException* sysEx = (SystemException*)any.value_;
		    val = dynamic_cast<SystemException*>(sysEx -> _OB_clone());
		    assert(val);
		}
		else if(any.info_)
		{
		    val = any.info_ -> dup(any.value_);
		}
		else
		{
                    InputStreamImpl* in = (InputStreamImpl*)any.value_;
                    val = in -> _OB_clone();
		}
		break;
		
	    case tk_string:
		val = string_dup((const char*)any.value_);
		break;
		
	    case tk_native:
                if(!strcmp(p -> id(), "IDL:omg.org/CORBA/AbstractBase:1.0"))
		    //
		    // Abstract not supported in OBE
		    //
		    assert(false);
                else
                    assert(false);
		break;

	    case tk_alias:
		assert(false);
		break;

	    //
	    // These are not supported in OBE
	    //
	    case tk_fixed:
	    case tk_wchar:
	    case tk_wstring:
	    case tk_value:
	    case tk_value_box:
	    case tk_abstract_interface:
		assert(false);
		break;
	    }
	
	    replace(any.type_, val, true, any.info_);
	}
	else
	    replace(any.type_, 0);
    }
    
    return *this;
}

// ----------------------------------------------------------------------
// Insertion into Any
// ----------------------------------------------------------------------

void
Any::operator<<=(Short v)
{
    replace(_tc_short, new Short(v), true);
}

void
Any::operator<<=(UShort v)
{
    replace(_tc_ushort, new UShort(v), true);
}

void
Any::operator<<=(Long v)
{
    replace(_tc_long, new Long(v), true);
}

void
Any::operator<<=(ULong v)
{
    replace(_tc_ulong, new ULong(v), true);
}

void
Any::operator<<=(LongLong v)
{
    replace(_tc_longlong, new LongLong(v), true);
}

void
Any::operator<<=(ULongLong v)
{
    replace(_tc_ulonglong, new ULongLong(v), true);
}

void
Any::operator<<=(Float v)
{
    replace(_tc_float, new Float(v), true);
}

void
Any::operator<<=(Double v)
{
    replace(_tc_double, new Double(v), true);
}

void
Any::operator<<=(LongDouble v)
{
#if SIZEOF_LONG_DOUBLE >= 12
    replace(_tc_longdouble, new LongDouble(v), true);
#endif
}

void
Any::operator<<=(const char* v)
{
    replace(_tc_string, string_dup(v), true);
}

void
Any::operator<<=(const String_var& v)
{
    *this <<= v.in();
}

void
Any::operator<<=(const StrForSeq& v)
{
    *this <<= v.in();
}

void
Any::operator<<=(const Any& v)
{
    *this <<= new Any(v);
}

void
Any::operator<<=(Any* v)
{
    replace(_tc_any, v, true);
}

void
Any::operator<<=(TypeCode_ptr v)
{
    TypeCode_ptr val = TypeCode::_duplicate(v);
    *this <<= &val;
}

void
Any::operator<<=(TypeCode_ptr* v)
{
    replace(_tc_TypeCode, *v, true);
}

void
Any::operator<<=(Principal_ptr v)
{
    Principal_ptr val = Principal::_duplicate(v);
    *this <<= &val;
}

void
Any::operator<<=(Principal_ptr* v)
{
    replace(_tc_Principal, *v, true);
}

void
Any::operator<<=(Object_ptr v)
{
    Object_ptr val = Object::_duplicate(v);
    *this <<= &val;
}

void
Any::operator<<=(Object_ptr* v)
{
    replace(_tc_Object, *v, true);
}

//
// The IDL type 'boolean' is mapped to bool, so a separate insertion
// operator is possible
//
void
Any::operator<<=(Boolean v)
{
    replace(_tc_boolean, new Boolean(v), true);
}

void
Any::operator<<=(from_boolean v)
{
    replace(_tc_boolean, new Boolean(v.val), true);
}

void
Any::operator<<=(from_char v)
{
    replace(_tc_char, new Char(v.val), true);
}

void
Any::operator<<=(from_octet v)
{
    replace(_tc_octet, new Octet(v.val), true);
}

void
Any::operator<<=(from_string v)
{
    TypeCode_var tc = _tc_string._OB_bounded(v.bound);

    if(v.nocopy)
	replace(tc, v.val, true);
    else
	replace(tc, string_dup(v.val), true);
}

// ----------------------------------------------------------------------
// Extraction from Any
// ----------------------------------------------------------------------

Boolean
Any::operator>>=(Short& v) const
{
    if(check_type(_tc_short))
    {
	v = *(Short*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(UShort& v) const
{
    if(check_type(_tc_ushort))
    {
	v = *(UShort*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(Long& v) const
{
    if(check_type(_tc_long))
    {
	v = *(Long*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(ULong& v) const
{
    if(check_type(_tc_ulong))
    {
	v = *(ULong*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(LongLong& v) const
{
    if(check_type(_tc_longlong))
    {
	v = *(LongLong*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(ULongLong& v) const
{
    if(check_type(_tc_ulonglong))
    {
	v = *(ULongLong*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(Float& v) const
{
    if(check_type(_tc_float))
    {
	v = *(Float*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(Double& v) const
{
    if(check_type(_tc_double))
    {
	v = *(Double*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(LongDouble& v) const
{
#if SIZEOF_LONG_DOUBLE >= 12
    if(check_type(_tc_longdouble))
    {
	v = *(LongDouble*)value_;
	return true;
    }
    else
	return false;
#else
    return false;
#endif
}

Boolean
Any::operator>>=(const char*& v) const
{
    if(check_type(_tc_string))
    {
	v = (const char*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(const Any*& v) const
{
    if(check_type(_tc_any))
    {
	v = (const Any*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(const SystemException*& v) const
{
    TypeCode_ptr origType = GetOrigType_nodup(type_);
    if(origType -> _OB_isSystemException())
    {
	v = (SystemException*)value_;
        return true;
    }
    else
        return false;
}

Boolean 
Any::operator>>=(TypeCode_ptr& v) const
{
    if(check_type(_tc_TypeCode))
    {
	v = (TypeCode_ptr)value_;
	return true;
    }
    else
	return false;
}

Boolean 
Any::operator>>=(Principal_ptr& v) const
{
    if(check_type(_tc_Principal))
    {
	v = (Principal_ptr)value_;
	return true;
    }
    else
	return false;
}

Boolean 
Any::operator>>=(Object_ptr& v) const
{
    if(check_type(_tc_Object))
    {
	v = (Object_ptr)value_;
	return true;
    }
    else
	return false;
}

//
// The IDL type 'boolean' is mapped to bool, so a separate extraction
// operator is possible
//
Boolean
Any::operator>>=(Boolean& v) const
{
    if(check_type(_tc_boolean))
    {
	v = *(Boolean*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(to_boolean v) const
{
    if(check_type(_tc_boolean))
    {
	v.val = *(Boolean*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(to_char v) const
{
    if(check_type(_tc_char))
    {
	v.val = *(Char*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(to_octet v) const
{
    if(check_type(_tc_octet))
    {
	v.val = *(Octet*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(to_string v) const
{
    TypeCode_ptr origType = GetOrigType_nodup(type_);
    if(origType -> kind() == tk_string && origType -> _OB_length() == v.bound)
    {
	v.val = (const char*)value_;
	return true;
    }
    else
	return false;
}

Boolean
Any::operator>>=(to_object v) const
{
    TypeCode_ptr origType = GetOrigType_nodup(type_);

    if(origType -> kind() == tk_objref)
    {
	v.val = Object::_duplicate((Object_ptr)value_);
	return true;
    }
    else if(origType -> kind() == tk_abstract_interface)
    {
	//
	// Not allowed in OBE
	//
	assert(false);
	return false;
    }

    return false;
}

// ----------------------------------------------------------------------
// Replace type
// ----------------------------------------------------------------------

void
Any::replace(TypeCode_ptr tc, void* val, Boolean rel)
{
    replace(tc, val, rel, 0);
}

// ----------------------------------------------------------------------
// TypeCode accessor and modifier
// ----------------------------------------------------------------------

TypeCode_ptr
Any::type() const
{
    return TypeCode::_duplicate(type_);
}

void
Any::type(TypeCode_ptr tc)
{
    if(!type_ -> equivalent(tc))
        throw BAD_TYPECODE();
    else
        type_ = TypeCode::_duplicate(tc);
}

// ----------------------------------------------------------------------
// Return type and value
// ----------------------------------------------------------------------

const void*
Any::value() const
{
    return value_;
}

// ----------------------------------------------------------------------
// Additional ORBacus functions
// ----------------------------------------------------------------------

void
Any::replace(TypeCode_ptr tc, void* val, Boolean rel,
             const ConstructedInfo* info)
{
    if(rel_ && value_)
    {
	TypeCode_ptr p = GetOrigType_nodup(type_);
	
	switch(p -> kind())
	{
	case tk_null:
	case tk_void:
	    break;
	    
	case tk_short:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (Short*)value_;
	    break;
	    
	case tk_long:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (Long*)value_;
	    break;

	case tk_longlong:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (LongLong*)value_;
	    break;

	case tk_ushort:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (UShort*)value_;
	    break;
	    
	case tk_ulong:
	case tk_enum:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (ULong*)value_;
	    break;

	case tk_ulonglong:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (ULongLong*)value_;
	    break;

	case tk_float:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (Float*)value_;
	    break;
	    
	case tk_double:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (Double*)value_;
	    break;
	    
	case tk_longdouble:
#if SIZEOF_DOUBLE != 0
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (LongDouble*)value_;
#endif
	    break;

	case tk_boolean:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (Boolean*)value_;
	    break;
	    
	case tk_char:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (Char*)value_;
	    break;
	    
	case tk_octet:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (Octet*)value_;
	    break;
	    
	case tk_any:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    delete (Any*)value_;
	    break;
	    
	case tk_TypeCode:
	    release((TypeCode_ptr)value_);
	    break;
	    
	case tk_Principal:
	    release((Principal_ptr)value_);
	    break;
	    
	case tk_objref:
	    release((Object_ptr)value_);
	    break;
	    
	case tk_struct:
	case tk_sequence:
	case tk_union:
	case tk_array:
	case tk_except:

	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);

	    if(p -> _OB_isSystemException())
	    {
		delete (SystemException*)value_;
	    }
	    else if(info_)
	    {
		info_ -> free(value_);
	    }
	    else
	    {
                delete (InputStreamImpl*)value_;
	    }
	    break;
	    
	case tk_string:
	    assert_nca(value_ == 0 || value_ != val, NCAAnySelfReplace);
	    string_free((char*)value_);
	    break;
	    
	case tk_native:
            if(!strcmp(p -> id(), "IDL:omg.org/CORBA/AbstractBase:1.0"))
		//
		// Abstract not supported in OBE
		//
		assert(false);
            break;

	case tk_alias: // No alias possible here,
	    break;     // this is to keep the compiler happy

	//
	// These are not supported in OBE
	//
	case tk_fixed:
	case tk_wchar:
	case tk_wstring:
	case tk_value:
	case tk_value_box:
	case tk_abstract_interface:
	    assert(false);
	    break;
	}
    }

    type_ = TypeCode::_duplicate(tc);
    value_ = val;
    rel_ = rel;
    info_ = info;

    //
    // If we don't have an ORB instance and this Any contains an object
    // reference, then get the ORB instance from the object reference
    //
    if(is_nil(orbInstance_) && value_)
    {
        TypeCode_ptr origType = GetOrigType_nodup(type_);
        if(origType -> kind() == tk_objref)
            orbInstance_ = ((Object_ptr)value_) -> _OB_ORBInstance();
    }
}

Boolean
Any::check_type(TypeCode_ptr t) const
{
    return type_ -> equivalent(t);
}

InputStreamImpl*
Any::get_input_stream() const
{
    TypeCode_ptr origType = GetOrigType_nodup(type_);

    InputStreamImpl* in = 0;

    switch(origType -> kind())
    {
    case tk_struct:
    case tk_sequence:
    case tk_union:
    case tk_array:
    case tk_except:

        if(!info_ && !origType -> _OB_isSystemException())
        {
            in = (InputStreamImpl*)value_;
            in -> _OB_reset();
        }
        break;

    default:
	break;
    }

    assert(in);
    return in;
}

void
Any::_OB_ORBInstance(ORBInstance_ptr orbInstance)
{
    orbInstance_ = ORBInstance::_duplicate(orbInstance);
}

ORBInstance_ptr
Any::_OB_ORBInstance()
{
    return ORBInstance::_duplicate(orbInstance_);
}

TypeCode_ptr
Any::_OB_type_nodup() const
{
    return type_;
}

// ----------------------------------------------------------------------
// Insertion and extraction on Any_var
// ----------------------------------------------------------------------

void
operator<<=(Any_var& any, Short val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, UShort val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Long val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, ULong val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, LongLong val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, ULongLong val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Float val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Double val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, LongDouble val)
{
#if SIZEOF_DOUBLE != 0
    any.inout() <<= val;
#endif
}

void
operator<<=(Any_var& any, const char* val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, char* val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, const String_var& val)
{
    any.inout() <<= val.in();
}

void
operator<<=(Any_var& any, const StrForSeq& val)
{
    any.inout() <<= val.in();
}

void
operator<<=(Any_var& any, const Any& val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Any* val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, TypeCode_ptr val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, TypeCode_ptr* val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Principal_ptr val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Principal_ptr* val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Object_ptr val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Object_ptr* val)
{
    any.inout() <<= val;
}

Boolean
operator>>=(const Any_var& any, Short& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, UShort& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Long& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, ULong& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, LongLong& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, ULongLong& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Float& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Double& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, LongDouble& val)
{
#if SIZEOF_DOUBLE != 0
    return any.in() >>= val;
#endif
}

Boolean
operator>>=(const Any_var& any, const char*& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, const Any*& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, const SystemException*& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, TypeCode_ptr& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Principal_ptr& val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Object_ptr& val)
{
    return any.in() >>= val;
}

void
operator<<=(Any_var& any, Any::from_boolean val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Any::from_octet val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Any::from_char val)
{
    any.inout() <<= val;
}

void
operator<<=(Any_var& any, Any::from_string val)
{
    any.inout() <<= val;
}

Boolean
operator>>=(const Any_var& any, Any::to_boolean val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Any::to_octet val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Any::to_char val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Any::to_string val)
{
    return any.in() >>= val;
}

Boolean
operator>>=(const Any_var& any, Any::to_object val)
{
    return any.in() >>= val;
}

// ----------------------------------------------------------------------
// Insertion and extraction for system exceptions
// ----------------------------------------------------------------------

#define OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(T) \
\
void \
operator<<=(Any& any, T* v) \
{ \
    SystemException* ex = v; \
    any.replace(_tc_##T, ex, true); \
} \
\
void \
operator<<=(Any& any, const T& v) \
{ \
    SystemException* ex = new T(v); \
    any.replace(_tc_##T, ex, true); \
} \
\
Boolean \
operator>>=(const Any& any, const T*& v) \
{ \
    if(any.check_type(_tc_##T)) \
    { \
	SystemException* ex = (SystemException*)any.value(); \
        v = dynamic_cast<T*>(ex); \
        assert(v); \
        return true; \
    } \
    else \
        return false; \
}

OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(UNKNOWN)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(BAD_PARAM)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(NO_MEMORY)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(IMP_LIMIT)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(COMM_FAILURE)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(INV_OBJREF)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(NO_PERMISSION)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(INTERNAL)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(MARSHAL)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(INITIALIZE)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(NO_IMPLEMENT)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(BAD_TYPECODE)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(BAD_OPERATION)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(NO_RESOURCES)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(NO_RESPONSE)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(BAD_INV_ORDER)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(TRANSIENT)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(OBJ_ADAPTER)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(OBJECT_NOT_EXIST)
OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I(INV_POLICY)
