// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/Basic.h>

#include <ctype.h>

//
// The ORBacus version, as string and as integer
//
namespace OB
{

const char* Version = OBE_VERSION;
unsigned long IntegerVersion = OBE_INTEGER_VERSION;

} // End of namespace OB

#ifndef NDEBUG

// ----------------------------------------------------------------------
// Helper function for assert_nca()
// ----------------------------------------------------------------------

void
OB::NCA(OB::NCAReason reason, const char* ex,
	const char* file, unsigned int line)
{
    //
    // Update the documentation if error messages are changed here !!!!
    //

    CORBA::String_var err =
	CORBA::string_dup("non-compliant application detected:\n");

    switch(reason)
    {
    case NCANotImpl:
	err +=
"application requested a feature that is not yet implemented";
	break;

    case NCADeprecated:
	err +=
"application used a deprecated feature that is not implemented anymore";
	break;

    case NCAWrongAlloc:
	err +=
"application used wrong memory allocation function";
	break;

    case NCADoubleFree:
	err +=
"memory that was already deallocated was deallocated again";
	break;

    case NCARefCountNotZero:
	err +=
"object was deleted without an object reference count of zero";
	break;

    case NCARefCountZero:
	err +=
"object was already deleted (object reference count was already zero)";
	break;

    case NCASeqLenGrMax:
	err +=
"sequence length was greater than maximum sequence length";
	break;

    case NCASeqRange:
	err +=
"index for sequence operator[] was out of range";
	break;

    case NCASeqBound:
	err +=
"buffer size not equal to sequence bound";
	break;

    case NCANullInit:
	err +=
"null pointer was used to initialize T_var type";
	break;

    case NCANullPtr:
	err +=
"operator->() was used on null pointer or nil object reference";
	break;

    case NCANullDeref:
	err +=
"application tried to dereference a null pointer";
	break;

    case NCANullString:
	err +=
"null pointer was passed as string parameter or return value";
	break;

    case NCANullValue:
        err +=
"null value passed as parameter.";
        break;

    case NCASelfAssign:
	err +=
"self assignment caused a dangling pointer";
	break;

    case NCAAnySelfReplace:
	err +=
"replacement of Any content by its own value caused a dangling pointer";
	break;

    case NCADiscType:
	err +=
"invalid union discriminator type used";
	break;

    case NCADiscMismatch:
	err +=
"union discriminator mismatch";
	break;

    case NCAUnionInit:
	err +=
"uninitialized union used";
	break;

    case NCAUnembeddedRecTC:
	err +=
"an operation on an unembedded recursive TypeCode was invoked";
	break;

    case NCAEmbeddedRecTC:
	err +=
"an already embedded TypeCode was reused";
	break;

    case NCANoLongDouble:
	err +=
"LongDouble type is not supported on this platform";
	break;
    }
 
    err += "\nassertion failed: ";
    err += ex;
    err += "\nfile ";
    err += file;
    err += ", line ";
    err += line;

    fputs(err, stderr);
}

#endif

// ----------------------------------------------------------------------
// Special ORBacus memory allocation that keeps track of the number of
// elements in a buffer
// ----------------------------------------------------------------------

static const size_t more = sizeof(CORBA::ULong) * 2;
static const CORBA::ULong magic     = 0x02468ACE;
static const CORBA::ULong magicFree = 0x0E2C4A68;

void*
OB::Alloc(size_t sz, CORBA::ULong n)
{
    char* p = (char*)malloc(more + n * sz);
    if(p)
    {
        CORBA::ULong* np = (CORBA::ULong*)p;
        *np++ = magic;
        *np = n;
        return p + more;
    }

    return 0;
}

void
OB::Free(void* vp)
{
    char* p = (char*)vp;

#ifndef NDEBUG

    CORBA::ULong* magicp = (CORBA::ULong*)(p - more);
    assert_nca(*magicp != magicFree, NCADoubleFree);
    assert_nca(*magicp == magic, NCAWrongAlloc);
    *magicp = magicFree;

#endif

    free(p - more);
}

CORBA::ULong
OB::Allocated(const void* vp)
{
    const CORBA::ULong* magicp = (const CORBA::ULong*)((const char*)vp - more);
    assert_nca(*magicp == magic, NCAWrongAlloc);
    const CORBA::ULong* np = magicp + 1;
    return *np;
}

#ifndef NDEBUG

bool
OB::AllocCheck(const void* vp)
{
    const CORBA::ULong* magicp = (const CORBA::ULong*)((const char*)vp - more);
    return *magicp == magic;
}

#endif

// ----------------------------------------------------------------------
// Memory allocation functions used by string sequences
// ----------------------------------------------------------------------

char**
OB::StrSeqAllocbuf(CORBA::ULong n)
{
    char** p = (char**)OB::Alloc(sizeof(char*), n);
    
    for(CORBA::ULong i = 0 ; i < n ; i++)
	p[i] = OB::EmptyString;

    return p;
}

void
OB::StrSeqFreebuf(char** p)
{
    if(p)
    {
	CORBA::ULong n = OB::Allocated(p);
	
	for(CORBA::ULong i = 0 ; i < n ; i++)
	    CORBA::string_free(p[i]);

	OB::Free(p);
    }
}

// ----------------------------------------------------------------------
// Convert between octet sequence and ascii string
// ----------------------------------------------------------------------

char*
OB::OctetsToAscii(const CORBA::Octet* oct, CORBA::ULong len)
{
    char* str = CORBA::string_alloc(len * 2);
    char s[3];
    
    for(CORBA::ULong i = 0 ; i < len ; i++, oct++)
    {
	sprintf(s, "%02x", *oct);
	strcpy(&str[i * 2], s);
    }

    return str;
}

CORBA::Octet*
OB::AsciiToOctets(const char* str)
{
    CORBA::ULong len = strlen(str) / 2;
    CORBA::Octet* oct = new CORBA::Octet[len];

    int val;
    char hex[3];
    hex[2] = '\0';

    const char* s = str;
    for(CORBA::ULong i = 0 ; i < len ; i++)
    {
	hex[0] = *s++;
	hex[1] = *s++;
	sscanf(hex, "%x", &val);
	oct[i] = val;
    }

    return oct;
}

// ----------------------------------------------------------------------
// Convert between LongLong/ULongLong and string
// ----------------------------------------------------------------------

bool
OB::StringToULongLong(const char* str, const char** end, CORBA::UShort base,
                      CORBA::ULongLong& result)
{
    const char *cp = str;
    const CORBA::ULongLong MAX = (~CORBA::ULongLong(0));

    //
    // Skip leading whitespace
    //
    while(*cp && isspace(*cp))
        cp++;

    if(*cp == '\0')
        return false;

    //
    // Check for sign
    //
    if(*cp == '-')
        return false;
    else if(*cp == '+')
        cp++;

    if(*cp == '\0')
        return false;

    //
    // Check for leading zero
    //
    if(*cp == '0' && *(cp + 1) != '\0')
    {
        cp++;
        if(*cp == 'x' || *cp == 'X') // hexadecimal
        {
            if(base != 16 && base != 0)
                return false;
            base = 16;
            cp++;
        }
        else if(isdigit(*cp)) // octal
        {
            if(base != 8 && base != 0)
                return false;
            base = 8;
        }
    }
    else if(base == 0)
        base = 10;

    CORBA::ULongLong bound = MAX / base;
    CORBA::ULongLong last = MAX % base;
    const char* start = cp;
    result = 0;
    while(*cp)
    {
        CORBA::UShort num;
        char c = toupper(*cp);
        if(isdigit(c))
        {
            num = c - '0';
        }
        else if(isalpha(c) && c >= 'A' && c <= 'F' && base == 16)
        {
            num = c - 'A' + 10;
        }
        else
            break;

        if(num >= base)
            return false;

        //
        // Check for range error
        //
        if(result >= bound && num > last)
            return false;

        result *= base;
        result += num;
        cp++;
    }

    if(cp == start)
        return false;

    if(end)
        *end = cp;
    return true;
}

bool
OB::StringToLongLong(const char* str, const char** end, CORBA::UShort base,
                     CORBA::LongLong& result)
{
    const char* cp = str;
    bool neg = false;
    const CORBA::ULongLong UMAX = (~CORBA::ULongLong(0));
    const CORBA::LongLong LMAX = (CORBA::LongLong(UMAX >> 1));
    const CORBA::LongLong LMIN = (CORBA::LongLong(~LMAX));

    //
    // Skip leading whitespace
    //
    while(*cp && isspace(*cp))
        cp++;

    if(*cp == '-')
    {
        neg = true;
        cp++;
    }

    CORBA::ULongLong val;
    if(StringToULongLong(cp, end, base, val))
    {
        if(neg)
        {
            CORBA::ULongLong max =
                CORBA::ULongLong(-(LMIN + OB_LONGLONG(1))) + OB_ULONGLONG(1);
            if(val > max)
                return false;
        }
        else if(val > CORBA::ULongLong(LMAX))
            return false;

        if(neg)
            result = -CORBA::LongLong(val);
        else
            result = CORBA::LongLong(val);

        return true;
    }

    return false;
}

char*
OB::ULongLongToString(CORBA::ULongLong val)
{
    char buff[25];
    int len = 0;

    do
    {
        buff[len++] = (char)(val % 10) + '0';
        val /= 10;
    } while(val > 0);

    char* result = CORBA::string_alloc(len);
    for(int i = len - 1 ; i >= 0 ; i--)
        result[len - i - 1] = buff[i];
    result[len] = '\0';
    return result;
}

char*
OB::LongLongToString(CORBA::LongLong val)
{
    char buff[25];
    int len = 0;
    bool neg = (val < 0);

    CORBA::ULongLong v =
        neg ? CORBA::ULongLong(-(val + OB_LONGLONG(1)) + OB_LONGLONG(1)) :
              CORBA::ULongLong(val);

    do
    {
        buff[len++] = (char)(v % 10) + '0';
        v /= 10;
    } while(v > 0);

    if(neg)
        buff[len++] = '-';

    char* result = CORBA::string_alloc(len);
    for(int i = len - 1 ; i >= 0 ; i--)
        result[len - i - 1] = buff[i];
    result[len] = '\0';
    return result;
}

// ----------------------------------------------------------------------
// String memory management
// ----------------------------------------------------------------------

namespace OB
{

char* EmptyString = (char*)"";

} // End of namespace OB

char*
CORBA::string_alloc(CORBA::ULong len)
{
    return (char*)OB::Alloc(sizeof(char), len + 1);
}

void
CORBA::string_free(char* p)
{
    if(p && p != OB::EmptyString)
	OB::Free(p);
}

char*
CORBA::string_dup(const char* p)
{
    if(p)
    {
	if(p == OB::EmptyString)
	    return OB::EmptyString;

	char* ret = CORBA::string_alloc(strlen(p));
        if(ret)
            strcpy(ret, p);
	return ret;
    }
    else
	return 0;
}

// ----------------------------------------------------------------------
// String_var type
// ----------------------------------------------------------------------

CORBA::String_var::String_var(char* p)
    : ptr_(p)
{
    assert_nca(p != 0, OB::NCANullInit);
    assert_nca(p == 0 || p == OB::EmptyString || OB::AllocCheck(p),
	       OB::NCAWrongAlloc);
}

CORBA::String_var::String_var(const CORBA::String_var& r)
    : ptr_(string_dup(r.ptr_))
{
}

CORBA::String_var::String_var(const OB::StrForStruct& r)
    : ptr_(string_dup(r.ptr_))
{
}

CORBA::String_var::String_var(const OB::StrForSeq& r)
    : ptr_(string_dup(r))
{
}

CORBA::String_var&
CORBA::String_var::operator=(char* p)
{
    assert_nca(p == 0 || p == OB::EmptyString || OB::AllocCheck(p),
	       OB::NCAWrongAlloc);
    assert_nca(ptr_ == 0 || ptr_ == OB::EmptyString || p != ptr_,
	       OB::NCASelfAssign);

    string_free(ptr_);
    ptr_ = p;

    return *this;
}

CORBA::String_var&
CORBA::String_var::operator=(const char* p)
{
    if(p != ptr_)
    {
	string_free(ptr_);
	ptr_ = string_dup(p);
    }

    return *this;
}

CORBA::String_var&
CORBA::String_var::operator=(const CORBA::String_var& r)
{
    if(r.ptr_ != ptr_)
    {
	string_free(ptr_);
	ptr_ = string_dup(r.ptr_);
    }

    return *this;
}

CORBA::String_var&
CORBA::String_var::operator=(const OB::StrForStruct& r)
{
    if(r.ptr_ != ptr_)
    {
	string_free(ptr_);
	ptr_ = string_dup(r.ptr_);
    }

    return *this;
}

CORBA::String_var&
CORBA::String_var::operator=(const OB::StrForSeq& r)
{
    if(r.in() != ptr_)
    {
	string_free(ptr_);
	ptr_ = string_dup(r);
    }

    return *this;
}

char*&
CORBA::String_var::out()
{
    string_free(ptr_);
    ptr_ = 0;
    return ptr_;
}

char*
CORBA::String_var::_retn()
{
    char* ret = ptr_;
    ptr_ = 0;
    return ret;
}

CORBA::String_var&
CORBA::String_var::operator+=(const char* p)
{
    if(p)
    {
	if(ptr_)
	{
	    ULong max = 0;

	    if(ptr_ && ptr_ != OB::EmptyString)
		max = OB::Allocated(ptr_) - 1;

	    ULong newLen = strlen(ptr_) + strlen(p);

	    if(newLen > max)
	    {
		max = newLen > max * 2 ? newLen : max * 2;
		char* oldPtr = ptr_;
		ptr_ = string_alloc(max);
		strcpy(ptr_, oldPtr);
		string_free(oldPtr);
	    }

	    strcat(ptr_, p);
	}
	else
	    ptr_ = string_dup(p);
    }

    return *this;
}

CORBA::String_var&
CORBA::String_var::operator+=(char c)
{
    char s[2];
    s[0] = c;
    s[1] = '\0';
    return operator+=(s);
}

CORBA::String_var&
CORBA::String_var::operator+=(unsigned char c)
{
    return operator+=((char)c);
}

CORBA::String_var&
CORBA::String_var::operator+=(CORBA::LongLong n)
{
    String_var str = OB::LongLongToString(n);
    return operator+=(str);
}

CORBA::String_var&
CORBA::String_var::operator+=(CORBA::ULongLong n)
{
    String_var str = OB::ULongLongToString(n);
    return operator+=(str);
}

// ----------------------------------------------------------------------
// String type for string sequences
// ----------------------------------------------------------------------

OB::StrForSeq&
OB::StrForSeq::operator=(char* p)
{
    assert_nca(p == 0 || p == OB::EmptyString || OB::AllocCheck(p),
	       OB::NCAWrongAlloc);
    assert_nca(*ptr_ == 0 || *ptr_ == OB::EmptyString || p != *ptr_,
	       OB::NCASelfAssign);

    if(rel_)
	CORBA::string_free(*ptr_);

    *ptr_ = p;

    return *this;
}

OB::StrForSeq&
OB::StrForSeq::operator=(const char* p)
{
    if(p != *ptr_)
    {
	if(rel_)
	    CORBA::string_free(*ptr_);

	*ptr_ = CORBA::string_dup(p);
    }

    return *this;
}

OB::StrForSeq&
OB::StrForSeq::operator=(const CORBA::String_var& r)
{
    if(r.in() != *ptr_)
    {
	if(rel_)
	    CORBA::string_free(*ptr_);

	*ptr_ = CORBA::string_dup(r);
    }

    return *this;
}

OB::StrForSeq&
OB::StrForSeq::operator=(const OB::StrForStruct& r)
{
    if(r.in() != *ptr_)
    {
	if(rel_)
	    CORBA::string_free(*ptr_);

	*ptr_ = CORBA::string_dup(r);
    }

    return *this;
}

OB::StrForSeq&
OB::StrForSeq::operator=(const OB::StrForSeq& r)
{
    if(*(r.ptr_) != *ptr_)
    {
	if(rel_)
	    CORBA::string_free(*ptr_);

	*ptr_ = CORBA::string_dup(*(r.ptr_));
    }

    return *this;
}

char*&
OB::StrForSeq::out()
{
    if(rel_)
	CORBA::string_free(*ptr_);
    *ptr_ = 0;
    return *ptr_;
}

char*
OB::StrForSeq::_retn()
{
    char* ret = *ptr_;
    *ptr_ = 0;
    return ret;
}
