// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/Client.h>
#include <OBE/Except.h>
#include <OBE/OCIClient.h>
#include <OBE/ORBInstance.h>
#include <OBE/OBPolicies.h>
#include <OBE/RefCounted.h>

#include <GIOPClient.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::Client_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::Client_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(OB::ClientManager_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::ClientManager_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Client constructor and destructor
// ----------------------------------------------------------------------

Client::Client()
    : usage_(0)
{
}

Client::~Client()
{
}

// ----------------------------------------------------------------------
// ClientManager constructor and destructor
// ----------------------------------------------------------------------

ClientManager::ClientManager(ORBInstance_ptr instance)
    : destroy_(false),
      orbInstance_(ORBInstance::_duplicate(instance))
{
}

ClientManager::~ClientManager()
{
    assert(destroy_);
    assert(allClients_.empty());
    assert(reusableClients_.empty());
}

// ----------------------------------------------------------------------
// ClientManager public member implementation
// ----------------------------------------------------------------------

void
ClientManager::destroy()
{
    JTCSynchronized synchronized(*this);

    //
    // Don't destroy twice
    //
    if(destroy_)
	return;

    //
    // Set the destroy flag
    //
    destroy_ = true;

    //
    // Destroy all clients
    //
    for(OB_STL::list<Client_var>::iterator iter = allClients_.begin();
        iter != allClients_.end();
	iter++)
    {
        (*iter) -> destroy();
    }

    //
    // Reset internal data
    //
    orbInstance_ = ORBInstance::_nil();
    allClients_.clear();
    reusableClients_.clear();
}

OB_STL::list<ClientProfilePair_var>*
ClientManager::getClientProfilePairs(const IOP::IOR& ior,
                                     const PolicyList& policies)
{
    assert_nca(ior.type_id.in() != 0, NCANullString);

    if(::strlen(ior.type_id) == 0 && ior.profiles.length() == 0)
	throw INV_OBJREF(); // Can't create Client for a nil object

    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if this operation is called after ORB destruction
    //
    if(destroy_)
	throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    OB_STL::list<ClientProfilePair_var>* pairs =
        new OB_STL::list<ClientProfilePair_var>;
    try
    {
        ULong i, j;

        //
        // Find out whether private clients are requested
        //
        OB::ConnectionReusePolicy_var reusePolicy;
        for(j = 0; j < policies.length(); j++)
            if(policies[j] -> policy_type() == CONNECTION_REUSE_POLICY_ID)
                reusePolicy = ConnectionReusePolicy::_narrow(policies[j]);
        bool privateClients;
        if(!is_nil(reusePolicy) && reusePolicy -> value() == false)
            privateClients = true;
        else
            privateClients = false;

        //
        // First try to create CollocatedClient pairs
        //
	getCollocatedPairs(ior, policies, pairs);

        //
        // If no private clients are requested, add all existing reusable
        // clients, which are usable for the given IOR and policies
        //
        if(!privateClients)
        {
            for(OB_STL::list<Client_var>::iterator iter =
                reusableClients_.begin() ;
                iter != reusableClients_.end() ;
                iter++)
            {
                OCI::ProfileInfoSeq_var profileInfos =
                    (*iter) -> getUsableProfiles(ior, policies);
                for(j = 0 ; j < profileInfos -> length() ; j++)
                {
                    ClientProfilePair_var pair = new ClientProfilePair();
                    pair -> client = (*iter);
                    pair -> profile = new RefCountProfileInfo(profileInfos[j]);
                    pairs -> push_back(pair);
                }
            }
        }

        //
        // Finally, create new GIOPClients for all connectors we can get
        //
        OCI::ConFactoryRegistry_var conFactoryRegistry =
            orbInstance_ -> getConFactoryRegistry();
        OCI::ConFactorySeq_var factories =
            conFactoryRegistry -> get_factories();
        for(i = 0 ; i < factories -> length() ; i++)
        {
            OCI::ConnectorSeq_var connectors =
                factories[i] -> create_connectors(ior, policies);
            for(j = 0 ; j < connectors -> length() ; j++)
            {
                //
                // Get all usable profiles
                //
                OCI::ProfileInfoSeq_var profileInfos =
                    connectors[j] -> get_usable_profiles(ior, policies);
                assert(profileInfos -> length() >= 1);

                //
                // Create a new GIOPClient for each usable profile, and set
                // the concurrency model and code converters. Filter out
                // clients that are equivalent to other clients we already
                // have.
                //
                for(ULong k = 0 ; k < profileInfos -> length() ; k++)
                {
                    Client_var client =
                        new GIOPClient(orbInstance_, connectors[j]);

                    if(pairs -> size() > 0)
                    {
                        OB_STL::list<ClientProfilePair_var>::iterator iter;
                        
                        for(iter = pairs -> begin() ;
                            iter != pairs -> end() ;
                            iter++)
                        {
                            if((*iter) -> client -> equal(client))
                                break;
                        }
                        
                        if(iter != pairs -> end())
                        {
                            client -> destroy();
                            continue;
                        }
                    }

                    //
                    // Add the new client to the list of all clients
                    //
                    allClients_.push_back(client);

                    //
                    // Add client/profile pairs
                    //
                    OCI::ProfileInfoSeq_var clientProfileInfos =
                        client -> getUsableProfiles(ior, policies);
                    for(ULong l = 0 ; l < clientProfileInfos -> length() ; l++)
                    {
                        ClientProfilePair_var pair = new ClientProfilePair();
                        pair -> client = client;
                        pair -> profile =
                            new RefCountProfileInfo(clientProfileInfos[l]);
                        pairs -> push_back(pair);
                    }

                    //
                    // If no private clients have been requested, add it also
                    // to the list of existing reusable clients
                    //
                    if(!privateClients) 
                        reusableClients_.push_back(client);
                }
            }
        }

        //
        // If we still don't have any client/profile pairs, throw a
        // NO_RESOURCES exception
        //
        if(pairs -> size() == 0)
            throw NO_RESOURCES();

        //
        // Increment the usage count on all clients
        //
        OB_STL::list<ClientProfilePair_var>::iterator iter;
        for(iter = pairs -> begin() ; iter != pairs -> end() ; iter++)
            (*iter) -> client -> incUsage();
    }
    catch(...)
    {
        delete pairs;
        throw;
    }

    return pairs;
}

void
ClientManager::releaseClient(Client_ptr client)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if this operation is called after ORB destruction
    //
    if(destroy_)
	return;

    //
    // TODO:
    //
    // We need to examine the entire shutdown sequence again since the
    // GIOPClientWorkers and what-not get destroyed after the ORB has
    // been marked as destroyed
    //
    // if(destroy_)
    //     throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    bool inUse = client -> decUsage();

    if(!inUse)
    {
	OB_STL::list<Client_var>::iterator iter;

	for(iter = reusableClients_.begin();
	    iter != reusableClients_.end();
	    iter++)
	{
	    if((*iter).in() == client)
	    {
		reusableClients_.erase(iter);
		break;
	    }
	}
	
	for(iter = allClients_.begin();
	    iter != allClients_.end();
	    iter++)
	{
	    if((*iter).in() == client)
	    {
		client -> destroy();
		allClients_.erase(iter);
		return;
	    }
	}
    }
}
