// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/Config.h>

#ifndef HAVE_OBE_NO_COLLOC

#include <OBE/OAInterface.h>
#include <OBE/Except.h>
#include <OBE/ORBInstance.h>
#include <OBE/Downcall.h>
#include <OBE/RefCounted.h>
#include <OBE/Stream.h>

#include <CollocatedServer.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::CollocatedServer_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::CollocatedServer_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// CollocatedUpcallReturn public member implementation
// ----------------------------------------------------------------------

void
CollocatedUpcallReturn::upcallBeginReply(UpcallImpl*)
{
}

void
CollocatedUpcallReturn::upcallEndReply(UpcallImpl* upcall)
{
    OCI::BufferImpl* upOutBuf = upcall -> output() -> _OB_bufferImpl();
    upOutBuf -> pos(0);
    OCI::BufferImpl* downInBuf = down_ -> input() -> _OB_bufferImpl();
    downInBuf -> consume(upOutBuf);
    down_ -> setNoException();
}

void
CollocatedUpcallReturn::upcallUserException(UpcallImpl*,
					    const UserException& ex)
{
    down_ -> setUserException(ex);
}

void
CollocatedUpcallReturn::upcallSystemException(UpcallImpl*,
					      const SystemException& ex)
{
    down_ -> setSystemException(ex);
}

void
CollocatedUpcallReturn::upcallForward(UpcallImpl*,
				      RefCountIOR_ptr ior)
{
    down_ -> setLocationForward(ior);
}

// ----------------------------------------------------------------------
// CollocatedServer constructor and destructor
// ----------------------------------------------------------------------

CollocatedServer::CollocatedServer(OAInterface_ptr oaInterface)
    : Server(),
      destroy_(false),
      oaInterface_(OAInterface::_duplicate(oaInterface))
{
}

CollocatedServer::~CollocatedServer()
{
    assert(destroy_);
}

// ----------------------------------------------------------------------
// CollocatedServer public member implementation
// ----------------------------------------------------------------------

void
CollocatedServer::destroy()
{
    //
    // Set the destroy flag
    //
    destroy_ = true;
}

void
CollocatedServer::activate()
{
}

void
CollocatedServer::send(DowncallImpl* down)
    throw()
{
    //
    // May only be called if no response is expected
    //
    assert(!down -> responseExpected());

    //
    // Check whether we are destroyed
    //
    if(destroy_)
    {
	down -> setFailureException(TRANSIENT(0, COMPLETED_NO),
				    COMPLETED_NO);
	return;
    }

    //
    // This is a oneway call, and if there was no exception so far,
    // everything is fine
    //
    down -> setNoException();
    
    //
    // Create the upcall
    //
    UpcallImpl upcall(down -> orbInstance_nodup());
    upcall.setup(0,
		 &down -> profileInfo_nodup() -> value.key,
		 down -> requestId(),
		 down -> operation());
    
    //
    // Invoke the upcall
    //
    bool invoke = oaInterface_ -> setupUpcall(&upcall);
    if(invoke)
    {
	OCI::BufferImpl* downOutBuf = down -> output() -> _OB_bufferImpl();
	downOutBuf -> pos(0);
	OCI::BufferImpl* upInBuf = upcall.input() -> _OB_bufferImpl();
	upInBuf -> consume(downOutBuf);
	upcall.invoke();
    }
}

void
CollocatedServer::sendReceive(DowncallImpl* down)
    throw()
{
    //
    // May only be called if a response is expected
    //
    assert(down -> responseExpected());

    //
    // Check whether we are destroyed
    //
    if(destroy_)
    {
	down -> setFailureException(TRANSIENT(0, COMPLETED_NO),
				    COMPLETED_NO);
	return;
    }

    //
    // From here on, we consider the Downcall as pending
    //
    down -> setPending();
    
    //
    // Create the upcall
    //
    CollocatedUpcallReturn upcallReturn(down);
    UpcallImpl upcall(down -> orbInstance_nodup());
    upcall.setup(&upcallReturn,
		 &down -> profileInfo_nodup() -> value.key,
		 down -> requestId(),
		 down -> operation());

    //
    // Invoke the upcall
    //
    bool invoke = oaInterface_ -> setupUpcall(&upcall);
    if(invoke)
    {
	OCI::BufferImpl* downOutBuf = down -> output() -> _OB_bufferImpl();
	downOutBuf -> pos(0);
	OCI::BufferImpl* upInBuf = upcall.input() -> _OB_bufferImpl();
	upInBuf -> consume(downOutBuf);
	upcall.invoke();
    }

    //
    // The downcall may not be pending anymore at this point
    //
    assert(!down -> pending());
}

ULong
CollocatedServer::requestId()
{
    return 0;
}

OCI::ProfileInfoSeq*
CollocatedServer::getUsableProfiles(const OCI::IOR& ior,
				    const PolicyList& policies)
{
    return oaInterface_ -> getUsableProfiles(ior, policies);
}

#endif // !HAVE_OBE_NO_COLLOC
