// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/Except.h>
#include <OBE/Stream.h>
#include <OBE/ORBInstance.h>
#include <OBE/RefCounted.h>
#include <OBE/InitialServiceManager.h>
#include <OBE/IIOP.h>
#include <OBE/Logger.h>

#include <IIOPUtil.h>
#include <CorbaURL.h>

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <ctype.h>

using namespace CORBA;
using namespace OB;

// 
// corbaname / corbaloc default port number
// 
const UShort IIOPAddr::DEFAULT_PORT = 2809;

// 
// Corba URL Scheme Names
// 
// OMG Standard
// 
const char CorbaURL::iorPrefix_[] = "ior";
const char CorbaURL::corbalocPrefix_[] = "corbaloc";

#ifndef NORELFILE
// 
// ORBacus Extensions              
//
const char CorbaURL::filePrefix_[] = "file";
const char CorbaURL::relfilePrefix_[] = "relfile";
#endif

// 
// corbaloc / corbname protocol table
// 
const char CorbaURL::iiopProt_[] = "iiop:";
const char CorbaURL::iiopProtColon_[] = ":";  // iiop: shorthand
const char CorbaURL::rirProt_[] = "rir:"; 

// ----------------------------------------------------------------------
// CorbaURL constructor 
// ----------------------------------------------------------------------

CorbaURL::CorbaURL(const char* urlStr) : urlStr_(urlStr)
{
    parseScheme();
}

// ----------------------------------------------------------------------
// CorbaURL public member implementation
// ----------------------------------------------------------------------
    
OCI::ObjectKey& 
CorbaURL::strToKey(const char* start, const char* end, OCI::ObjectKey& key)
{
    if(start == 0)
        end = 0;
    else if(end == 0)
        end = start + strlen(start);

    ULong len = end - start;
    key.length(len);

    if(start)
        memcpy((char*)key.get_buffer(), start, len);

    return key;
}

Object_ptr
CorbaURL::toObject(ORBInstance_ptr orb)
{
#ifndef NORELFILE
    // 
    // WIN32 doesn't have an absolute root
    // for the `file:' scheme. A drive letter 
    // or network machine name may proceed the 
    // filename
    // 
#ifdef WIN32
    const char* fileAbsolutePrefix = 0;
#else
    const char* fileAbsolutePrefix = "/";
#endif
#endif

    if(strcmp(scheme_, iorPrefix_) == 0)
        return IORToObject(orb);

    if(strcmp(scheme_, corbalocPrefix_) == 0)
        return CORBALOCToObject(orb);

#ifndef NORELFILE
    if(strcmp(scheme_, filePrefix_) == 0)
        return OBFileToObject(orb, fileAbsolutePrefix);

    if(strcmp(scheme_, relfilePrefix_) ==0)
        return OBFileToObject(orb, 0);
#endif
    
    throw BAD_PARAM("URL: unrecognized scheme",
                    MinorBadSchemeName, COMPLETED_NO);

    return Object::_nil();      // some compiler need this
}

const char* 
CorbaURL::getScheme()
{
    return scheme_;
}

int
CorbaURL::urlStrToInt(const char* start, const char* end, 
                         int minVal, int maxVal)
{
    if(end == 0)
        end = start + strlen(start);    

    // 
    // Make sure string is in simple integer form
    // 
    const char* intCheck = start;
    
    while(intCheck != end)
    {
        if(!isdigit(*intCheck++))
            throw BAD_PARAM(
                "URL: numeric value expected",
                MinorBadAddress, COMPLETED_NO);
    }

    int val = atoi(start);

    // 
    //  check resulting value range if enabled
    // 
    if(minVal != maxVal && (val < minVal || val > maxVal))
            throw BAD_PARAM(
                "URL: numeric value out of range",
                MinorBadAddress, COMPLETED_NO);

    return val;
}

char*  
CorbaURL::lowerStr(char* start)
{
    char* strStart = start;
    
    while(*start != '\0')
    {
        *start = tolower(*start);
        start++;
    }

    return strStart;
}

char*
CorbaURL::unescapeURL(char* start, const char* end, const char*& unescapeEnd)
{
    if(end == 0)
        end = start + strlen(start);
    
    char* retnStr = start;
    char* cpChar = start;
    char ch;
        
    while(start != end)
    {
        ch = *start;
        // 
        // Escape sequence '%' must be followed by
        // 2 valid characters 0-9, A-F, a-f
        // 
        if(ch == '%')
        {
            if(start + 2 >= end)
                throw BAD_PARAM(
                    "URL: bad escape sequence length",
                    MinorBadAddress,COMPLETED_NO);

            char c1 = *++start;
            char c2 = *++start;
            ch = (convertHex(c1) << 4) | convertHex(c2);
        }
        *cpChar++ = ch;
        ++start;
    }

    *cpChar = '\0';
    
    // 
    // save true end so strings ending in %00 escapes are not lost
    // 
    if(&unescapeEnd != 0)
        unescapeEnd = cpChar;

    return retnStr;
}

int 
CorbaURL::convertHex(char ch)    
{
    if(ch >= '0' && ch <= '9')
        return ch - '0';
    if(ch >= 'A' && ch <= 'F')
        return (ch - 'A') + 10;
    if(ch >= 'a' && ch <= 'f')
        return (ch - 'a') + 10;

    throw BAD_PARAM(
        "URL: bad escape sequence character",
        MinorBadAddress,COMPLETED_NO);

    return 0;                   // some compilers need this
}

char* 
CorbaURL::copyStr(const char* start, const char* end)
{
    if(end == 0)
        end = start + strlen(start);
    
    char* newStr = string_alloc(end-start);
    strncpy(newStr, start, end-start);
    *(newStr + (end - start) ) = '\0';
    return newStr;
}

#ifndef NORELFILE
Object_ptr
CorbaURL::OBFileToObject(ORBInstance_ptr orbInstance, const char* dirPrefix)
{
    const char* start = urlStr_ + strlen(scheme_) + 1; // past the ":"

    //
    // allow up to three leading '/''s to match common use
    // forms of the "file:" URL scheme
    // 
    for(int s = 0; s < 3; ++s, ++start)
    {
        if(*start != '/')
            break;
    }
    
    if(*start == '\0')
        throw BAD_PARAM("URL: no file name specified", 
                        MinorBadSchemeSpecificPart, 
                        COMPLETED_NO);

    String_var fileName = dirPrefix;
    String_var urlFile = copyStr(start, 0);
    const char* urlEnd;
    unescapeURL(urlFile.inout(), 0, urlEnd);
    fileName += urlFile;
    
    FILE* f = fopen(fileName, "r");
    if(f != 0)
    {
        String_var s;
	char buf[64];

	while(fgets(buf, 64, f))
	    s += string_dup(buf);

        // 
        // Explicit close required for NT file locking
        // 
	fclose(f);

        //
        // Null terminate the read string
        //
        char *p = strpbrk(s.inout(), "\r\n");
        if(p != 0)
            *p = '\0';

        // 
        // TODO: Address potential recursion problem here
        // 
        ObjectFactory_var objectFactory =
            orbInstance -> getObjectFactory();
        return objectFactory -> stringToObject(s);
    }
    throw BAD_PARAM("URL: bad file name", 
                    MinorBadSchemeSpecificPart, 
                    COMPLETED_NO);
    
    return Object::_nil();      // some compiler need this
}
#endif

Object_ptr
CorbaURL::CORBALOCToObject(ORBInstance_ptr orbInstance)
{
    assert(strcmp(getScheme(), corbalocPrefix_) == 0);

    const char* addrEnd = parseCORBALOCAddresses();
    
    if(addresses_.length() == 0)
        throw BAD_PARAM(
            "URL: no protocol address", 
            MinorBadAddress, COMPLETED_NO);

    const char* keyStart;
    const char* keyEnd;
    
    // 
    // Normal key form scheme:protocol:addr[/key]
    // 
    if(addrEnd != 0 && *addrEnd == '/')
    {
        keyStart = addrEnd+1;
        keyEnd = parseKey(keyStart);
    }
    else if(addrEnd == 0)
    {
        // 
        // no key, use scheme or protocol default key 
        // 
        if(rirURL_ == true)
        {
            keyStart = "NameService";
            keyEnd = keyStart + strlen(keyStart);
            
        }
        else
        {
            keyStart = 0;
            keyEnd = 0;
        }

        strToKey(keyStart, keyEnd, key_);
    }
    else
    {
        // 
        // Should never happen
        // 
        throw INTERNAL("URL: unexpected characters",
                             MinorOther,
                             COMPLETED_NO);
    }
    
    //
    // Although no object-key is allowed by the INS specification
    // we're going to make this an invalid IOR
    //
    if(key_.length() == 0)
            throw BAD_PARAM("URL: no key specified",
                                  MinorOther,
                                  COMPLETED_NO);
    if(rirURL_ == true)
    {
        String_var initialToken = copyStr(keyStart, keyEnd);
        try 
        {
            return orbInstance -> getInitialServiceManager() ->
                resolveInitialReferences(initialToken.in());
        }
        // 
        // TODO - Any other exceptions to catch here?
        // 
        catch(const ORB::InvalidName& )
        {
            throw BAD_PARAM(
                "URL: invalid initial reference token", 
                MinorBadSchemeSpecificPart, 
                COMPLETED_NO);
        }
    }
    else
    {
        IOP::IOR_var ior = new IOP::IOR;
        ior -> type_id = (const char* )"";
        ior ->profiles.length(addresses_.length());

        for(ULong pr = 0; pr < addresses_.length(); ++pr)
        {
            addresses_[pr] -> setProfile(ior.inout(), pr, key_);
        }

        ObjectFactory_var objectFactory = orbInstance -> getObjectFactory();
        RefCountIOR_var refIOR = new RefCountIOR(ior);
        Object_ptr obj = objectFactory -> createObject(refIOR);
        
        return obj;
    }

    return Object::_nil();      // some compilers need this
}

Object_ptr
CorbaURL::IORToObject(ORBInstance_ptr orbInstance)
{
    assert(strcmp(getScheme(), iorPrefix_) == 0);

    const char* s = urlStr_+strlen(iorPrefix_)+1; // +1 = ":"
    ULong len = strlen(s); 

    if(len % 2)
	throw BAD_PARAM(
            "URL: invalid length", 
            MinorBadSchemeSpecificPart, COMPLETED_NO);

    Octet* octBuf = 0;
    octBuf = AsciiToOctets(s);
    InputStreamImpl in(len / 2, 0, false);
    memcpy(in._OB_bufferImpl() -> data(), octBuf, len / 2);
    delete [] octBuf;

    try
    {
        in._OB_readEndian();
        IOP::IOR_var ior = new IOP::IOR;
        IOP::IOR::_OB_unmarshal(ior.inout(), &in);
        ObjectFactory_var objectFactory = orbInstance -> getObjectFactory();
        RefCountIOR_var refIOR = new RefCountIOR(ior);
        return objectFactory -> createObject(refIOR);
    }
    catch(const CORBA::MARSHAL&)
    {
        // In this case, a marshal error is really 
        // a bad "IOR:..." string 
        // 
        delete [] octBuf;
        throw CORBA::BAD_PARAM("URL: invalid",
                               MinorBadSchemeSpecificPart,
                               COMPLETED_NO);
    }

    return Object::_nil(); // Some compilers need this
}

const char* 
CorbaURL::parseKey(const char* keyStrStart)
{
    String_var keyStrCopy = string_dup(keyStrStart);
    const char* keyStrEnd;

    unescapeURL(keyStrCopy.inout(), 0, keyStrEnd);
    strToKey(keyStrCopy, keyStrEnd, key_);

    return keyStrEnd;
}

URLAddr*
CorbaURL::parseSingleProtocolAddr(const char* addrStart, 
                                  const char* addrEnd)
{
    String_var addr = copyStr(addrStart, addrEnd);

    // 
    // trueAddrEnd will not be used by iiop and rir but
    // is required for generic protocol processing 
    // 
    const char* trueAddrEnd;
    unescapeURL(addr.inout(), 0, trueAddrEnd);

    // 
    // IIOP:
    // 
    if(strncmp(addr.in(), iiopProt_, sizeof iiopProt_ - 1) == 0 ||
        strncmp(addr.in(), iiopProtColon_, sizeof iiopProtColon_ - 1) == 0 )
        return new IIOPAddr(addr.in(), trueAddrEnd);

    // 
    // RIR:
    // 
    if(strncmp(addr.in(), rirProt_, sizeof rirProt_ - 1) == 0)
    {
        rirURL_ = true; 
        return new RIRAddr(addr.in(), trueAddrEnd);
    }

    // 
    // Add New Schemes Here (ATM....)
    // 

    throw BAD_PARAM("URL: unknown protocol",
                           MinorBadSchemeSpecificPart,
                          COMPLETED_NO);
    return 0;                   // some compilers need this
}

void 
CorbaURL::parseScheme()
{
    char* scheme = 0;
    
    if(urlStr_ != 0)
    {
        const char* schemeEnd = strchr(urlStr_, ':');

        if(schemeEnd != 0)
            scheme = lowerStr(copyStr(urlStr_, schemeEnd));
    }
    
    if(scheme == 0)
        throw BAD_PARAM(
            "URL: no scheme found",
            MinorBadSchemeName, COMPLETED_NO);

    scheme_ = scheme;
}

const char* 
CorbaURL::parseCORBALOCAddresses()
{
    // 
    // addr starts scheme's ':'
    // 
    const char* addrStart = urlStr_+strlen(scheme_)+1; // +1=':' 
    rirURL_ = false;
    const char* addr = addrStart;
    const char* addrEnd = addr;

    // 
    // parse addresses (separated by ',') before '/' or end of string
    // 
    while(addrEnd != 0 && *addrEnd != '/')
    {
        addrEnd = strpbrk(addr, ",/");

        URLAddr *urlAddr = parseSingleProtocolAddr(addr, addrEnd);
        CORBA::ULong addressesLength = addresses_.length();
        addresses_.length(addressesLength + 1);
        addresses_[addressesLength] = urlAddr;

        if(rirURL_ == true && addresses_.length() > 1)
            throw BAD_PARAM(
                "URL: rir cannot be used with other protocols",
                MinorBadSchemeSpecificPart, COMPLETED_NO);

        addr = addrEnd + 1;
    }

    return addrEnd;
}

// ----------------------------------------------------------------------
// URLAddr destructor 
// ----------------------------------------------------------------------

URLAddr::~URLAddr()
{
}

// ----------------------------------------------------------------------
// RIRAddr constructor
// ----------------------------------------------------------------------

RIRAddr::RIRAddr(const char* addrStart, const char* ) 
{
    addrStart += sizeof CorbaURL::rirProt_ - 1;

    if(*addrStart != '\0')
        throw BAD_PARAM(
            "URL: rir does not allow an address",
            MinorBadAddress, COMPLETED_NO);
}

// ----------------------------------------------------------------------
// RIRAddr public member implementation
// ----------------------------------------------------------------------

void
RIRAddr::setProfile(IOP::IOR& ior, int profileNumber, 
			 const OCI::ObjectKey& key)
{
    // 
    // rir: should never be adding a profile to an IOR
    // 
    throw INTERNAL("URL: rir internal evluation failure", 
                         MinorOther, COMPLETED_NO);
}

const char*
RIRAddr::getProtocolName() 
{
    return CorbaURL::rirProt_;
}

// ----------------------------------------------------------------------
// IIOPAddr constructor
// ----------------------------------------------------------------------

IIOPAddr::IIOPAddr(const char* addrStart, const char* )
{
    addrStart += (*addrStart == CorbaURL::iiopProtColon_[0]) ? 
        sizeof CorbaURL::iiopProtColon_ - 1 : 
        sizeof CorbaURL::iiopProt_ - 1 ;
    
    //
    // Do we have an iiop version 'X.Y@' portion? (default is 1.0)
    //
    const char* at = strchr(addrStart, '@');

    if(at == 0)
    {
        major_ = 1;
        minor_ = 0;
    }
    else
    {
        const char* dot = strchr(addrStart, '.');
        if(dot == 0)
            throw BAD_PARAM(
                "URL: iiop version must be of the form `X.Y'", 
                MinorBadAddress, COMPLETED_NO);
        
        // 
        // Major Version Must start at 1
        // Minor Version can start at 0
        // 
        major_ = CorbaURL::urlStrToInt(addrStart, dot, 1, 255);
        minor_ = CorbaURL::urlStrToInt(dot + 1, at, 0, 255);

        //
        // The end of the address portion is now the character
        // after the '@'
        //
        addrStart = at + 1;
    }

    // 
    // empty hostname is illegal (as is port ':YYYY' by itself)
    // 
    if(addrStart == 0 || *addrStart=='\0' || *addrStart==':')
        throw BAD_PARAM("URL: iiop hostname must be specified", 
                              MinorBadAddress, COMPLETED_NO);
    
    //
    // Find optional ':' port number separator or end of string 
    //
    const char* colon = strchr(addrStart, ':');
    host_ = CorbaURL::copyStr(addrStart, colon);
    
    // 
    // port specified ? Valid range 1 - 65535
    // 
    if(colon != 0)
    {
        const char* portStr = colon + 1;
        port_ = (UShort)CorbaURL::urlStrToInt(portStr, 0, 1, 65535);
    }
    else 
        port_ = DEFAULT_PORT;           
}

// ----------------------------------------------------------------------
// RIRAddr public member implementation
// ----------------------------------------------------------------------

void
IIOPAddr::setProfile(IOP::IOR& ior, int profileNumber, 
		     const OCI::ObjectKey& key)
{   
    ior.profiles[profileNumber].tag = IOP::TAG_INTERNET_IOP;
    
    ::IIOP::ProfileBody_1_0 body;
    body.iiop_version.major = 1;
    body.iiop_version.minor = 0;
    body.host = host_;
    body.port = port_;
    body.object_key.length(key.length());
    memcpy(body.object_key.get_buffer(), key.get_buffer(),
	   key.length());
    
    OutputStreamImpl out;
    out._OB_writeEndian();
    body._OB_marshal(&out);
    ULong count = out._OB_pos();
    ior.profiles[profileNumber].profile_data.length(count);
    Octet* oct = ior.profiles[profileNumber].profile_data.get_buffer();
    memcpy(oct, out._OB_bufferImpl() -> data(), count);
}

const char*
IIOPAddr::getProtocolName()
{
    return CorbaURL::iiopProtColon_;
}
