// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/MarshalStubImpl.h>
#include <OBE/Downcall.h>
#include <OBE/Client.h>
#include <OBE/ORBInstance.h>
#include <OBE/Stream.h>
#include <OBE/RefCounted.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(DowncallEmitter_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(DowncallEmitter_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// DowncallImpl private and protected members
// ----------------------------------------------------------------------

void
DowncallImpl::checkForException()
    throw(ExceptionBase)
{
    switch(state_)
    {
    case DowncallStateUserException:
	if(ex_) // Only raise if a user exception has been set
	    ex_ -> _raise();
	break;

    case DowncallStateSystemException:
        assert(ex_);
	ex_ -> _raise();
	break;

    case DowncallStateFailureException:
	assert(ex_);
	//
	// COMPILERFIX HPUX 64
	//
	// was throw FailureException(*SystemException::_downcast(ex_));
	//
        throw FailureException(*(dynamic_cast<SystemException*>(ex_)));

    case DowncallStateForward:
	assert(!is_nil(forwardIOR_));
	throw LocationForward(forwardIOR_);

    default:
	break;
    }
}

// ----------------------------------------------------------------------
// DowncallImpl constructor and destructor
// ----------------------------------------------------------------------

DowncallImpl::DowncallImpl(ORBInstance_ptr orbInstance,
			   Client_ptr client,
			   RefCountProfileInfo_ptr profileInfo,
			   RefCountPolicyList_ptr policies,
			   const char* op, ULong opLen,
			   bool resp)
    : orbInstance_(orbInstance),
      client_(client),
      profileInfo_(profileInfo),
      policies_(policies),
      op_(op),
      opLen_(opLen),
      out_(12),
      in_(12, 0, false),
      state_(DowncallStateUnsent),
      ex_(0),
      responseExpected_(resp)
{
    //
    // Get the next request ID
    //
    reqId_ = client_ -> requestId();
}

DowncallImpl::~DowncallImpl()
{
    delete ex_;
}

// ----------------------------------------------------------------------
// DowncallImpl public member implementation
// ----------------------------------------------------------------------

OutputStreamImpl*
DowncallImpl::preMarshal()
    throw(ExceptionBase)
{
    emitter_ = client_ -> startDowncall(this);
    checkForException();
    return &out_;
}

void
DowncallImpl::marshalEx(const SystemException& ex)
    throw(ExceptionBase)
{
    setFailureException(ex, COMPLETED_NO);
    checkForException();
    assert(false);
}

void
DowncallImpl::postMarshal()
    throw(ExceptionBase)
{
}

void
DowncallImpl::request()
    throw(ExceptionBase)
{
    assert(responseExpected_);

    emitter_ -> sendReceive(this);
    checkForException();
}

void
DowncallImpl::oneway()
    throw(ExceptionBase)
{
    assert(!responseExpected_);

    emitter_ -> send(this);
    checkForException();
}

InputStreamImpl*
DowncallImpl::preUnmarshal()
    throw(ExceptionBase)
{
    return &in_;
}

void
DowncallImpl::unmarshalEx(const SystemException& ex)
    throw(ExceptionBase)
{
    setFailureException(ex, COMPLETED_MAYBE);
    checkForException();
    assert(false);
}

void
DowncallImpl::postUnmarshal()
    throw(ExceptionBase)
{
    //
    // If the result of this downcall is a user exception, but no user
    // exception could be unmarshalled, then use the system exception
    // UNKNOWN
    //
    if(state_ == DowncallStateUserException && !ex_)
	setSystemException(UNKNOWN(MinorUnknownUserException, COMPLETED_YES));

    checkForException();
}

char*
DowncallImpl::unmarshalExceptionId()
{
    assert(state_ == DowncallStateUserException);
    ULong pos = in_._OB_pos();
    String_var id = in_.read_string();
    in_._OB_pos(pos);
    return id._retn();
}

void
DowncallImpl::setPending()
    throw()
{
    assert(responseExpected_);
    state_ = DowncallStatePending;
}

void
DowncallImpl::setNoException()
    throw()
{
    state_ = DowncallStateNoException;
    in_._OB_ORBInstance(orbInstance_);
}

void
DowncallImpl::setUserException()
    throw()
{
    assert(responseExpected_);
    state_ = DowncallStateUserException;
    in_._OB_ORBInstance(orbInstance_);
}

void
DowncallImpl::setUserException(UserException* ex)
    throw()
{
    assert(responseExpected_);
    assert(!ex_);
    state_ = DowncallStateUserException;
    ex_ = ex;
}

void
DowncallImpl::setUserException(const UserException& ex)
    throw()
{
    assert(responseExpected_);
    assert(!ex_);
    state_ = DowncallStateUserException;
    ex_ = ex._OB_clone();
}

void
DowncallImpl::setSystemException(const SystemException& ex)
    throw()
{
    assert(responseExpected_);
    assert(!ex_);
    state_ = DowncallStateSystemException;
    ex_ = ex._OB_clone();
}

void
DowncallImpl::setFailureException(const SystemException& ex,
				  CompletionStatus completed)
    throw()
{
    assert(!ex_);
    state_ = DowncallStateFailureException;
    ex_ = ex._OB_clone();
    (dynamic_cast<SystemException*>(ex_)) ->completed(completed);
}

void
DowncallImpl::setLocationForward(RefCountIOR_ptr ior)
    throw()
{
    assert(responseExpected_);
    assert(is_nil(forwardIOR_));
    state_ = DowncallStateForward;
    forwardIOR_ = RefCountIOR::_duplicate(ior);
}
