// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/Except.h>
#include <OBE/MarshalStubImpl.h>
#include <OBE/Stream.h>
#include <OBE/RefCounted.h>

#include <stdio.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// Exception public member implementation
// ----------------------------------------------------------------------

const char*
Exception::_rep_id() const
{
    assert(false);
    return 0;
}

// ----------------------------------------------------------------------
// LocationForward constructor
// ----------------------------------------------------------------------

LocationForward::LocationForward(RefCountIOR_ptr i)
    : ior(RefCountIOR::_duplicate(i))
{
}

// ----------------------------------------------------------------------
// FailureException constructors and destructor
// ----------------------------------------------------------------------

FailureException::FailureException(const SystemException& ex)
    : exception(dynamic_cast<SystemException*>(ex._OB_clone()))
{
}

FailureException::FailureException(const FailureException& ex)
    : exception(dynamic_cast<SystemException*>(ex.exception -> _OB_clone()))
{
}

FailureException::~FailureException()
{
    delete exception;
}

// ----------------------------------------------------------------------
// FailureException public member implementation
// ----------------------------------------------------------------------

FailureException&
FailureException::operator=(const FailureException& ex)
{
    if(&ex != this)
    {
	delete exception;
	exception = dynamic_cast<SystemException*>(ex.exception -> _OB_clone());
    }
    return *this;
}

// ----------------------------------------------------------------------
// SystemException public member implementation
// ----------------------------------------------------------------------

SystemException::SystemException(const char* reason)
    : reason_(reason), minor_(0), status_(COMPLETED_NO)
{
}

SystemException::SystemException(ULong minor, CompletionStatus status)
    : reason_(string_dup("")), minor_(minor), status_(status)
{
}

SystemException::SystemException(const char* reason,
				 ULong minor,
				 CompletionStatus status)
    : reason_(reason), minor_(minor), status_(status)
{
}

SystemException::SystemException(const SystemException& se)
    : reason_(se.reason_),
      minor_(se.minor_), status_(se.status_)
{
}

SystemException&
SystemException::operator=(const SystemException& se)
{
    if(this != &se)
    {
	reason_ = se.reason_;
	minor_ = se.minor_;
	status_ = se.status_;
    }
    return *this;
}

char*
SystemException::_OB_defaultToString(const char* exceptionName) const
{
    String_var s = string_dup("System exception `");
    s += exceptionName;
    s += "'";

    if(reason_.in() && strlen(reason_) > 0)
    {
	s += "\nReason: ";
	s += reason_;
    }

    s += "\nCompleted: ";

    switch(status_)
    {
    case COMPLETED_YES:
	s += "yes";
	break;

    case COMPLETED_NO:
	s += "no";
	break;

    case COMPLETED_MAYBE:
	s += "maybe";
	break;
    }

    s += "\nMinor code: ";
    char minorHex[64];
    sprintf(minorHex, "0x%lx", (unsigned long)minor_);
    s += minorHex;

    return s._retn();
}

// ----------------------------------------------------------------------
// UserException public member implementation
// ----------------------------------------------------------------------

char*
UserException::_OB_defaultToString(const char* exceptionName) const
{
    String_var s = string_dup("User exception `");
    s += exceptionName;
    s += "'";

    return s._retn();
}

void
UserException::_OB_marshal(OutputStreamImpl* out) const
{
    out -> write_string(_rep_id());
}

// ----------------------------------------------------------------------
// All specific system exception public member implementation
// ----------------------------------------------------------------------

#define OB_MAKE_STANDARD_EXCEPTION_I(T) \
\
T::T(const char* r) \
    : SystemException(r) \
{ \
} \
T::T(ULong m, CompletionStatus s) \
    : SystemException(m, s) \
{ \
} \
T::T(const char* r, ULong m, CompletionStatus s) \
    : SystemException(r, m, s) \
{ \
} \
T::T(const T& ex) \
    : SystemException(ex) \
{ \
} \
\
const char* \
T::_rep_id() const \
{ \
    return "IDL:omg.org/CORBA/" #T ":1.0"; \
} \
\
Exception* \
T::_OB_clone() const \
{ \
    return new T(*this); \
}

//
// UNKNOWN
//
OB_MAKE_STANDARD_EXCEPTION_I(UNKNOWN)

//
// BAD_PARAM
//
OB_MAKE_STANDARD_EXCEPTION_I(BAD_PARAM)

//
// NO_MEMORY
//
OB_MAKE_STANDARD_EXCEPTION_I(NO_MEMORY)

//
// IMP_LIMIT
//
OB_MAKE_STANDARD_EXCEPTION_I(IMP_LIMIT)

//
// COMM_FAILURE
//
OB_MAKE_STANDARD_EXCEPTION_I(COMM_FAILURE)

//
// INV_OBJREF
//
OB_MAKE_STANDARD_EXCEPTION_I(INV_OBJREF)

//
// NO_PERMISSION
//
OB_MAKE_STANDARD_EXCEPTION_I(NO_PERMISSION)

//
// INTERNAL
//
OB_MAKE_STANDARD_EXCEPTION_I(INTERNAL)

//
// MARSHAL
//
OB_MAKE_STANDARD_EXCEPTION_I(MARSHAL)

//
// INITIALIZE
//
OB_MAKE_STANDARD_EXCEPTION_I(INITIALIZE)

//
// NO_IMPLEMENT
//
OB_MAKE_STANDARD_EXCEPTION_I(NO_IMPLEMENT)

//
// BAD_TYPECODE
//
OB_MAKE_STANDARD_EXCEPTION_I(BAD_TYPECODE)

//
// BAD_OPERATION
//
OB_MAKE_STANDARD_EXCEPTION_I(BAD_OPERATION)

//
// NO_RESOURCES
//
OB_MAKE_STANDARD_EXCEPTION_I(NO_RESOURCES)

//
// NO_RESPONSE
//
OB_MAKE_STANDARD_EXCEPTION_I(NO_RESPONSE)

//
// BAD_INV_ORDER
//
OB_MAKE_STANDARD_EXCEPTION_I(BAD_INV_ORDER)

//
// TRANSIENT
//
OB_MAKE_STANDARD_EXCEPTION_I(TRANSIENT)

//
// OBJ_ADAPTER
//
OB_MAKE_STANDARD_EXCEPTION_I(OBJ_ADAPTER)

//
// OBJECT_NOT_EXIST
//
OB_MAKE_STANDARD_EXCEPTION_I(OBJECT_NOT_EXIST)

//
// INV_POLICY
//
OB_MAKE_STANDARD_EXCEPTION_I(INV_POLICY)

// ----------------------------------------------------------------------
// CDR functions
// ----------------------------------------------------------------------

void
SystemException::_OB_marshal(OutputStreamImpl* out) const
{
    out -> write_string(_rep_id());
    out -> write_ulong(minor_);
    out -> write_ulong(status_);
}

SystemException*
SystemException::_OB_unmarshal(InputStreamImpl* in)
{
    const char* id = in -> read_string_nodup();
    ULong minor = in -> read_ulong();
    CompletionStatus status = (CompletionStatus)in -> read_ulong();

    if(strcmp(id, "IDL:omg.org/CORBA/BAD_PARAM:1.0") == 0)
    {
	return new BAD_PARAM(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_MEMORY:1.0") == 0)
    {
	return new NO_MEMORY(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/IMP_LIMIT:1.0") == 0)
    {
	return new IMP_LIMIT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/COMM_FAILURE:1.0") == 0)
    {
	return new COMM_FAILURE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INV_OBJREF:1.0") == 0)
    {
	return new INV_OBJREF(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_PERMISSION:1.0") == 0)
    {
	return new NO_PERMISSION(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INTERNAL:1.0") == 0)
    {
	return new INTERNAL(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/MARSHAL:1.0") == 0)
    {
	return new MARSHAL(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INITIALIZE:1.0") == 0)
    {
	return new INITIALIZE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_IMPLEMENT:1.0") == 0)
    {
	return new NO_IMPLEMENT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/BAD_TYPECODE:1.0") == 0)
    {
	return new BAD_TYPECODE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/BAD_OPERATION:1.0") == 0)
    {
	return new BAD_OPERATION(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_RESOURCES:1.0") == 0)
    {
	return new NO_RESOURCES(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/NO_RESPONSE:1.0") == 0)
    {
	return new NO_RESPONSE(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/BAD_INV_ORDER:1.0") == 0)
    {
	return new BAD_INV_ORDER(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/TRANSIENT:1.0") == 0)
    {
	return new TRANSIENT(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/OBJ_ADAPTER:1.0") == 0)
    {
	return new OBJ_ADAPTER(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/OBJECT_NOT_EXIST:1.0") == 0)
    {
	return new OBJECT_NOT_EXIST(minor, status);
    }
    else if(strcmp(id, "IDL:omg.org/CORBA/INV_POLICY:1.0") == 0)
    {
	return new INV_POLICY(minor, status);
    }

    //
    // Unknown exception
    //
    return new UNKNOWN(minor, status);
}
