// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/Object.h>
#include <OBE/Stream.h>
#include <OBE/Logger.h>
#include <OBE/OBPolicies.h>
#include <OBE/OCIClient.h>
#include <OBE/RefCounted.h>
#include <OBE/ORBInstance.h>
#ifndef HAVE_OBE_NO_TRACE
#  include <OBE/CoreTraceLevels.h>
#endif // !HAVE_OBE_NO_TRACE

#include <GIOPClient.h>
#include <GIOPClientWorker.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::GIOPClient_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPClient_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// GIOPClient private and protected member implementations
// ----------------------------------------------------------------------

GIOPClientWorker_ptr
GIOPClient::getWorker(bool create, Long t)
{
    JTCSynchronized synchronized(*this);

    if(destroy_)
	throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);
    
    if(is_nil(worker_) && create)
    {
#ifndef HAVE_OBE_NO_TRACE
	//
	// Trace connection attempt
	//
	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceConnections())
	{
	    String_var msg = string_dup("trying to establish connection\n");
	    msg += "timeout: ";
	    if(t >= 0)
	    {
		msg += t;
		msg += "ms\n";
	    }
	    else
		msg += "none\n";
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("outgoing", msg);
	}
#endif // !HAVE_OBE_NO_TRACE
	
	//
	// Create new transport, using the connector
	//
	// TODO: In reactive mode, this is not ideal. A non-blocking
	// connect() operation should be added to the OCI. Then a
	// GIOPClientStarterReactive class should be written, that
	// registers with the Reactor in order to listen to write
	// events on the filedescriptor of the non-blocking Connector
	// (handle() must be added to the Connector). If it receives a
	// write event, it gets the Transport from the Connector by
	// calling an operation complete_connect() (must also be
	// added). Then it creates an GIOPClientWorkerReactive with
	// the newly created Transport.
	//
	// For symetry reasons, GIOPClientStarterBlocking and
	// GIOPClientStarterThreaded should also be added, even though
	// these classes only have a trivial functionality. Or perhaps
	// the GIOPClientStarterThreaded tries to connect() in the
	// backgound? Just an idea...
	//
	
	OCI::Transport_var transport;
	
	if(t >= 0)
	{
	    transport = connector_ -> connect_timeout((ULong)t);
	    
	    //
	    // Was there a timeout?
	    //
	    if(is_nil(transport))
		throw NO_RESPONSE(0, COMPLETED_NO);
	}
	else
	{
	    transport = connector_ -> connect();
	    assert(!is_nil(transport));
	}
	
	worker_ = new GIOPClientWorker(orbInstance_, this, transport);
    }
    
    return GIOPClientWorker::_duplicate(worker_);
}

// ----------------------------------------------------------------------
// GIOPClient constructor and destructor
// ----------------------------------------------------------------------

GIOPClient::GIOPClient(ORBInstance_ptr orbInstance,
		       OCI::Connector_ptr connector)
    : Client(),
      orbInstance_(ORBInstance::_duplicate(orbInstance)),
      nextRequestId_(0),
      connector_(OCI::Connector::_duplicate(connector)),
      destroy_(false)
{
}

GIOPClient::~GIOPClient()
{
    assert(destroy_);
    assert(is_nil(worker_));
}

// ----------------------------------------------------------------------
// GIOPClient public member implementation
// ----------------------------------------------------------------------

void
GIOPClient::destroy()
{
    GIOPClientWorker_var worker;
    
    {
	JTCSynchronized synchronized(*this);
	    
	//
	// Don't destroy twice
	//
	if(destroy_)
	    return;

	//
	// Set the destroy flag
	//
	destroy_ = true;

	//
	// Use a copy of the worker, and destroy the worker outside
	// the synchronization, to avoid deadlocks
	//
	worker = worker_._retn();
    }

    if(!is_nil(worker))
	worker -> destroy();
}

void
GIOPClient::removeWorker(GIOPClientWorker_ptr worker)
{
    JTCSynchronized synchronized(*this);

    if(worker_.in() == worker)
	worker_ = GIOPClientWorker::_nil();
}

ULong
GIOPClient::requestId()
{
    JTCSynchronized synchronized(nextRequestIdMutex_);
    return nextRequestId_++;
}

OCI::ProfileInfoSeq*
GIOPClient::getUsableProfiles(const OCI::IOR& ior, const PolicyList& policies)
{
    //
    // Get all profiles usable for the connector
    //
    OCI::ProfileInfoSeq_var all =
	connector_ -> get_usable_profiles(ior, policies);
    
    return all._retn();
}

DowncallEmitter_ptr
GIOPClient::startDowncall(DowncallImpl* down)
    throw()
{
    GIOPClientWorker_var worker;
    try
    {
	//
	// Get the worker, creating a new one if there is none
	// available
	//
	worker = getWorker(true, down -> policies_nodup() -> connectTimeout);
    }
    catch(const SystemException& ex)
    {
	down -> setFailureException(ex, COMPLETED_NO);
	return DowncallEmitter::_nil();
    }

    try
    {
	//
	// Get the output stream and skip the message header
	//
	OutputStreamImpl* out = down -> output();
	out -> _OB_pos(12);

	out -> write_ulong(0); // service_context
        out -> write_ulong(down -> requestId()); // request_id
        out -> write_boolean(down -> responseExpected()); // response_expected
	
        //
        // reserved
        //
        static Octet reserved[3] = { 0, 0, 0 };
        out -> write_octet_array(reserved, 3);
	
        //
        // object_key
        //
	const OCI::ProfileInfo& profileInfo =
	    down -> profileInfo_nodup() -> value;
        ULong keyLen = profileInfo.key.length();
        out -> write_ulong(keyLen);
        out -> write_octet_array(profileInfo.key.get_buffer(), keyLen);

        //
        // operation (we use octets to avoid codeset conversion)
        //
	const char* op = down -> operation();
        ULong opLen = down -> operationLength() + 1;
        out -> write_ulong(opLen);
        out -> write_octet_array((Octet*)op, opLen);

        out -> write_ulong(0); // requesting_principal

	return worker._retn();
    }
    catch(const SystemException& ex)
    {
	down -> setFailureException(ex, COMPLETED_NO);
	return DowncallEmitter::_nil();
    }
}

bool
GIOPClient::equal(Client_ptr cl)
{
    GIOPClient_ptr client = dynamic_cast<GIOPClient_ptr>(cl);
    if(is_nil(client))
	return false;
    
    if(!connector_ -> equal(client -> connector_))
	return false;
    
    return true;
}
