// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/Object.h>
#include <OBE/Stream.h>
#include <OBE/OCIClient.h>
#include <OBE/RefCounted.h>
#include <OBE/ORBInstance.h>
#include <OBE/Logger.h>
#ifndef HAVE_OBE_NO_TRACE
#  include <OBE/CoreTraceLevels.h>
#endif
#include <OBE/Properties.h>
#include <OBE/NativeTypes.h>

#include <GIOPClient.h>
#include <GIOPClientWorker.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::GIOPClientWorker_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPClientWorker_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// GIOPClientWorker private and protected member implementations
// ----------------------------------------------------------------------

#ifndef HAVE_OBE_NO_TRACE
void
GIOPClientWorker::logCloseConnection()
{
    if(!closeLogged_)
    {
	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceConnections())
	{
	    String_var msg = string_dup("closing connection\n");
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("outgoing", msg);
	}
	
	closeLogged_ = true;
    }
}
#endif // !HAVE_OBE_NO_TRACE

void
GIOPClientWorker::gracefulShutdown()
{
    //
    // COMPILERFIX: witout this fix exceptions are not caught in
    // the following try/catch block.
    //
#if defined(__GNUC__) && defined(__OPTIMIZE__) 
    volatile int dummy = 1;
    dummy++;
#endif
    
    //
    // Now we gracefully shutdown the connection
    //
    try
    {
#ifndef HAVE_OBE_NO_TRACE
	logCloseConnection();
#endif // !HAVE_OBE_NO_TRACE
	transport_ -> shutdown();
	
/*
	while(true)
	{
	    //
	    // Setup the incoming message buffer
	    //
	    if(is_nil(buf))
		buf = new OCI::Buffer(12);
	    
	    if(buf -> pos() < 12)
	    {
		if(shutdownTimeout_ > 0)
		{
		    //
		    // Receive header, with timeout
		    //
		    transport_ -> receive_timeout(buf,
						  shutdownTimeout_ * 1000);
		    if(!buf -> is_full())
			throw TRANSIENT(MinorForcedShutdown,
					COMPLETED_MAYBE);
		}
		else
		{
		    //
		    // Receive header, blocking
		    //
		    transport_ -> receive(buf, true);
			assert(buf -> is_full());
		}
		
		//
		// Header is complete
		//
		incoming_.extractHeader(buf);
		buf -> realloc(12 + incoming_.size());
	    }
	    
	    if(!buf -> is_full())
	    {
		if(shutdownTimeout_ > 0)
		{
		    //
		    // Receive body, with timeout
		    //
		    transport_ -> receive_timeout(buf,
						  shutdownTimeout_ * 1000);
		    if(!buf -> is_full())
			throw TRANSIENT(MinorForcedShutdown,
					COMPLETED_MAYBE);
		}
		else
		{
		    //
		    // Receive body, blocking
		    //
		    transport_ -> receive(buf, true);
		    assert(buf -> is_full());
		}
	    }
	    
	    //
	    // Execute the message
	    //
	    execute();
	}
*/
    }
    catch(const SystemException& ex)
    {
	exception(StateClosed, ex);
	return;
    }

// Temporary
exception(StateClosed,
          TRANSIENT(MinorForcedShutdown, COMPLETED_MAYBE));
}

void
GIOPClientWorker::abortiveShutdown()
{    
    //
    // Send a MessageError message
    //
    try
    {
	OutputStreamImpl out(12);
	
	//
	// Write message header
	//
	Octet flags = 0;
	if(Endian)
	    flags |= 0x01;
	
	static Octet giop[] = { 'G', 'I', 'O', 'P' };
	
	out.write_octet_array(giop, 4); // magic
	out.write_octet(1); // GIOP_version.major
	out.write_octet(0); // GIOP_version.minor
	out.write_octet(flags); // flags
	out.write_octet(6); // message_type (GIOP::MessageError)
	out.write_ulong(0); // message_size
	
	transport_ -> send(out._OB_bufferImpl(), true);
	assert(out._OB_bufferImpl() -> is_full());
    }
    catch(const SystemException& ex)
    {
	exception(StateClosed, ex);
	return;
    }
    
    //
    // If we are in StateError, we don't go through all the hula hoop
    // with continuing to receive messages until the peer
    // closes. Instead, we just close the connection, meaning that we
    // can't be 100% sure that the peer gets the last message.
    //
    exception(StateClosed,
	      TRANSIENT(MinorForcedShutdown, COMPLETED_MAYBE));
}

void
GIOPClientWorker::exception(State state, const CORBA::SystemException& ex)
{
    //
    // Don't do anything if there is no state change
    //
    if(state_ == state)
	return;

    //
    // It is not possible to transition backwards
    //
    if(state < state_)
	return;

    //
    // Update the state
    //
    state_ = state;

    //
    // Set the failure exception
    //
    if(!exception_)
    {
	//
	// Save the exception, with a completion status of COMPLETED_NO
	//
	exception_ = dynamic_cast<SystemException*>(ex._OB_clone());
	exception_ -> completed(COMPLETED_NO);
    }

    switch(state_)
    {
    case StateActive:

	//
	// This operation should not be called for StateActive
	//
	assert(false);
	break;

    case StateClosing:

    	//
	// Detach the worker from the client
	//
	client_ -> removeWorker(this);

        //
	// Do a graceful shutdown
	//
	gracefulShutdown();

	break;

    case StateError:

	//
	// Detach the worker from the starter
	//
	client_ -> removeWorker(this);

        //
	// Do an abortive shutdown
	//
	abortiveShutdown();

	break;


    case StateClosed:

	//
	// Close the transport
	//
#ifndef HAVE_OBE_NO_TRACE
	logCloseConnection();
#endif // !HAVE_OBE_NO_TRACE
	transport_ -> close();

	//
	// Detach the worker from the client
	//
	client_ -> removeWorker(this);

	break;
    }
}

void
GIOPClientWorker::execute(DowncallImpl* down, Octet messageType)
    throw()
{
    assert(state_ == StateActive || state_ == StateClosing);

    //
    // Get the input stream
    //
    InputStreamImpl* in = down -> input();
    in -> _OB_pos(12);

    //
    // Execute message
    //
    switch(messageType)
    {
    case 1: // GIOP::Reply
    {
	ULong status;
	ULong reqId;
	
	try
	{
	    //
	    // Skip service context list
	    //
	    ULong len = in -> read_ulong();
	    for(ULong i = 0 ; i < len ; i++)
	    {
		// context_id
		in -> _OB_skipAlign(4);
		in -> _OB_skip(4);
		// context_data
		ULong datalen = in -> read_ulong();
		in -> _OB_skip(datalen);
	    }

	    reqId = in -> read_ulong(); // request_id
	    status = in -> read_ulong(); // reply_status
	}
	catch(const SystemException& ex)
	{
	    exception(StateError, ex);
	    down -> setFailureException(*exception_, COMPLETED_MAYBE);
	    return;
	}
	
	if(reqId != down -> requestId())
	{ 
	    //
	    // Request id is unknown
	    //
	    exception(StateError,
		      COMM_FAILURE(MinorUnknownReqId, COMPLETED_MAYBE));
	    down -> setFailureException(*exception_, COMPLETED_MAYBE);
	    return;
	}
	
	switch(status)
	{
	case 0: // GIOP::NO_EXCEPTION
	    down -> setNoException();
	    break;
	    
	case 1: // GIOP::USER_EXCEPTION
	    down -> setUserException();
	    break;
	    
	case 2: // GIOP::SYSTEM_EXCEPTION
	{
	    SystemException* ex = SystemException::_OB_unmarshal(in);
	    down -> setSystemException(*ex);
	    delete ex;
	    break;
	}
	
	case 3: // GIOP::LOCATION_FORWARD
	{
	    IOP::IOR ior;
	    IOP::IOR::_OB_unmarshal(ior, in);
	    RefCountIOR_var refIOR = new RefCountIOR(ior);
	    down -> setLocationForward(refIOR);
	    break;
	}
	    
	default:
	    
	    //
	    // Unknown message
	    // TODO: We need a better minor code
	    //
	    exception(StateError,
		      COMM_FAILURE(MinorUnknownMessage, COMPLETED_MAYBE));
	    down -> setFailureException(*exception_, COMPLETED_MAYBE);
	    break;
	}

	break;
    }
	
    case 0: // GIOP::Request
    case 2: // GIOP::CancelRequest:
    case 3: // GIOP::LocateRequest:
	
	//
	// These messages may not be sent to a client
	//
	exception(StateError,
		  COMM_FAILURE(MinorWrongMessage, COMPLETED_MAYBE));
	down -> setFailureException(*exception_, COMPLETED_MAYBE);
	break;
	
    case 4: // GIOP::LocateReply:
	
	//
	// Since I am not sending LocateRequest messages, LocateReply
	// may not be sent to this client
	//
	exception(StateError,
		  COMM_FAILURE(MinorWrongMessage, COMPLETED_MAYBE));
	down -> setFailureException(*exception_, COMPLETED_MAYBE);
	break;
	
    case 5: // GIOP::CloseConnection:

	//
	// If the peer closes the connection, all outstanding requests
	// can safely be reissued. Thus we send all of them a
	// TRANSIENT exception with a completion status of
	// COMPLETED_NO. This is done by calling exception() with the
	// notCompleted parameter set to true.
	//
	exception(StateClosed,
		  TRANSIENT(MinorCloseConnection, COMPLETED_MAYBE));
	down -> setFailureException(*exception_, COMPLETED_NO);
	break;
	
    case 6: // GIOP::MessageError:
	
	//
	// How should we react to MessageError messages? We just send
	// a MessageError back, and close the connection.
	//
	exception(StateError,
		  COMM_FAILURE(MinorMessageError, COMPLETED_MAYBE));
	down -> setFailureException(*exception_, COMPLETED_MAYBE);
	break;
	
    default:
	
	//
        // Unknown message
	//
	exception(StateError,
		  COMM_FAILURE(MinorUnknownMessage, COMPLETED_MAYBE));
	down -> setFailureException(*exception_, COMPLETED_MAYBE);
	break;
    }
}

bool
GIOPClientWorker::sendNoSync(DowncallImpl* down)
    throw()
{
    //
    // If a failure exception is set, don't add new downcall
    // objects. Just set the exception in the downcall.
    //
    if(exception_)
    {
	down -> setFailureException(*exception_, COMPLETED_NO);
	return true;
    }

    //
    // If we are in StateClosing, StateError, or StateClosed, then why
    // is no exception set?
    //
    assert(state_ != StateClosing && state_ != StateError &&
	   state_ != StateClosed);

    //
    // Get the output stream and the buffer
    //
    OutputStreamImpl* out = down -> output();
    OCI::BufferImpl* buf = out -> _OB_bufferImpl();

    //
    // Add the message header
    //
    try
    {
	Octet flags = 0;
	if(Endian)
	    flags |= 0x01;
	
	static Octet giop[] = { 'G', 'I', 'O', 'P' };
	
	ULong pos = out -> _OB_pos();
	out -> _OB_pos(0);
	out -> write_octet_array(giop, 4); // magic
	out -> write_octet(1); // GIOP_version.major
	out -> write_octet(0); // GIOP_version.minor
	out -> write_octet(flags); // flags
	out -> write_octet(0); // message_type (GIOP::Request)
	out -> write_ulong(pos - 12); // message_size
    }
    catch(const SystemException&)
    {
	down -> setFailureException(*exception_, COMPLETED_NO);
	return true;
    }

    //
    // Get the timeout
    //
    Long t = down -> policies_nodup() -> requestTimeout;

    try
    {
	buf -> pos(0);

	if(t < 0)
	{
	    //
	    // Send buffer, blocking
	    //
	    transport_ -> send(buf, true);
	    assert(buf -> is_full());
	}
	else
	{
	    //
	    // Send buffer, with timeout
	    //
	    transport_ -> send_timeout(buf, (ULong)t);
	    if(!buf -> is_full())
		throw NO_RESPONSE(); // Timeout
	}
    }
    catch(const SystemException& ex)
    {
	exception(StateClosed, ex);
	down -> setFailureException(*exception_, COMPLETED_NO);
	return true;
    }
    
    if(down -> responseExpected())
    {
	//
	// Response expected, so we're not finished yet
	//
	down -> setPending();
	return false;
    }
    else
    {
	//
	// No response expected, so we're finished
	//
	down -> setNoException();
	return true;
    }
}

// ----------------------------------------------------------------------
// GIOPClientWorker constructor and destructor
// ----------------------------------------------------------------------

GIOPClientWorker::GIOPClientWorker(ORBInstance_ptr orbInstance,
				   GIOPClient_ptr client,
				   OCI::Transport_ptr transport)
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      client_(GIOPClient::_duplicate(client)),
      transport_(OCI::Transport::_duplicate(transport)),
#ifndef HAVE_OBE_NO_TRACE
      closeLogged_(false),
#endif // !HAVE_OBE_NO_TRACE
      state_(StateActive),
      exception_(0)
{
    //
    // Cache properties
    //
    Properties_var properties = orbInstance_ -> getProperties();
    const char* value;

    value = properties -> getProperty("ooc.orb.client_shutdown_timeout");
    if(value)
	shutdownTimeout_ = atoi(value);
    else
	shutdownTimeout_ = 2;

#ifndef HAVE_OBE_NO_TRACE
    //
    // Trace new outgoing connection
    //
    CoreTraceLevels_var coreTraceLevels = orbInstance_ -> getCoreTraceLevels();
    if(coreTraceLevels -> traceConnections())
    {
	String_var msg = string_dup("new connection\n");
	Logger_var logger = orbInstance_ -> getLogger();
	logger -> trace("outgoing", msg);
    }
#endif // !HAVE_OBE_NO_TRACE
}

GIOPClientWorker::~GIOPClientWorker()
{
    assert(state_ == StateClosed);
    delete exception_;
}

// ----------------------------------------------------------------------
// GIOPClientWorker public member implementations
// ----------------------------------------------------------------------

void
GIOPClientWorker::destroy()
{
    JTCSynchronized synchronized(*this);

    exception(GIOPClientWorker::StateClosing,
	      INITIALIZE(MinorORBDestroyed, COMPLETED_NO));
}

OCI::Transport_ptr
GIOPClientWorker::transport()
{
    return OCI::Transport::_duplicate(transport_);
}

void
GIOPClientWorker::send(DowncallImpl* down)
    throw()
{
    JTCSynchronized synchronized(*this);

    //
    // May only be called if no response is expected
    //
    assert(!down -> responseExpected());

    //
    // Send the downcall
    //
    bool finished = sendNoSync(down);
    assert(finished);
}

void
GIOPClientWorker::sendReceive(DowncallImpl* down)
    throw()
{
    JTCSynchronized synchronized(*this);

    //
    // May only be called if a response is expected
    //
    assert(down -> responseExpected());

    //
    // Send the downcall
    //
    bool finished = sendNoSync(down);
    if(finished)
	return;

    //
    // Get the timeout
    //
    Long t = down -> policies_nodup() -> requestTimeout;

    //
    // Get the input stream stream and the buffer
    //
    InputStreamImpl* in = down -> input();
    OCI::BufferImpl* buf = in -> _OB_bufferImpl();

    try
    {
	buf -> pos(0);

	if(t < 0)
	{
	    //
	    // Receive header, blocking
	    //
	    transport_ -> receive(buf, true);
	    assert(buf -> is_full());
	}
	else
	{
	    //
	    // Receive header, with timeout
	    //
	    transport_ -> receive_timeout(buf, (ULong)t);
	    if(!buf -> is_full())
		throw NO_RESPONSE(); // Timeout
	}
    }
    catch(const SystemException& ex)
    {
	exception(StateClosed, ex);
	down -> setFailureException(*exception_, COMPLETED_MAYBE);
	return;
    }
    
    //
    // Header is complete
    //
    Octet messageType = 255;
    try
    {
	const Octet* data = buf -> data();
	if(data[0] != 'G' || data[1] != 'I' ||
	   data[2] != 'O' || data[3] != 'P')
	{
	    throw COMM_FAILURE("", MinorNoGIOP, COMPLETED_MAYBE);
	}
	
	//
	// Peek at version
	//
	Octet major = data[4];
	Octet minor = data[5];
	
	if(major != 1 || minor != 0)
	    throw COMM_FAILURE("", MinorVersion, COMPLETED_MAYBE);
	
	buf -> pos(6); // Skip magic + GIOP_version
        bool byteOrder = in -> read_boolean();
        in -> _OB_swap(byteOrder != Endian);
        messageType = in -> read_octet();
	ULong size = in -> read_ulong();
	buf -> realloc(12 + size);
    }
    catch(const SystemException& ex)
    {
	exception(StateError, ex);
	down -> setFailureException(*exception_, COMPLETED_MAYBE);
	return;
    }
    
    if(!buf -> is_full())
    {
	try
	{
	    if(t < 0)
	    {
		//
		// Receive body, blocking
		//
		transport_ -> receive(buf, true);
		assert(buf -> is_full());
	    }
	    else
	    {
		//
		// Receive body, with timeout
		//
		transport_ -> receive_timeout(buf, (ULong)t);
		if(!buf -> is_full())
		    throw NO_RESPONSE(); // Timeout
	    }
	}
	catch(const SystemException& ex)
	{
	    exception(StateClosed, ex);
	    down -> setFailureException(*exception_, COMPLETED_MAYBE);
	    return;
	}
    }
    
    //
    // Execute the message
    //
    execute(down, messageType);
}
