// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/OAInterface.h>
#include <OBE/Except.h>
#include <OBE/OCI.h>
#include <OBE/ORBInstance.h>
#include <OBE/Logger.h>
#ifndef HAVE_OBE_NO_TRACE
#  include <OBE/CoreTraceLevels.h>
#endif
#include <OBE/NativeTypes.h>

#include <GIOPServerStarter.h>
#include <GIOPServerWorker.h>
#include <ORBInstanceS.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::GIOPServerStarter_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPServerStarter_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// GIOPServerStarter private and protected member implementation
// ----------------------------------------------------------------------

#ifndef HAVE_OBE_NO_TRACE
void
GIOPServerStarter::logCloseAcceptor()
{
    CoreTraceLevels_var coreTraceLevels =
	orbInstance_ -> getCoreTraceLevels();
    if(coreTraceLevels -> traceConnections())
    {
        String_var msg = string_dup("stopped accepting connections\n");
	Logger_var logger = orbInstance_ -> getLogger();
	logger -> trace("incoming", msg);
    }
}
#endif // !HAVE_OBE_NO_TRACE

// ----------------------------------------------------------------------
// GIOPServerStarter constructor and destructor
// ----------------------------------------------------------------------

GIOPServerStarter::GIOPServerStarter(
    ORBInstance_ptr orbInstance, OCI::Acceptor_ptr acceptor,
    OAInterface_ptr oaInterface)
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      acceptor_(OCI::Acceptor::_duplicate(acceptor)),
      oaInterface_(OAInterface::_duplicate(oaInterface)),
      state_(StateHolding) // Must be holding initially
{
    try
    {
#ifndef HAVE_OBE_NO_TRACE
        //
        // Trace acceptor creation
        //
	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceConnections())
        {
            String_var msg = string_dup("accepting connections\n");
	    Logger_var logger = orbInstance_ -> getLogger();
            logger -> trace("incoming", msg);
        }
#endif // !HAVE_OBE_NO_TRACE

	//
	// Start listening
	//
	acceptor_ -> listen();
    }
    catch(const SystemException&)
    {
	acceptor_ -> close();
	state_ = StateClosed;
        _OB_setRef(0);
        throw;
    }

    try
    {
	//
	// Get the server ORBInstance
	//
	ORBInstanceS* orbInstance =
	    dynamic_cast<ORBInstanceS*>(orbInstance_.in());
	assert(orbInstance);

	//
	// Retrieve the thread group for the servers
	//
	JTCThreadGroupHandle group = orbInstance -> getServerWorkerGroup();

	//
	// Start starter thread
	//
	starterThread_ = new StarterThread(group, this);
	starterThread_ -> start();
    }
    catch(const JTCException&)
    {
	acceptor_ -> close();
	state_ = StateClosed;
	_OB_setRef(0);
	throw IMP_LIMIT(MinorThreadLimit, COMPLETED_NO);
    }
}

GIOPServerStarter::~GIOPServerStarter()
{
    assert(state_ == StateClosed);
}

// ----------------------------------------------------------------------
// GIOPServerStarter public member implementation
// ----------------------------------------------------------------------

void
GIOPServerStarter::removeWorker(GIOPServerWorker_ptr worker)
{
    JTCSynchronized synchronized(*this);

    OB_STL::list<GIOPServerWorker_var>::iterator iter;
    for(iter = workers_.begin() ; iter != workers_.end() ; iter++)
    {
        if((*iter).in() == worker)
	{
	    workers_.erase(iter);
	    return;
	}
    }
}

bool
GIOPServerStarter::setState(State state)
{
    OB_STL::list<GIOPServerWorker_var> workers;

    {
	JTCSynchronized synchronized(*this);

	//
	// Don't do anything if there is no state change
	//
	if(state_ == state)
	    return false;

	//
	// It is not possible to transition backwards, except if we are
	// in holding state
	//
	if(state_ != StateHolding && state < state_)
	    return false;

	switch(state)
	{
	case StateActive:
	case StateHolding:
	    break;

	case StateClosed:

	    //
	    // Connect to this starter's acceptor, to unblock the call
	    // to accept() in the starter thread
	    //
	    try
	    {
		OCI::Transport_var tr = acceptor_ -> connect_self();
		tr -> close();
	    }
	    catch(const SystemException&)
	    {
		//
		// Ignore all system exceptions
		//
	    }

	    break;
	}

	//
	// Update the state and notify about the state change
	//
	state_ = state;
	notifyAll();

	//
	// Use a copy of the workers list, and set the state of the
	// workers outside the synchronization, to avoid deadlocks
	//
	workers = workers_;
    }

    OB_STL::list<GIOPServerWorker_var>::iterator iter;
    for(iter = workers.begin() ; iter != workers.end() ; iter++)
    {
	switch(state_)
	{
	case StateActive:
	    (*iter) -> setState(GIOPServerWorker::StateActive);
	    break;
	    
	case StateHolding:
	    (*iter) -> setState(GIOPServerWorker::StateHolding);
	    break;
	    
	case StateClosed:
	    (*iter) -> setState(GIOPServerWorker::StateClosing);
	    break;
	}
    }

    return true;
}

void
GIOPServerStarter::starterRun()
{
    while(true)
    {
	//
	// Get new transport, blocking
	//
	OCI::Transport_var transport;
	try
	{
	    transport = acceptor_ -> accept(true);
	    assert(!is_nil(transport));
	}
	catch(const NO_PERMISSION&)
	{
	    //
	    // Ignore NO_PERMISSION exceptions
	    //
	}
	catch(const SystemException&)
	{
	    //
	    // Ignore exception. This probably means that the server
	    // exceeded the number available file descriptors.
	    //
	}

	{
	    JTCSynchronized synchronized(*this);

	    //
	    // Check whether we are on hold
	    //
	    while(state_ == StateHolding)
	    {
		try
		{
		    wait();
		}
		catch(const JTCInterruptedException&)
		{
		}
	    }
	    
	    if(!is_nil(transport))
	    {
		try
		{
		    if(state_ == StateActive)
		    {
			//
			// If we're active, we create and add a new
			// worker to the worker list
			//
			GIOPServerWorker_var worker =
			    new GIOPServerWorker(orbInstance_,
							 this,
							 transport,
							 oaInterface_);
			
			workers_.push_back(worker);
			worker -> setState(GIOPServerWorker::StateActive);
		    }
		    else
		    {
			//
			// If we're closed, we create a new dummy
			// worker, only in order to set it to
			// StateClosing for proper connection shutdown
			//
			assert(state_ == StateClosed);
			
			GIOPServerWorker_var worker =
			    new GIOPServerWorker(orbInstance_,
							 _nil(),
							 transport,
							 oaInterface_);
			
			worker -> setState(GIOPServerWorker::StateClosing);
		    }
		}
		catch(const SystemException& ex)
		{
#ifndef HAVE_OBE_NO_TRACE
		    String_var msg = string_dup("can't accept connection\n");
		    String_var exMsg = ex._rep_id();
		    msg += exMsg;
		    Logger_var logger = orbInstance_ -> getLogger();
		    logger -> warning(msg);
#endif
		}
	    }

	    if(state_ == StateClosed)
		break;
	}
    }
}

// ----------------------------------------------------------------------
// GIOPServerStarter::SenderThread public member implementations
// ----------------------------------------------------------------------

void
GIOPServerStarter::StarterThread::run()
{
    try
    {
	starter_ -> starterRun();
    }
    catch(const Exception& ex)
    {
#ifndef HAVE_OBE_NO_TRACE
	Logger_var logger = starter_ -> orbInstance_ -> getLogger();
	logger -> error(ex._rep_id());
#endif
	assert(false);
    }
    catch(...)
    {
	assert(false);
    }

    //
    // Shutdown the acceptor so that no further connections are
    // accepted
    //
#ifndef HAVE_OBE_NO_TRACE
    starter_ -> logCloseAcceptor();
#endif // !HAVE_OBE_NO_TRACE
    starter_ -> acceptor_ -> shutdown();
    
    //
    // Accept all connections which might have queued up in the
    // listen() backlog
    //
    while(true)
    {
	OCI::Transport_var transport;

	try
	{
	    transport = starter_ -> acceptor_ -> accept(false);
	}
	catch(const SystemException&)
	{
	}

	if(is_nil(transport))
	    break;

	try
	{
	    GIOPServerWorker_ptr worker =
		new GIOPServerWorker(starter_ -> orbInstance_,
				     GIOPServerStarter::_nil(),
				     transport,
				     starter_ -> oaInterface_);

	    worker -> setState(GIOPServerWorker::StateClosing);
	}
	catch(const SystemException&)
	{
	    // Ignore SystemExceptions
	}
    }

    //
    // Close the acceptor
    //
    starter_ -> acceptor_ -> close();

    //
    // Break cyclic object dependency
    //
    starter_ = GIOPServerStarter::_nil();
}
