// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/OAInterface.h>
#include <OBE/Stream.h>
#include <OBE/Logger.h>
#include <OBE/OCI.h>
#include <OBE/RefCounted.h>
#include <OBE/ORBInstance.h>
#include <OBE/Logger.h>
#ifndef HAVE_OBE_NO_TRACE
#  include <OBE/CoreTraceLevels.h>
#endif // !HAVE_OBE_NO_TRACE
#include <OBE/Properties.h>
#include <OBE/NativeTypes.h>

#include <GIOPServerStarter.h>
#include <GIOPServerWorker.h>
#include <ORBInstanceS.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::GIOPServerWorker_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::GIOPServerWorker_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// GIOPServerWorker private and protected member implementations
// ----------------------------------------------------------------------

#ifndef HAVE_OBE_NO_TRACE
void
GIOPServerWorker::logCloseConnection()
{
    if(!closeLogged_)
    {
	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceConnections())
	{
	    String_var msg = string_dup("closing connection\n");
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("incoming", msg);
	}
	
	closeLogged_ = true;
    }
}
#endif // !HAVE_OBE_NO_TRACE

void
GIOPServerWorker::gracefulShutdown()
{
    //
    // Don't shut down if there are still outstanding upcalls
    //
    if(upcallCount_ > 0)
	return;

    try
    {
	//
	// Get the server ORBInstance
	//
	ORBInstanceS* orbInstance =
	    dynamic_cast<ORBInstanceS*>(orbInstance_.in());
	assert(orbInstance);

	//
	// Retrieve the thread group for the servers
	//
	JTCThreadGroupHandle group = orbInstance -> getServerWorkerGroup();

	//
	// Start the shutdown thread, which will take care of proper
	// connection termination
	//
	shutdownThread_ = new ShutdownThread(group, this);
	shutdownThread_ -> start();
    }
    catch(const JTCException&)
    {	
	exception(StateClosed, IMP_LIMIT(MinorThreadLimit, COMPLETED_NO));
	return;
    }
}

void
GIOPServerWorker::abortiveShutdown()
{    
    //
    // Send a MessageError message
    //
    try
    {
	OutputStreamImpl out(12);
	
	//
	// Write message header
	//
	Octet flags = 0;
	if(Endian)
	    flags |= 0x01;
	
	static Octet giop[] = { 'G', 'I', 'O', 'P' };
	
	out.write_octet_array(giop, 4); // magic
	out.write_octet(1); // GIOP_version.major
	out.write_octet(0); // GIOP_version.minor
	out.write_octet(flags); // flags
	out.write_octet(6); // message_type (GIOP::MessageError)
	out.write_ulong(0); // message_size
	
	out._OB_bufferImpl() -> pos(0);

	if(!transport_ -> send_detect(out._OB_bufferImpl(), true))
	{
	    setStateNoSync(StateClosed);
	    return; // Orderly shutdown
	}
	assert(out._OB_bufferImpl() -> is_full());
    }
    catch(const SystemException& ex)
    {
	exception(StateClosed, ex);
	return;
    }
    
    //
    // If we are in StateError, we don't go through all the hula hoop
    // with continuing to receive messages until the peer
    // closes. Instead, we just close the connection, meaning that we
    // can't be 100% sure that the peer gets the last message.
    //
    setStateNoSync(StateClosed);	
}

void
GIOPServerWorker::exception(State state, const SystemException& ex)
{
#ifndef HAVE_OBE_NO_TRACE
    String_var msg = string_dup("communications error with client\n");
    String_var exMsg = ex._rep_id();
    msg += exMsg;
    Logger_var logger = orbInstance_ -> getLogger();
    logger -> warning(msg);
#endif

    setStateNoSync(state);
}

void
GIOPServerWorker::setStateNoSync(State state)
{
    //
    // Don't do anything if there is no state change
    //
    if(state_ == state)
	return;

    //
    // It is not possible to transition backwards, except if we are
    // in holding state
    //
    if(state_ != StateHolding && state < state_)
	return;

    //
    // Update the state
    //
    state_ = state;

    switch(state_)
    {
    case StateActive:
    case StateHolding:

	//
	// There is nothing special to do
	//
	break;

    case StateClosing:

	//
	// Detach the worker from the starter
	//
 	if(!is_nil(starter_))
	    starter_ -> removeWorker(this);

        //
	// Do a graceful shutdown
	//
	gracefulShutdown();

	break;

    case StateError:

	//
	// Detach the worker from the starter
	//
 	if(!is_nil(starter_))
	    starter_ -> removeWorker(this);

        //
	// Do an abortive shutdown
	//
	abortiveShutdown();

	break;

    case StateClosed:

	//
	// Detach the worker from the starter
	//
 	if(!is_nil(starter_))
	    starter_ -> removeWorker(this);

	//
	// Close the transport
	//
#ifndef HAVE_OBE_NO_TRACE
	logCloseConnection();
#endif // !HAVE_OBE_NO_TRACE
	transport_ -> close();

	break;
    }
}

void
GIOPServerWorker::execute(UpcallImpl* upcall, Octet messageType)
    throw()
{
    assert(state_ == StateActive || state_ == StateClosing);

    //
    // Execute message
    //
    switch(messageType)
    {
    case 0: // GIOP::Request

	//
	// Only execute the request if we are in StateActive. If we
	// are in StateClosing, then the server signals the client to
	// reissue the request by sending a CloseConnection message.
	//
	if(state_ == StateActive)
	    executeRequest(upcall);
	break;
    
    case 2: // GIOP::CancelRequest
	
	//
	// Do nothing here (doing nothing is GIOP compliant)
	//
	break;
	
    case 3: // GIOP::LocateRequest

	//
	// Only execute the locate request if we are in StateActive
	//
	if(state_ == StateActive)
	    executeLocate(upcall);
        break;

    case 1: // GIOP::Reply
    case 4: // GIOP::LocateReply

	//
	// These messages may not be sent to a server
	//
	exception(StateError,
		  COMM_FAILURE(MinorWrongMessage, COMPLETED_MAYBE));
	break;

    case 5: // GIOP::CloseConnection
	
	//
	// This message may not be sent to a server in GIOP 1.0 or 1.1
	//
	exception(StateClosed,
		  COMM_FAILURE(MinorWrongMessage, COMPLETED_MAYBE));
        break;
	
    case 6: // GIOP::MessageError

	//
	// How should we react to MessageError messages? We just send
	// a MessageError back, and close the connection.
	//
	exception(StateError,
		  COMM_FAILURE(MinorMessageError, COMPLETED_MAYBE));
	break;

    default:
	
	//
        // Unknown message
	//
	exception(StateError,
		  COMM_FAILURE(MinorUnknownMessage, COMPLETED_MAYBE));
	break;
    }
}

void
GIOPServerWorker::executeRequest(UpcallImpl* upcall)
    throw()
{
    assert(state_ == StateActive);

    //
    // Get the input stream
    //
    InputStreamImpl* in = upcall -> input();
    in -> _OB_pos(12);

    ULong reqId;
    bool response;
    OCI::ObjectKey key;
    const char* op;
	
    try
    {
	ULong len;

	//
	// Skip service context list
	//
	len = in -> read_ulong();
	for(ULong i = 0 ; i < len ; i++)
	{
	    // context_id
	    in -> _OB_skipAlign(4);
	    in -> _OB_skip(4);
	    // context_data
	    ULong datalen = in -> read_ulong();
	    in -> _OB_skip(datalen);
	}
	
        reqId = in -> read_ulong(); // request_id
        response = in -> read_boolean(); // response_expected

        //
        // object_key
        //
        len = in -> read_ulong();
	CORBA::Octet* keyBuf;
	in -> read_octet_array_nodup(keyBuf, len);
	key.replace(len, len, keyBuf);

        //
	// operation. This is a zero-copy read.
        //
	op = in -> read_string_nodup();

        len = in -> read_ulong(); // requesting_principal
        if(len)
            in -> _OB_skip(len);
    }
    catch(const SystemException& ex)
    {
	exception(StateError, ex);
	return;
    }
    
    //
    // New upcall will be started
    //
    if(response)
	upcallCount_++;
    
    //
    // Setup upcall and invoke. Note that the Upcall object keeps a
    // pointer to the OCI::ObjectKey.
    //
    upcall -> setup(response ? this : 0, &key, reqId, op);
    if(oaInterface_ -> setupUpcall(upcall))
	upcall -> invoke();
}

void
GIOPServerWorker::executeLocate(UpcallImpl* upcall)
    throw()
{
    assert(state_ == StateActive);

    ULong reqId;
    OAInterface::LocateStatus status;
    RefCountIOR_var ior;

    try
    {
	//
	// Get the input stream
	//
	InputStreamImpl* in = upcall -> input();
	in -> _OB_pos(12);

        reqId = in -> read_ulong(); // request_id

        //
        // object_key
        //
        ULong len = in -> read_ulong();
	CORBA::Octet* keyBuf;
        in -> read_octet_array_nodup(keyBuf, len);
	OCI::ObjectKey key(len, len, keyBuf);

	//
	// Find the IOR for the key
	//
	status = oaInterface_ -> findByKey(key, ior);
    }
    catch(const SystemException& ex)
    {
	exception(StateError, ex);
	return;
    }
    
    try
    {
	//
	// Get the output stream
	//
	OutputStreamImpl* out = upcall -> output();
	out -> _OB_pos(12);
	
	//
	// Write the LocateReplyHeader
	//
	out -> write_ulong(reqId); // request_id
	out -> write_ulong((ULong)status); // locate_status

	//
	// Write the message body
	//
	if(status == OAInterface::OBJECT_FORWARD)
            ior -> value._OB_marshal(out);

	//
	// Add the message header
	//
	Octet flags = 0;
	if(Endian)
	    flags |= 0x01;
	
	static Octet giop[] = { 'G', 'I', 'O', 'P' };
	
	ULong pos = out -> _OB_pos();
	out -> _OB_pos(0);
	out -> write_octet_array(giop, 4); // magic
	out -> write_octet(1); // GIOP_version.major
	out -> write_octet(0); // GIOP_version.minor
	out -> write_octet(flags); // flags
	out -> write_octet(4); // message_type (GIOP::LocateReply)
	out -> write_ulong(pos - 12); // message_size

	//
	// Send buffer, blocking, detect connection loss
	//
	OCI::BufferImpl* buf = out -> _OB_bufferImpl();
	buf -> pos(0);
	if(!transport_ -> send_detect(buf, true))
	{
	    setStateNoSync(StateClosed);
	    return; // Orderly shutdown
	}
	assert(buf -> is_full());
    }
    catch(const SystemException& ex)
    {
	exception(StateClosed, ex);
	return;
    }
}

// ----------------------------------------------------------------------
// GIOPServerWorker constructor and destructor
// ----------------------------------------------------------------------

GIOPServerWorker::GIOPServerWorker(ORBInstance_ptr orbInstance,
				   GIOPServerStarter_ptr starter,
				   OCI::Transport_ptr transport,
				   OAInterface_ptr oaInterface)
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      starter_(GIOPServerStarter::_duplicate(starter)),
      transport_(OCI::Transport::_duplicate(transport)),
      oaInterface_(OAInterface::_duplicate(oaInterface)),
      upcallCount_(0),
#ifndef HAVE_OBE_NO_TRACE
      closeLogged_(false),
#endif // !HAVE_OBE_NO_TRACE
      state_(StateHolding) // Must be holding initially
{
    //
    // Cache properties
    //
    Properties_var properties = orbInstance_ -> getProperties();
    const char* value;

    value = properties -> getProperty("ooc.orb.server_shutdown_timeout");
    if(value)
	shutdownTimeout_ = atoi(value);
    else
	shutdownTimeout_ = 2;

#ifndef HAVE_OBE_NO_TRACE
    //
    // Trace new incoming connection
    //
    CoreTraceLevels_var coreTraceLevels = orbInstance_ -> getCoreTraceLevels();
    if(coreTraceLevels -> traceConnections())
    {
	String_var msg = string_dup("new connection\n");
	Logger_var logger = orbInstance_ -> getLogger();
	logger -> trace("incoming", msg);
    }
#endif // !HAVE_OBE_NO_TRACE

    try
    {
	//
	// Get the server ORBInstance
	//
	ORBInstanceS* orbInstance =
	    dynamic_cast<ORBInstanceS*>(orbInstance_.in());
	assert(orbInstance);

	//
	// Retrieve the thread group for the servers
	//
	JTCThreadGroupHandle group = orbInstance -> getServerWorkerGroup();

	//
	// Start receiver thread
	//
	receiverThread_ = new ReceiverThread(group, this);
	receiverThread_ -> start();
    }
    catch(const JTCException&)
    {
	transport_ -> close();
	state_ = StateClosed;
	_OB_setRef(0);
	throw IMP_LIMIT(MinorThreadLimit, COMPLETED_NO);
    }
}

GIOPServerWorker::~GIOPServerWorker()
{
    assert(state_ == StateClosed);
}

// ----------------------------------------------------------------------
// GIOPServerWorker public member implementations
// ----------------------------------------------------------------------

void
GIOPServerWorker::setState(State state)
{
    JTCSynchronized synchronized(*this);
    setStateNoSync(state);
}

void
GIOPServerWorker::upcallBeginReply(UpcallImpl* upcall)
{
    OutputStreamImpl* out = upcall -> output();
    out -> _OB_pos(12);

    try
    {
        out -> write_ulong(0); // service_context
        out -> write_ulong(upcall -> requestId()); // request_id
        out -> write_ulong(0); // reply_status (GIOP::NO_EXCEPTION)
    }
    catch(const SystemException&)
    {
	//
	// Nothing may go wrong here, otherwise we might have a
	// recursion
	//
	assert(false);
    }
}

void
GIOPServerWorker::upcallEndReply(UpcallImpl* upcall)
{
    //
    // If we are in StateError or StateClosed, don't send a reply for
    // the upcall
    //
    if(state_ == StateError || state_ == StateClosed)
	return;

    OutputStreamImpl* out = upcall -> output();
    
    //
    // Add the message header
    //
    try
    {
	Octet flags = 0;
	if(Endian)
	    flags |= 0x01;
	
	static Octet giop[] = { 'G', 'I', 'O', 'P' };
	
	ULong pos = out -> _OB_pos();
	out -> _OB_pos(0);
	out -> write_octet_array(giop, 4); // magic
	out -> write_octet(1); // GIOP_version.major
	out -> write_octet(0); // GIOP_version.minor
	out -> write_octet(flags); // flags
	out -> write_octet(1); // message_type (GIOP::Reply)
	out -> write_ulong(pos - 12); // message_size
    }
    catch(const SystemException&)
    {
	//
	// Nothing may go wrong here, otherwise we might have a
	// recursion
	//
	assert(false);
    }
    
    //
    // Send buffer, blocking, detect connection loss
    //
    OCI::BufferImpl* buf = out -> _OB_bufferImpl();
    buf -> pos(0);
    try
    {
	if(!transport_ -> send_detect(buf, true))
	{
	    setStateNoSync(StateClosed);
	    return; // Orderly shutdown
	}
	assert(buf -> is_full());
    }
    catch(const SystemException& ex)
    {
	exception(StateClosed, ex);
	return;
    }

    //
    // Upcall has finished
    //
    assert(upcallCount_ > 0);
    upcallCount_--;
    
    //
    // If this was the last upcall and we are in StateClosing, we now
    // gracefully shut down the connection
    //
    if(upcallCount_ == 0 && state_ == StateClosing)
	gracefulShutdown();
}

void
GIOPServerWorker::upcallUserException(UpcallImpl* upcall,
                                      const UserException& ex)
{
    OutputStreamImpl* out = upcall -> output();
    out -> _OB_pos(12);

    try
    {
	out -> write_ulong(0); // service_context
        out -> write_ulong(upcall -> requestId()); // request_id
        out -> write_ulong(1); // reply_status (GIOP::USER_EXCEPTION)
        ex._OB_marshal(out);
    }
    catch(const SystemException&)
    {
	//
	// Nothing may go wrong here, otherwise we might have a
	// recursion
	//
	assert(false);
    }

    upcallEndReply(upcall);
}

void
GIOPServerWorker::upcallSystemException(UpcallImpl* upcall,
					const SystemException& ex)
{
    OutputStreamImpl* out = upcall -> output();
    out -> _OB_pos(12);

    try
    {
	out -> write_ulong(0); // service_context
        out -> write_ulong(upcall -> requestId()); // request_id
        out -> write_ulong(2); // reply_status (GIOP::SYSTEM_EXCEPTION)
        ex._OB_marshal(out);
    }
    catch(const SystemException&)
    {
	//
	// Nothing may go wrong here, otherwise we might have a
	// recursion
	//
	assert(false);
    }

    upcallEndReply(upcall);
}

void
GIOPServerWorker::upcallForward(UpcallImpl* upcall, RefCountIOR_ptr ior)
{
    OutputStreamImpl* out = upcall -> output();
    out -> _OB_pos(12);

    try
    {
	out -> write_ulong(0); // service_context
        out -> write_ulong(upcall -> requestId()); // request_id
        out -> write_ulong(3); // reply_status (GIOP::LOCATION_FORWARD)
	ior -> value._OB_marshal(out);
    }
    catch(const SystemException&)
    {
	//
	// Nothing may go wrong here, otherwise we might have a
	// recursion
	//
	assert(false);
    }

    upcallEndReply(upcall);
}

void
GIOPServerWorker::receiverRun()
{
    while(true)
    {
	UpcallImpl upcall(orbInstance_);
	
	//
	// Get the input stream stream and the buffer
	//
	InputStreamImpl* in = upcall.input();
	OCI::BufferImpl* buf = in -> _OB_bufferImpl();
	
	//
	// Receive header, blocking, detect connection loss
	//
	try
	{
	    buf -> pos(0);

	    if(!transport_ -> receive_detect(buf, true))
	    {
		JTCSynchronized synchronized(*this);
		setStateNoSync(StateClosed);
		return; // Orderly shutdown
	    }
	    assert(buf -> is_full());
	}
	catch(const SystemException& ex)
	{
	    JTCSynchronized synchronized(*this);
	    exception(StateClosed, ex);
	    return;
	}
    
	//
	// Header is complete
	//
	Octet messageType = 255;
	try
	{
	    const Octet* data = buf -> data();
	    if(data[0] != 'G' || data[1] != 'I' ||
	       data[2] != 'O' || data[3] != 'P')
	    {
		throw COMM_FAILURE("", MinorNoGIOP, COMPLETED_MAYBE);
	    }
	    
	    //
	    // Peek at version
	    //
	    Octet major = data[4];
	    Octet minor = data[5];
	    
	    if(major != 1 || minor != 0)
		throw COMM_FAILURE("", MinorVersion, COMPLETED_MAYBE);
	    
	    buf -> pos(6); // Skip magic + GIOP_version
	    bool byteOrder = in -> read_boolean();
	    in -> _OB_swap(byteOrder != Endian);
	    messageType = in -> read_octet();
	    ULong size = in -> read_ulong();
	    buf -> realloc(12 + size);
	}
	catch(const SystemException& ex)
	{
	    JTCSynchronized synchronized(*this);
	    exception(StateError, ex);
	    return;
	}
    
	if(!buf -> is_full())
	{
	    //
	    // Receive body, blocking, detect connection loss
	    //
	    try
	    {
		if(!transport_ -> receive_detect(buf, true))
		{
		    JTCSynchronized synchronized(*this);
		    setStateNoSync(StateClosed);
		    return; // Orderly shutdown
		}
		assert(buf -> is_full());
	    }
	    catch(const SystemException& ex)
	    {
		JTCSynchronized synchronized(*this);
		exception(StateClosed, ex);
		return;
	    }
	}
	    
	{
	    JTCSynchronized synchronized(*this);
	    
	    //
	    // If we are not in StateActive or StateClosing, stop this
	    // thread. We do *not* stop this thread if we are in
	    // StateClosing, since we must continue to read data from
	    // the Transport to make sure that no messages can get
	    // lost upon close, and to make sure that CloseConnection
	    // messages from the peer are processed.
	    //
	    if(state_ != StateActive && state_ != StateClosing)
		return;

	    //
	    // Execute the message
	    //
	    execute(&upcall, messageType);
	}
    }
}

void
GIOPServerWorker::shutdownRun()
{
    //
    // Send a CloseConnection message
    //
    try
    {
	OutputStreamImpl out(12);
	
	//
	// Write message header
	//
	Octet flags = 0;
	if(Endian)
	    flags |= 0x01;
	
	static Octet giop[] = { 'G', 'I', 'O', 'P' };
	
	out.write_octet_array(giop, 4); // magic
	out.write_octet(1); // GIOP_version.major
	out.write_octet(0); // GIOP_version.minor
	out.write_octet(flags); // flags
	out.write_octet(5); // message_type (GIOP::CloseConnection)
	out.write_ulong(0); // message_size
	
	out._OB_bufferImpl() -> pos(0);

	if(!transport_ -> send_detect(out._OB_bufferImpl(), true))
	{
	    JTCSynchronized synchronized(*this);
	    setStateNoSync(StateClosed);
	    return; // Orderly shutdown
	}
	assert(out._OB_bufferImpl() -> is_full());
    }
    catch(const SystemException& ex)
    {
	exception(StateClosed, ex);
	return;
    }
    
    //
    // Now we gracefully shutdown the connection
    //
#ifndef HAVE_OBE_NO_TRACE
    logCloseConnection();
#endif // !HAVE_OBE_NO_TRACE
    transport_ -> shutdown();
    
    while(true)
    {
	try
	{
	    if(shutdownTimeout_ > 0)
	    {
		//
		// Join with the receiver thread, with timeout
		//
		receiverThread_ -> join(shutdownTimeout_ * 1000);
	    }
	    else
	    {
		//
		// Join with the receiver thread, no timeout
		//
		receiverThread_ -> join();
	    }
	    
	    break;
	}
	catch(const JTCInterruptedException&)
	{
	    continue;
	}
    }
    
    //
    // We now close the connection actively, since it may still be
    // open under certain circumstances. (For example, if the receiver
    // thread has not terminated yet, or if the receiver thread sets
    // the state to StateError before termination.)
    //
    {
	JTCSynchronized synchronized(*this);
	setStateNoSync(StateClosed);
    }
}

// ----------------------------------------------------------------------
// GIOPServerWorker::ReceiverThread public member implementations
// ----------------------------------------------------------------------

void
GIOPServerWorker::ReceiverThread::run()
{
    try
    {
	worker_ -> receiverRun();
    }
    catch(const Exception& ex)
    {
#ifndef HAVE_OBE_NO_TRACE
	Logger_var logger = worker_ -> orbInstance_ -> getLogger();
	logger -> error(ex._rep_id());
#endif
	assert(false);
    }
    catch(const JTCException& ex)
    {
#ifndef HAVE_OBE_NO_TRACE
	Logger_var logger = worker_ -> orbInstance_ -> getLogger();
	String_var err(ex.getType());
	err += (const char*)": ";
	err += ex.getMessage();
	logger -> error(err);
#endif
	assert(false);
    }
    catch(...)
    {
	assert(false);
    }

    //
    // Break cyclic object dependency
    //
    worker_ = GIOPServerWorker::_nil();
}

// ----------------------------------------------------------------------
// GIOPServerWorker::ShutdownThread public member implementations
// ----------------------------------------------------------------------

void
GIOPServerWorker::ShutdownThread::run()
{
    try
    {
	worker_ -> shutdownRun();
    }
    catch(const Exception& ex)
    {
#ifndef HAVE_OBE_NO_TRACE
	Logger_var logger = worker_ -> orbInstance_ -> getLogger();
	logger -> error(ex._rep_id());
#endif
	assert(false);
    }
    catch(const JTCException& ex)
    {
#ifndef HAVE_OBE_NO_TRACE
	Logger_var logger = worker_ -> orbInstance_ -> getLogger();
	String_var err(ex.getType());
	err += (const char*)": ";
	err += ex.getMessage();
	logger -> error(err);
#endif
	assert(false);
    }
    catch(...)
    {
	assert(false);
    }

    //
    // Break cyclic object dependency
    //
    worker_ = GIOPServerWorker::_nil();
}
