// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_GIOP_SERVER_WORKER_H
#define OB_GIOP_SERVER_WORKER_H

#include <OBE/Server_fwd.h>
#include <OBE/OAInterface_fwd.h>
#include <OBE/OCI_fwd.h>
#include <OBE/ORBInstance_fwd.h>

#include <GIOPServerStarter_fwd.h>
#include <GIOPServerWorker_fwd.h>

#include <OBE/Upcall.h>
#include <OBE/list.h>

namespace OB
{

//
// The GIOPServerWorker class
//
class GIOPServerWorker : public RefCount, 
		         public UpcallReturn,
		         public JTCRecursiveMutex
{
    //
    // Hide copy-constructor and assignment operator
    //
    GIOPServerWorker(const GIOPServerWorker&);
    void operator=(const GIOPServerWorker&);

protected:

    ORBInstance_var orbInstance_; // The ORBInstance
    GIOPServerStarter_var starter_; // The server starter
    OCI::Transport_var transport_; // The transport
    OAInterface_var oaInterface_; // The object adapter interface
    CORBA::ULong upcallCount_; // Number of outstanding upcalls
#ifndef HAVE_OBE_NO_TRACE
    bool closeLogged_; // True if connection closure was already logged
#endif // !HAVE_OBE_NO_TRACE
    CORBA::ULong shutdownTimeout_; // The shutdown timeout

    //
    // The worker's state
    //
public:
    enum State
    {
	StateActive,
	StateHolding,
	StateClosing,
	StateError,
	StateClosed
    };
protected:
    State state_;

    //
    // The receiver thread
    //
    class ReceiverThread : public JTCThread
    {
	GIOPServerWorker_var worker_;

    public:
	
	ReceiverThread(JTCThreadGroupHandle& group,
		       GIOPServerWorker_ptr worker)
	    : JTCThread(group),
	      worker_(GIOPServerWorker::_duplicate(worker)) { }
	virtual ~ReceiverThread() { }
	virtual void run();
    };
    friend class ReceiverThread;
    JTCThreadHandle receiverThread_;

    //
    // The shutdown thread
    //
    class ShutdownThread : public JTCThread
    {
	GIOPServerWorker_var worker_;

    public:
	
	ShutdownThread(JTCThreadGroupHandle& group,
		       GIOPServerWorker_ptr worker)
	    : JTCThread(group),
	      worker_(GIOPServerWorker::_duplicate(worker)) { }
	virtual ~ShutdownThread() { }
	virtual void run();
    };
    friend class ShutdownThread;
    JTCThreadHandle shutdownThread_;

#ifndef HAVE_OBE_NO_TRACE
    //
    // Emit a trace message when closing the connection
    //
    void logCloseConnection();
#endif // !HAVE_OBE_NO_TRACE

    //
    // Shut down the connection, graceful and abortive
    //
    void gracefulShutdown();
    void abortiveShutdown();

    //
    // Change the state of the worker due to an exception
    //
    void exception(State, const CORBA::SystemException&);

    //
    // Change the state of the worker without synchronization
    //
    void setStateNoSync(State);

    //
    // Execute incoming message
    //
    void execute(UpcallImpl*, CORBA::Octet)
	throw();

    //
    // Execute Request message
    //
    void executeRequest(UpcallImpl*)
	throw();

    //
    // Execute LocateRequest message
    //
    void executeLocate(UpcallImpl*)
	throw();

public:

    GIOPServerWorker(ORBInstance_ptr, GIOPServerStarter_ptr,
		     OCI::Transport_ptr, OAInterface_ptr);
    virtual ~GIOPServerWorker();

    static inline GIOPServerWorker_ptr _duplicate(GIOPServerWorker_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline GIOPServerWorker_ptr _nil()
    { return 0; }

    //
    // Change the state of the worker
    //
    void setState(State);

    //
    // Called upon return from an upcall
    //
    virtual void upcallBeginReply(UpcallImpl*);
    virtual void upcallEndReply(UpcallImpl*);
    virtual void upcallUserException(UpcallImpl*,
				     const CORBA::UserException&);
    virtual void upcallSystemException(UpcallImpl*,
				       const CORBA::SystemException&);
    virtual void upcallForward(UpcallImpl*, RefCountIOR_ptr);

    //
    // Run method for receiver and shutdown thread
    //
    void receiverRun();
    void shutdownRun();
};

} // End of namespace OB

namespace CORBA
{

inline void
release(OB::GIOPServerWorker_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::GIOPServerWorker_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

#endif
