// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/InitialServiceManager.h>
#include <OBE/ORBInstance.h>
#include <OBE/Properties.h>
#include <OBE/RefCounted.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::InitialServiceManager_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(OB::InitialServiceManager_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// InitialServiceManager constructor and destructor
// ----------------------------------------------------------------------

InitialServiceManager::InitialServiceManager(ORBInstance_ptr instance)
    : services_(new ServiceMap(37)),
      destroy_(false),
      orbInstance_(ORBInstance::_duplicate(instance))
{

    try
    {
	//
	// Populate the services map
	//
	Properties_var properties = orbInstance_ -> getProperties();

	//
	// Obtain the INS default initial reference URL
	//
	const char* value =
	    properties -> getProperty("ooc.orb.default_init_ref");
	if(value == 0)
	    defaultInitRef_ = string_dup("");
	else
	    defaultInitRef_ = value;

	//
	// Add those services configured in the "ooc.orb.service" property
	//
	Properties::KeySeq serviceKeys =
	    properties -> getKeys("ooc.orb.service.");
	for(ULong i = 0 ; i < serviceKeys.length() ; i++)
	{
	    const char* key = serviceKeys[i];
	    const char* value = properties -> getProperty(key);
	    assert(value != 0);
	    assert(strncmp(key, "ooc.orb.service.",
			   strlen("ooc.orb.service.")) == 0);
	    key += strlen("ooc.orb.service.");
	    addInitialReference(key, value, true);
	}
    }
    catch(...)
    {
	_OB_setRef(0);
	throw;
    }
}

InitialServiceManager::~InitialServiceManager()
{
    assert(destroy_);
}

// ----------------------------------------------------------------------
// InitialServiceManager public member functions
// ----------------------------------------------------------------------

void
InitialServiceManager::destroy()
{
    JTCSynchronized synchronized(*this);

    assert(!destroy_); // May only be destroyed once
    destroy_ = true;

    delete services_;
    orbInstance_ = ORBInstance::_nil();
}

Object_ptr
InitialServiceManager::resolveInitialReferences(const char* identifier)
    throw(SystemException, ORB::InvalidName)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
        throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    assert_nca(identifier != 0, NCANullString);

    ObjectFactory_var objectFactory = orbInstance_ -> getObjectFactory();

    Object_var obj;

    //
    // Search the list of initial references
    //
    Service svc;
    if(services_ -> get(identifier, svc))
    {
        if(!is_nil(svc.obj))
            obj = Object::_duplicate(svc.obj);
        else if(strlen(svc.ref.in()) > 0)
        {
            obj = objectFactory -> stringToObject(svc.ref.in());
            svc.obj = Object::_duplicate(obj);
            services_ -> put(identifier, svc);
        }
    }

    //
    // If no match was found, and there's a default initial
    // reference "template", then try to compose a URL using
    // the identifier as the object-key
    //
    if(is_nil(obj) && strlen(defaultInitRef_) > 0)
    {
        String_var url = defaultInitRef_;
        url += '/';
        url += identifier;
        obj = objectFactory -> stringToObject(url);
    }

    if(is_nil(obj))
        throw ORB::InvalidName();

    //
    // If the object is a l-c object, return the object now
    //
    if(obj -> _OB_isLC())
        return obj._retn();

    //
    // If the object is remote, return a new reference with the
    // current set of policies applied
    //
    RefCountPolicyList_var policies = objectFactory -> policies();
    return obj -> _set_policy_overrides(policies -> value, SET_OVERRIDE);
}

void
InitialServiceManager::addInitialReference(const char* name, Object_ptr obj)
    throw(SystemException, ORB::InvalidName)
{
    addInitialReference(name, obj, false);
}

void
InitialServiceManager::addInitialReference(const char* name,
					   const char* iorString,
					   bool override)
    throw(SystemException, ORB::InvalidName)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
        throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    assert_nca(name != 0, NCANullString);
    assert_nca(iorString != 0, NCANullString);

    if(services_ -> containsKey(name) && !override)
	throw ORB::InvalidName();

    Service svc;
    svc.ref = iorString;
    services_ -> put(name, svc);
}

void
InitialServiceManager::addInitialReference(const char* name,
					   Object_ptr p,
					   bool override)
    throw(SystemException, ORB::InvalidName)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
        throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    assert_nca(name != 0, NCANullString);

    if(services_ -> containsKey(name) && !override)
	throw ORB::InvalidName();

    Service svc;
    svc.ref = string_dup("");
    svc.obj = Object::_duplicate(p);
    services_ -> put(name, svc);
}
