// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/MarshalStubImpl.h>
#include <OBE/OBPolicies.h>
#include <OBE/Downcall.h>
#include <OBE/Client.h>
#include <OBE/Stream.h>
#include <OBE/ORBInstance.h>
#include <OBE/RefCounted.h>
#include <OBE/Logger.h>
#ifndef HAVE_OBE_NO_TRACE
#  include <OBE/CoreTraceLevels.h>
#endif // !HAVE_OBE_NO_TRACE

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::MarshalStubImpl_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::MarshalStubImpl_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// MarshalStubImpl private and protected member implementation
// ----------------------------------------------------------------------

void
MarshalStubImpl::_OB_getClientProfilePair_nodup(
    Client_ptr& client, RefCountProfileInfo_ptr& profile)
{
    //
    // Lazy initialization of the client/profile pairs
    //
    if(!clientProfilePairs_)
    {
	//
	// Get all clients that can be used
	//
	ClientManager_var clientManager =
	    _ob_orbInstance_ -> getClientManager();
	clientProfilePairs_ =
	    clientManager -> getClientProfilePairs(IOR_ -> value,
                                                   _ob_policies_ -> value);
    }

    //
    // If we can't get any client/profile pairs, set and raise the
    // failure exception, and let the stub handle this.
    //
    if(clientProfilePairs_ -> size() == 0)
    {
#ifndef HAVE_OBE_NO_TRACE
	CoreTraceLevels_var coreTraceLevels =
	    _ob_orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceRetry() >= 2)
	{
	    Logger_var logger = _ob_orbInstance_ -> getLogger();
	    logger -> trace("retry", "no profiles available");
	}
#endif // !HAVE_OBE_NO_TRACE

	throw FailureException(NO_RESOURCES(0, COMPLETED_NO));
    }

    client = clientProfilePairs_ -> front() -> client.in();
    profile = clientProfilePairs_ -> front() -> profile.in();
}

void
MarshalStubImpl::_OB_handleFailureException(DowncallImpl& down,
					    const FailureException& ex)
{
    //
    // Only called if there is really a failure
    //
    assert(ex.exception);

    //
    // If there was a failure, release the client and remove the
    // faulty client/profile pair, whether we retry or not
    //
    Client_ptr client = down.client_nodup();
    RefCountProfileInfo_ptr profile = down.profileInfo_nodup();

    OB_STL::list<ClientProfilePair_var>::iterator iter;
    for(iter = clientProfilePairs_ -> begin();
	iter != clientProfilePairs_ -> end();
	iter++)
    {
        if((*iter) -> client.in() == client &&
	   (*iter) -> profile.in() == profile)
        {
	    ClientManager_var clientManager =
		_ob_orbInstance_ -> getClientManager();
	    clientManager -> releaseClient((*iter) -> client);
	    clientProfilePairs_ -> erase(iter);
	    break;
        }
    }

    //
    // We only retry upon COMM_FAILURE, TRANSIENT, and NO_RESPONSE
    //
    if(dynamic_cast<COMM_FAILURE*>(ex.exception) == 0 &&
       dynamic_cast<TRANSIENT*>(ex.exception) == 0 &&
       dynamic_cast<NO_RESPONSE*>(ex.exception) == 0)
	ex._raise();

    //
    // We can't retry if RETRY_STRICT is set and the completion status
    // is not COMPLETED_NO
    //
    if(ex.exception -> completed() != COMPLETED_NO)
	ex._raise();

    //
    // If no client/profile pairs are left, we cannot retry either
    //
    if(clientProfilePairs_ -> size() == 0)
    {
#ifndef HAVE_OBE_NO_TRACE
	CoreTraceLevels_var coreTraceLevels =
	    _ob_orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceRetry() >= 2)
	{
	    String_var msg = string_dup("no profiles left to try\n");
	    String_var exMsg = ex.exception -> _rep_id();
	    msg += exMsg;
	    
	    Logger_var logger = _ob_orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
#endif // !HAVE_OBE_NO_TRACE
	
	ex._raise();
    }

#ifndef HAVE_OBE_NO_TRACE
    //
    // OK, let's continue with the next profile
    //
    CoreTraceLevels_var coreTraceLevels =
	_ob_orbInstance_ -> getCoreTraceLevels();
    if(coreTraceLevels -> traceRetry() >= 2)
    {
	String_var msg = string_dup("trying next profile\n");
	String_var exMsg = ex.exception -> _rep_id();
	msg += exMsg;
	
	Logger_var logger = _ob_orbInstance_ -> getLogger();
	logger -> trace("retry", msg);
    }
#endif // !HAVE_OBE_NO_TRACE
}

// ----------------------------------------------------------------------
// MarshalStubImpl constructor and destructor
// ----------------------------------------------------------------------

MarshalStubImpl::MarshalStubImpl()
    : clientProfilePairs_(0)
{
}

MarshalStubImpl::~MarshalStubImpl()
{
    //
    // If the ORB has been destroyed then the clientManager can be nil
    //
    ClientManager_var clientManager = _ob_orbInstance_ -> getClientManager();
    if(!is_nil(clientManager) && clientProfilePairs_ != 0)
    {
	OB_STL::list<ClientProfilePair_var>::iterator iter;
        for(iter = clientProfilePairs_ -> begin() ;
	    iter != clientProfilePairs_ -> end() ;
	    iter++)
        {
	    clientManager -> releaseClient((*iter) -> client);
        }
    }

    delete clientProfilePairs_;
}

// ----------------------------------------------------------------------
// MarshalStubImpl public member implementation
// ----------------------------------------------------------------------

Boolean
MarshalStubImpl::_is_a(const char* id)
{
    while(true)
    {
	Client_ptr _ob_client;	
	RefCountProfileInfo_ptr _ob_profile;
	_OB_getClientProfilePair_nodup(_ob_client, _ob_profile);
	DowncallImpl _ob_down(_ob_orbInstance_, _ob_client, _ob_profile,
			      _ob_policies_, "_is_a", 5, true);
	try
	{
	    OutputStreamImpl* _ob_out = _ob_down.preMarshal();
	    try
	    {
		_ob_out -> write_string(id);
	    }
	    catch(const CORBA::SystemException& _ob_ex)
	    {
		_ob_down.marshalEx(_ob_ex);
	    }
	    _ob_down.postMarshal();
	    _ob_down.request();
	    InputStreamImpl* _ob_in = _ob_down.preUnmarshal();
	    Boolean _ob_r = false;
	    try
	    {
		_ob_r = _ob_in -> read_boolean();
	    }
	    catch(const CORBA::SystemException& _ob_ex)
	    {
		_ob_down.unmarshalEx(_ob_ex);
	    }
	    _ob_down.postUnmarshal();
	    return _ob_r;
	}
	catch(const FailureException& _ob_ex)
	{
	    _OB_handleFailureException(_ob_down, _ob_ex);
	}
    }
}

Boolean
MarshalStubImpl::_non_existent()
{
    while(true)
    {
	Client_ptr _ob_client;	
	RefCountProfileInfo_ptr _ob_profile;
	_OB_getClientProfilePair_nodup(_ob_client, _ob_profile);
	DowncallImpl _ob_down(_ob_orbInstance_, _ob_client, _ob_profile,
			      _ob_policies_, "_non_existent", 13, true);
	try
	{
	    _ob_down.preMarshal();
	    _ob_down.postMarshal();
	    _ob_down.request();
	    InputStreamImpl* _ob_in = _ob_down.preUnmarshal();
	    Boolean _ob_r = false;
	    try
	    {
		_ob_r = _ob_in -> read_boolean();
	    }
	    catch(const SystemException& _ob_ex)
	    {
		_ob_down.unmarshalEx(_ob_ex);
	    }
	    _ob_down.postUnmarshal();
	    return _ob_r;
	}
	catch(const FailureException& _ob_ex)
	{
	    _OB_handleFailureException(_ob_down, _ob_ex);
	}
    }
}
void
MarshalStubImpl::_OB_setup(ORBInstance_ptr orbInstance,
			   RefCountIOR_ptr ior,
			   RefCountIOR_ptr origIOR,
			   RefCountPolicyList_ptr policies)
{
    //
    // Save the ORBInstance object
    //
    _ob_orbInstance_ = ORBInstance::_duplicate(orbInstance);

    //
    // Save the IOR
    //
    IOR_ = RefCountIOR::_duplicate(ior);
    origIOR_ = RefCountIOR::_duplicate(origIOR);

    //
    // Save the policies
    //
    _ob_policies_ = RefCountPolicyList::_duplicate(policies);
}
