// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/Object.h>
#include <OBE/Stream.h>
#include <OBE/IIOP.h>
#include <OBE/OBPolicies.h>
#include <OBE/OCIClient_IIOP_init.h>
#include <OBE/seq_converter.h>

#include <OCIClient_IIOP_impl.h>
#include <IIOPUtil.h>

//
// Some operating systems don't know socklen_t
//
#ifdef HAVE_SOCKLEN_T
    typedef socklen_t socklen_type;
#else
#   ifdef HAVE_SOCKETS_WITH_SIZE_T
        typedef size_t socklen_type;
#   else
        typedef int socklen_type;
#   endif
#endif

//
// Undefine minor (this is needed for SunOS 4.1 and some other Unix
// flavors)
//
#undef minor

using namespace CORBA;
using namespace OB;

//#define OB_OCI_DEBUG

// ----------------------------------------------------------------------
// Some utility functions
// ----------------------------------------------------------------------

bool
OB::IIOPHostMatch(const char* host1, const char* host2, bool loopbackMatches)
{
    //
    // Direct host name comparison
    //
    if(strcmp(host1, host2) != 0)
    {
        //
        //
        // Direct host name comparision failed - must look up
        // addresses to be really sure if the hosts differ
        //
        try
        {
            struct sockaddr_in addr1;
            GetInAddr(host1, addr1);
            
            struct sockaddr_in addr2;
            GetInAddr(host2, addr2);
            
            if(addr1.sin_addr.s_addr != addr2.sin_addr.s_addr)
            {
                //
                // Address comparison failed - shall I extract
                // the key if the profile body contains the
                // loopback address?
                //
                if(loopbackMatches)
                {
                    if(addr2.sin_addr.s_addr != htonl(INADDR_LOOPBACK))
                        return false;
                }
                else
                    return false;
            }
        }
        catch(const COMM_FAILURE&)
        {
            //
            // Continue on hostname lookup failure
            //
            return false;
        }
    }

    return true;
}

void
OB::IIOPExtractAllProfileInfos(const OCI::IOR& ior,
                               OCI::ProfileInfoSeq& profileInfoSeq,
                               bool performMatch,
                               const char* host, UShort port,
                               bool loopbackMatches)
{
    for(ULong i = 0 ; i < ior.profiles.length() ; i++)
    {
	if(ior.profiles[i].tag == IOP::TAG_INTERNET_IOP)
	{
	    //
	    // Get the IIOP profile body
	    //
	    const Octet* coct = ior.profiles[i].profile_data.get_buffer();
            ULong len = ior.profiles[i].profile_data.length();
	    InputStreamImpl in((Octet*)coct, len);
            in._OB_readEndian();
	    ::IIOP::ProfileBody_1_0 body;
	    ::IIOP::ProfileBody_1_0::_OB_unmarshal(body, &in);

            if(performMatch)
            {
                //
                // Check primary host/port
                //
                bool match = false;
                if(port == body.port &&
                   IIOPHostMatch(host, body.host, loopbackMatches))
                {
                    match = true;
                }

                if(!match)
                    continue;
            }

            //
            // OK, found a match
            //
            len = profileInfoSeq.length() + 1;
            profileInfoSeq.length(len);
            profileInfoSeq[len - 1].key.length(body.object_key.length());
            memcpy(profileInfoSeq[len - 1].key.get_buffer(),
                   body.object_key.get_buffer(), body.object_key.length());
            profileInfoSeq[len - 1].minor = body.iiop_version.minor;
            profileInfoSeq[len - 1].major = body.iiop_version.major;
            profileInfoSeq[len - 1].id = ior.profiles[i].tag;
            profileInfoSeq[len - 1].index = i;
	}
    }
}

static Boolean
IIOPCompareBodies(const ::IIOP::ProfileBody_1_0& body1,
		  const ::IIOP::ProfileBody_1_0& body2)
{
    //
    // Compare versions
    //
    if(body1.iiop_version.major != body2.iiop_version.major ||
       body1.iiop_version.minor != body2.iiop_version.minor)
	return false;
    
    //
    // Compare ports
    //
    if(body1.port != body2.port)
	return false;
    
    //
    // Compare object keys
    //
    if(body1.object_key.length() != body2.object_key.length())
	return false;
    
    if(memcmp(body1.object_key.get_buffer(), body2.object_key.get_buffer(),
	      body1.object_key.length()) != 0)
	return false;
    
    //
    // Direct host name comparison
    //
    if(strcmp(body1.host, body2.host) != 0)
    {
	//
	// Direct host name comparision failed - must look up
	// addresses to be really sure if the hosts differ
	//
	try
	{
	    struct sockaddr_in addr1;
	    GetInAddr(body1.host, addr1);
	    
	    struct sockaddr_in addr2;
	    GetInAddr(body2.host, addr2);
	    
	    if(addr1.sin_addr.s_addr != addr2.sin_addr.s_addr)
		return false;
	}
	catch(const COMM_FAILURE&)
	{
	    //
	    // Return false on hostname lookup failure
	    //
	    return false;
	}
    }
    
    //
    // OK, found a match
    //
    return true;
}

static Boolean
IIOPEquivalent(const OCI::IOR& ior1, const OCI::IOR& ior2)
{
    ULong p1, p2, b1, b2;
    ULong cnt1 = 0, cnt2 = 0;
    ::IIOP::ProfileBody_1_0** bodies1 = 0;
    ::IIOP::ProfileBody_1_0** bodies2 = 0;

    //
    // Calculate number of IIOP profiles in ior1
    //
    for(p1 = 0 ; p1 < ior1.profiles.length() ; p1++)
	if(ior1.profiles[p1].tag == IOP::TAG_INTERNET_IOP)
	    cnt1++;
    
    //
    // Calculate number of IIOP profiles in ior2
    //
    for(p2 = 0 ; p2 < ior2.profiles.length() ; p2++)
	if(ior2.profiles[p2].tag == IOP::TAG_INTERNET_IOP)
	    cnt2++;

    //
    // Return false now if the number of IIOP profile bodies do not
    // match
    //
    if(cnt1 != cnt2)
	return false;

    //
    // Create an array with all IIOP profile bodies of ior1
    //
    bodies1 = new ::IIOP::ProfileBody_1_0*[cnt1];
    for(p1 = 0, b1 = 0 ; p1 < ior1.profiles.length() ; p1++)
	if(ior1.profiles[p1].tag == IOP::TAG_INTERNET_IOP)
	{
	    const Octet* coct =	ior1.profiles[p1].profile_data.get_buffer();
            ULong len = ior1.profiles[p1].profile_data.length();
	    InputStreamImpl in((Octet*)coct, len);
            in._OB_readEndian();
	    ::IIOP::ProfileBody_1_0* p = new ::IIOP::ProfileBody_1_0;
	    ::IIOP::ProfileBody_1_0::_OB_unmarshal(*p, &in);
	    bodies1[b1++] = p;
	}
    
    assert(b1 == cnt1);
    
    //
    // Create an array with all IIOP profile bodies of ior2
    //
    bodies2 = new ::IIOP::ProfileBody_1_0*[cnt2];
    for(p2 = 0, b2 = 0 ; p2 < ior2.profiles.length() ; p2++)
	if(ior2.profiles[p2].tag == IOP::TAG_INTERNET_IOP)
	{
	    const Octet* coct =	ior2.profiles[p2].profile_data.get_buffer();
	    ULong len = ior2.profiles[p2].profile_data.length();
	    InputStreamImpl in((Octet*)coct, len);
            in._OB_readEndian();
	    ::IIOP::ProfileBody_1_0* p = new ::IIOP::ProfileBody_1_0;
	    ::IIOP::ProfileBody_1_0::_OB_unmarshal(*p, &in);
	    bodies2[b2++] = p;
	}
    
    assert(b2 == cnt2);

    //
    // Check for profile body matches
    //
    for(b1 = 0 ; b1 < cnt1 ; b1++)
    {
	for(b2 = 0 ; b2 < cnt2 ; b2++)
	{
	    if(!bodies2[b2])
		continue;
	    
	    //
	    // Compare profile bodies
	    //
	    if(IIOPCompareBodies(*bodies1[b1], *bodies2[b2]))
	    {
		//
		// OK, found a match
		//
		delete bodies1[b1];
		bodies1[b1] = 0;
		delete bodies2[b2];
		bodies2[b2] = 0;
		break;
	    }
	}
    }

    //
    // Check whether there are any unmatched IIOP profile bodies
    //
    ULong unmatched1 = 0;
    for(b1 = 0 ; b1 < cnt1 ; b1++)
	if(bodies1[b1])
	{
	    unmatched1++;
	    delete bodies1[b1];
	}

    delete [] bodies1;

    ULong unmatched2 = 0;
    for(b2 = 0 ; b2 < cnt2 ; b2++)
	if(bodies2[b2])
	{
	    unmatched2++;
	    delete bodies2[b2];
	}

    delete [] bodies2;

    return unmatched1 == 0 && unmatched2 == 0;
}

static ULong
IIOPHash(const OCI::IOR& ior, ULong maximum)
{
    ULong hash = 0;

    for(ULong i = 0 ; i < ior.profiles.length() ; i++)
    {
	if(ior.profiles[i].tag == IOP::TAG_INTERNET_IOP)
	{
	    //
	    // Get the IIOP profile body
	    //
	    const Octet* coct = ior.profiles[i].profile_data.get_buffer();
            ULong len = ior.profiles[i].profile_data.length();
	    InputStreamImpl in((Octet*)coct, len);
            in._OB_readEndian();
	    ::IIOP::ProfileBody_1_0 body;
	    ::IIOP::ProfileBody_1_0::_OB_unmarshal(body, &in);

	    //
	    // Add port to hash
	    //
	    hash ^= body.port;

	    //
	    // Add object key to hash
	    //
	    for(ULong j = 0 ; j + 1 < body.object_key.length() ; j += 2)
		hash ^= body.object_key[j + 1] * 256 + body.object_key[j];
	}
    }

    return hash % (maximum + 1);
}

// ----------------------------------------------------------------------
// Transport_impl private and protected member implementations
// ----------------------------------------------------------------------

void
OCI::IIOP::Transport_impl::setBlock(Boolean block)
{
    if(block != block_)
    {
	block_ = block;
	
	if(block)
	    SetBlock(fd_);
	else
	    SetNoBlock(fd_);
    }
}

// ----------------------------------------------------------------------
// Transport_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::IIOP::Transport_impl::Transport_impl(OCI::Handle fd)
    : fd_(fd), block_(true)
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "Transport_impl" << OB_ENDL;
#endif

    //
    // Set the socket to blocking initially
    //
    SetBlock(fd_);

#ifdef WIN32
    //
    // Set a larger send buffer size to avoid a performance problem
    // on Win32. See send_detect().
    //
    // TODO: This is a total hack!
    //
    int bufsz = 64 * 1024; // 64K
    if(setsockopt(fd_, SOL_SOCKET, SO_SNDBUF, (char*)&bufsz,
       sizeof(int)) == -1)
    {
        int err = GetError();
        CloseSocket(fd_);
        _OB_setRef(0);
        SetError(err);
        throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }
#endif
}

OCI::IIOP::Transport_impl::~Transport_impl()
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "~Transport_impl" << OB_ENDL;
#endif

    if(fd_ >= 0)
	close();
}

// ----------------------------------------------------------------------
// Transport_impl public member implementation
// ----------------------------------------------------------------------

void
OCI::IIOP::Transport_impl::close()
    throw(SystemException)
{
    //
    // Close the socket
    //
    int saveFd = fd_;
    fd_ = -1; // Must be set to -1 before the shutdown/close
    ::shutdown(saveFd, SHUT_RDWR); // This helps to unblock threads blocking
                                   // in recv()
    CloseSocket(saveFd);
}

void
OCI::IIOP::Transport_impl::shutdown()
    throw(SystemException)
{
    ::shutdown(fd_, SHUT_WR); // Shutdown send side only
}

void
OCI::IIOP::Transport_impl::receive(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    setBlock(block);

    while(!buf -> is_full())
    {
	int result = ::recv(fd_, (char*)buf -> rest(),
			    buf -> rest_length(), 0);

	if(result == 0)
	{
	    throw COMM_FAILURE("", MinorRecvZero, COMPLETED_NO);
	}
	else if(result == -1)
	{
            if(IsInterrupted())
                continue;
            
	    if(WouldBlock())
	    {
		if(!block)
		    return;
		
		setBlock(true);
		continue;
	    }

	    throw COMM_FAILURE(LastError(), MinorRecv, COMPLETED_NO);
	}

	buf -> advance(result);
    }
}

Boolean
OCI::IIOP::Transport_impl::receive_detect(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    setBlock(block);

    while(!buf -> is_full())
    {
        int result = ::recv(fd_, (char*)buf -> rest(),
			    buf -> rest_length(), 0);

	if(result == 0)
	{
	    return false;
	}
	else if(result == -1)
	{
            if(IsInterrupted())
                continue;
            
	    if(fd_ == -1 || ConnectionDown())
		return false;

	    if(WouldBlock())
	    {
		if(!block)
		    return true;
		
		setBlock(true);
		continue;
	    }

	    throw COMM_FAILURE(LastError(), MinorRecv, COMPLETED_NO);
	}

	buf -> advance(result);
    }

    return true;
}

void
OCI::IIOP::Transport_impl::receive_timeout(OCI::Buffer_ptr buf, ULong t)
    throw(SystemException)
{
    if(t == 0 || fd_ == -1 /*TODO*/)
    {
	receive(buf, false);
	return;
    }

    setBlock(false);

    while(!buf -> is_full())
    {
	fd_set fdSet;
	FD_ZERO(&fdSet);

        //
        // It's possible for the fd_ member to become -1 if this
        // transport is closed by another thread
        //
        int fd = fd_;
        if(fd != -1)
        {
            //
            // Use fd here, otherwise FD_SET will crash if given a
            // file descriptor of -1
            //
            FD_SET(fd, &fdSet);

            struct timeval tv;
            tv.tv_sec = t / 1000;
            tv.tv_usec = (t - tv.tv_sec * 1000) * 1000;

            //
            // Use fd here, since if fd_ == -1, select will simply
            // block for the specified timeout period
            //
            int resultS = ::select(fd + 1, &fdSet, 0, 0, &tv);
            
            if(resultS == 0)
            {
                return;
            }
            else if(resultS == -1)
            {
                if(IsInterrupted())
                    continue;
                
                throw COMM_FAILURE(LastError(), MinorSelect, COMPLETED_NO);
            }
        }

        //
        // The fd_ member may be -1 at this point. If so, recv will fail.
        //
        int result = ::recv(fd_, (char*)buf -> rest(), 
			    buf -> rest_length(), 0);

        if(result == 0)
        {
            throw COMM_FAILURE("", MinorRecvZero, COMPLETED_NO);
        }
        else if(result == -1)
        {
            if(IsInterrupted())
                continue;
            
	    if(WouldBlock())
		return;

            throw COMM_FAILURE(LastError(), MinorRecv, COMPLETED_NO);
        }

        buf -> advance(result);
    }
}

void
OCI::IIOP::Transport_impl::send(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    setBlock(block);

    ULong sizeOfPacket = buf -> rest_length();

    while(!buf -> is_full())
    {
        int result = ::send(fd_, (const char*)buf -> rest(), sizeOfPacket, 0);

	if(result == 0)
	{
	    throw COMM_FAILURE("", MinorSendZero, COMPLETED_NO);
	}
	else if(result == -1)
	{
	    if(NoBufs() && sizeOfPacket > 1024)
	    {
		sizeOfPacket /= 2;
		continue;
	    }

            if(IsInterrupted())
                continue;
            
	    if(WouldBlock())
	    {
		if(!block)
		    return;
		
		setBlock(true);
		continue;
	    }

	    throw COMM_FAILURE(LastError(), MinorSend, COMPLETED_NO);
	}

	buf -> advance(result);

	if(buf -> rest_length() < sizeOfPacket)
	    sizeOfPacket = buf -> rest_length();
    }
}

Boolean
OCI::IIOP::Transport_impl::send_detect(OCI::Buffer_ptr buf, Boolean block)
    throw(SystemException)
{
    setBlock(block);

    ULong sizeOfPacket = buf -> rest_length();

    while(!buf -> is_full())
    {
#ifdef WIN32
        //
        // Limit the maximum size of a message, in order to avoid
        // a performance problem on Win32.
        //
        // Note: The maximum size must be less than the socket's
        // send buffer size set by the constructor.
        //
        // TODO: This is a total hack!
        //
        ULong toSend = sizeOfPacket;
        if(sizeOfPacket > 65535)
            toSend = 65535;

        int result = ::send(fd_, (const char*)buf -> rest(), toSend, 0);
#else
	int result = ::send(fd_, (const char*)buf -> rest(), sizeOfPacket, 0);
#endif

	if(result == 0)
	{
	    return false;
	}
	else if(result == -1)
	{
	    if(NoBufs() && sizeOfPacket > 1024)
	    {
		sizeOfPacket /= 2;
		continue;
	    }

            if(IsInterrupted())
                continue;
            
	    if(fd_ == -1 || ConnectionDown())
		return false;

	    if(WouldBlock())
	    {
		if(!block)
		    return true;
		
		setBlock(true);
		continue;
	    }

	    throw COMM_FAILURE(LastError(), MinorSend, COMPLETED_NO);
	}

	buf -> advance(result);

	if(buf -> rest_length() < sizeOfPacket)
	    sizeOfPacket = buf -> rest_length();
    }

    return true;
}

void
OCI::IIOP::Transport_impl::send_timeout(OCI::Buffer_ptr buf, ULong t)
    throw(SystemException)
{
    if(t == 0 || fd_ == -1 /*TODO*/)
    {
	send(buf, false);
	return;
    }

    setBlock(false);
    
    ULong sizeOfPacket = buf -> rest_length();

    while(!buf -> is_full())
    {
	fd_set fdSet;
	FD_ZERO(&fdSet);

        //
        // It's possible for the fd_ member to become -1 if this
        // transport is closed by another thread
        //
        int fd = fd_;
        if(fd != -1)
        {
            //
            // Use fd here, otherwise FD_SET will crash if given a
            // file descriptor of -1
            //
            FD_SET(fd, &fdSet);

            struct timeval tv;
            tv.tv_sec = t / 1000;
            tv.tv_usec = (t - tv.tv_sec * 1000) * 1000;

            //
            // Use fd here, since if fd_ == -1, select will simply
            // block for the specified timeout period
            //
            int resultS = ::select(fd + 1, 0, &fdSet, 0, &tv);
            
            if(resultS == 0)
            {
                return;
            }
            else if(resultS == -1)
            {
                if(IsInterrupted())
                    continue;
                
                throw COMM_FAILURE(LastError(), MinorSelect, COMPLETED_NO);
            }
        }

        //
        // The fd_ member may be -1 at this point. If so, send will fail.
        //
        int result = ::send(fd_, (const char*)buf -> rest(), sizeOfPacket, 0);

        if(result == 0)
        {
            throw COMM_FAILURE("", MinorSendZero, COMPLETED_NO);
        }
        else if(result == -1)
        {
	    if(NoBufs() && sizeOfPacket > 1024)
	    {
		sizeOfPacket /= 2;
		continue;
	    }

            if(IsInterrupted())
                continue;
            
	    if(WouldBlock())
		return;

            throw COMM_FAILURE(LastError(), MinorSend, COMPLETED_NO);
        }

        buf -> advance(result);

	if(buf -> rest_length() < sizeOfPacket)
	    sizeOfPacket = buf -> rest_length();
    }
}

// ----------------------------------------------------------------------
// Connector_impl private and protected member implementations
// ----------------------------------------------------------------------

void
OCI::IIOP::Connector_impl::close()
{
    //
    // Close the socket
    //
    int saveFd = fd_;
    fd_ = -1; // Must be set to -1 before the close
    CloseSocket(saveFd);
}

// ----------------------------------------------------------------------
// OCI::IIOP::Connector_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::IIOP::Connector_impl::Connector_impl(const char* host, UShort port)
    : host_(host), port_(port), fd_(-1)
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "OCI::IIOP::Connector_impl" << OB_ENDL;
#endif
}

OCI::IIOP::Connector_impl::~Connector_impl()
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "~OCI::IIOP::Connector_impl" << OB_ENDL;
#endif

    if(fd_ >= 0)
	close();
}

// ----------------------------------------------------------------------
// Connector_impl public member implementation
// ----------------------------------------------------------------------

OCI::Transport_ptr
OCI::IIOP::Connector_impl::connect()
    throw(SystemException)
{
    if(fd_ >= 0)
	close();

    //
    // Create socket
    //
    if((fd_ = CreateSocket()) == -1)
    {
	fd_ = -1;
	throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    }

    //
    // Set TCP_NODELAY option
    //
    int no = 1;
    if(setsockopt(fd_, IPPROTO_TCP, TCP_NODELAY, (char*)&no, sizeof(int))
       == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	fd_ = -1;
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }
   
    //
    // Get the address
    //
    struct sockaddr_in addr;
    memset(&addr, 0, sizeof(struct sockaddr_in));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port_);
    GetInAddr(host_, addr);

    //
    // Connect
    //
    while(true)
    {
        if(::connect(fd_, OB_REINTERPRET_CAST(struct sockaddr*, &addr),
		     sizeof(addr)) == -1)
        {
            if(IsInterrupted())
                continue;
	    
	    int err = GetError();
	    CloseSocket(fd_);
	    fd_ = -1;
	    SetError(err);
	    if(ConnectFailed())
		throw TRANSIENT(LastError(), MinorConnectFailed, COMPLETED_NO);
	    else
		throw COMM_FAILURE(LastError(), MinorConnect, COMPLETED_NO);
        }

        break;
    }
    
    //
    // Create new transport
    //
    Handle fd = fd_;
    fd_ = -1;
    OCI::Transport_var tr = new Transport_impl(fd);

    //
    // Return new transport
    //
    return tr._retn();
}

OCI::Transport_ptr
OCI::IIOP::Connector_impl::connect_timeout(ULong t)
    throw(SystemException)
{
    if(fd_ >= 0)
	close();

    //
    // Create socket
    //
    if((fd_ = CreateSocket()) == -1)
    {
	fd_ = -1;
	throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    }

    //
    // Set TCP_NODELAY option
    //
    int no = 1;
    if(setsockopt(fd_, IPPROTO_TCP, TCP_NODELAY, (char*)&no, sizeof(int))
       == -1)
    {
	int err = GetError();
	CloseSocket(fd_);
	fd_ = -1;
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }

    //
    // Set socket to non-blocking mode
    //
    SetNoBlock(fd_);
   
    //
    // Get the address
    //
    struct sockaddr_in addr;
    memset(&addr, 0, sizeof(struct sockaddr_in));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port_);
    GetInAddr(host_, addr);

    //
    // Connect
    //
    while(true)
    {
        if(::connect(fd_, OB_REINTERPRET_CAST(struct sockaddr*, &addr),
		     sizeof(addr)) == -1)
        {
            if(IsInterrupted())
                continue;

	    if(InProgress())
	    {
		//
		// Call select() until timeout expired or connection
		// fd_ becomes writeable.
		//
		while(true)
		{
		    
		    struct timeval tv;
		    tv.tv_sec = t / 1000;
		    tv.tv_usec = (t - tv.tv_sec * 1000) * 1000;

		    fd_set writeFdSet;
		    FD_ZERO(&writeFdSet);
		    FD_SET(fd_, &writeFdSet);
#ifndef WIN32
		    int resultS = ::select(fd_ + 1, 0, &writeFdSet, 0, &tv);
#else
		    //
		    // Under WIN32 ::connect failure notification is
		    // done through the exceptfds.
		    //
		    fd_set exceptFdSet;
		    FD_ZERO(&exceptFdSet);
		    FD_SET(fd_, &exceptFdSet);

		    int resultS = ::select(fd_ + 1, 0, &writeFdSet,
					   &exceptFdSet, &tv);
#endif
		    
		    if(resultS == 0)
		    {
			CloseSocket(fd_);
			fd_ = -1;
			return Transport::_nil();
		    }
		    else if(resultS == -1)
		    {
			if(IsInterrupted())
			    continue;
			
			throw COMM_FAILURE(LastError(), MinorSelect,
					   COMPLETED_NO);
		    }

#ifdef WIN32
		    //
		    // Windows seems to need this call to Sleep(),
		    // otherwise no error is reported through
		    // getsockopt. I have no idea why.
		    //
		    Sleep(0);
#endif
		    //
		    // Check for connect error
		    //
		    socklen_type optlen = sizeof(socklen_type);
		    int optval;
		    if(getsockopt(fd_, SOL_SOCKET, SO_ERROR, (char*)&optval,
				  &optlen) == -1)
		    {
			int err = GetError();
			CloseSocket(fd_);
			fd_ = -1;
			SetError(err);
			throw COMM_FAILURE(LastError(), MinorGetsockopt,
					   COMPLETED_NO);
		    }
		    
		    if(optval > 0)
		    {
			CloseSocket(fd_);
			fd_ = -1;
			SetError(optval);
			if(ConnectFailed())
			    throw TRANSIENT(LastError(), MinorConnectFailed,
					    COMPLETED_NO);
			else
			    throw COMM_FAILURE(LastError(), MinorConnect,
					       COMPLETED_NO);
		    }
		    break;
		}
	    }
	    else
	    {
		int err = GetError();
		CloseSocket(fd_);
		fd_ = -1;
		SetError(err);
		throw COMM_FAILURE(LastError(), MinorConnect, COMPLETED_NO);
	    }
        }
        break;
    }
    
    //
    // Create new transport
    //
    Handle fd = fd_;
    fd_ = -1;
    OCI::Transport_var tr = new Transport_impl(fd);

    //
    // Return new transport
    //
    return tr._retn();
}

OCI::ProfileInfoSeq*
OCI::IIOP::Connector_impl::get_usable_profiles(
    const IOP::IOR& ior,
    const CORBA::PolicyList& policies)
    throw(SystemException)
{
    OCI::ProfileInfoSeq_var seq = new OCI::ProfileInfoSeq;
    
    //
    // Make sure that the set of policies is met
    //
    for(ULong i = 0; i < policies.length(); i++)
    {
        if(policies[i] -> policy_type() == PROTOCOL_POLICY_ID)
        {
            ProtocolPolicy_var protocolPolicy =
                ProtocolPolicy::_narrow(policies[i]);
            assert(!is_nil(protocolPolicy));
            if(protocolPolicy -> value() != id())
                return seq._retn();
        }
    }

    IIOPExtractAllProfileInfos(ior, seq.inout(), true, host_, port_, false);

    return seq._retn();
}

CORBA::Boolean
OCI::IIOP::Connector_impl::equal(Connector_ptr con)
    throw(CORBA::SystemException)
{
    Connector_impl* impl = dynamic_cast<OCI::IIOP::Connector_impl*>(con);
    if(!impl)
	return false;
    
    //
    // Compare ports
    //
    if(port_ != impl -> port_)
	return false;
    
    //
    // Direct host name comparison
    //
    if(strcmp(host_, impl -> host_) != 0)
    {
	//
	//
	// Direct host name comparision failed - must look up
	// addresses to be really sure if the hosts differ
	//
	try
	{
	    struct sockaddr_in addr1;
	    GetInAddr(host_, addr1);
	    
	    struct sockaddr_in addr2;
	    GetInAddr(impl -> host_, addr2);
	    
	    if(addr1.sin_addr.s_addr != addr2.sin_addr.s_addr)
		return false;
	}
	catch(const COMM_FAILURE&)
	{
	    //
	    // Return false on hostname lookup failure
	    //
	    return false;
	}
    }

    //
    // OK, connectors are the same
    //
    return true;
}    

// ----------------------------------------------------------------------
// ConFactory_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::IIOP::ConFactory_impl::ConFactory_impl()
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "ConFactory_impl" << OB_ENDL;
#endif
}

OCI::IIOP::ConFactory_impl::~ConFactory_impl()
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "~ConFactory_impl" << OB_ENDL;
#endif
}

// ----------------------------------------------------------------------
// ConFactory_impl public member implementation
// ----------------------------------------------------------------------

OCI::ConnectorSeq*
OCI::IIOP::ConFactory_impl::create_connectors(const OCI::IOR& ior,
					      const PolicyList& policies)
    throw(SystemException)
{
    ULong i;

    //
    // Check whether policies are satisfied
    //
    for(i = 0; i < policies.length(); i++)
    {
	if(policies[i] -> policy_type() == PROTOCOL_POLICY_ID)
	{
	    ProtocolPolicy_var protocolPolicy =
		ProtocolPolicy::_narrow(policies[i]);
	    assert(!is_nil(protocolPolicy));
	    if(protocolPolicy -> value() != id())
		return new OCI::ConnectorSeq;
	}
    }

    //
    // Create Connectors from profiles
    //
    OCI::ConnectorSeq_var seq = new OCI::ConnectorSeq;
    for(i = 0 ; i < ior.profiles.length() ; i++)
    {
	if(ior.profiles[i].tag == tag())
	{
	    //
	    // Get the IIOP profile body
	    //
	    const Octet* coct = ior.profiles[i].profile_data.get_buffer();
	    ULong len = ior.profiles[i].profile_data.length();
	    InputStreamImpl in((Octet*)coct, len);
	    in._OB_readEndian();
	    ::IIOP::ProfileBody_1_0 body;
	    ::IIOP::ProfileBody_1_0::_OB_unmarshal(body, &in);
	    
	    //
	    // Create new connector for this profile
	    //
	    CORBA::ULong seqLength  = seq -> length();
	    seq -> length(seqLength + 1);
	    seq[seqLength] = new Connector_impl(body.host, body.port);
	}
    }

    return seq._retn();
}

Boolean
OCI::IIOP::ConFactory_impl::equivalent(const OCI::IOR& ior1,
				       const OCI::IOR& ior2)
    throw(SystemException)
{
    return IIOPEquivalent(ior1, ior2);
}

ULong
OCI::IIOP::ConFactory_impl::hash(const OCI::IOR& ior, ULong maximum)
    throw(SystemException)
{
    return IIOPHash(ior, maximum);
}

// ----------------------------------------------------------------------
// Initialization functions
// ----------------------------------------------------------------------

void
OCI::IIOP::initClient(OCI::ConFactoryRegistry_ptr registry)
{
    try
    {
        OCI::ConFactory_var factory = new ConFactory_impl;
        registry -> add_factory(factory);
    }
    catch(const OCI::FactoryAlreadyExists&)
    {
        // Ignore
    }
}
