// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/Object.h>
#include <OBE/Stream.h>
#include <OBE/IIOP.h>
#include <OBE/OBPolicies.h>
#include <OBE/OCI_IIOP_init.h>
#include <OBE/seq_converter.h>

#include <OCI_IIOP_impl.h>
#include <IIOPUtil.h>

//
// Some operating systems don't know socklen_t
//
#ifdef HAVE_SOCKLEN_T
    typedef socklen_t socklen_type;
#else
#   ifdef HAVE_SOCKETS_WITH_SIZE_T
        typedef size_t socklen_type;
#   else
        typedef int socklen_type;
#   endif
#endif

//
// Undefine minor (this is needed for SunOS 4.1 and some other Unix
// flavors)
//
#undef minor

using namespace CORBA;
using namespace OB;

//#define OB_OCI_DEBUG

// ----------------------------------------------------------------------
// Acceptor_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::IIOP::Acceptor_impl::Acceptor_impl(const StringSeq& hosts,
                                        bool multiProfile,
                                        UShort port, UShort backlog)
    : hosts_(hosts), multiProfile_(multiProfile), port_(port), fd_(-1)
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "OCI::IIOP::Acceptor_impl" << OB_ENDL;
#endif

    if(backlog == 0)
        backlog_ = 5;
    else
        backlog_ = backlog;

    //
    // Get the local address for use by connect_self. Since we are
    // binding to all network interfaces, we'll use the loopback
    // address.
    //
    memset(&localAddress_, 0, sizeof(localAddress_));
    localAddress_.sin_family = AF_INET;
    localAddress_.sin_port = htons(port_);
    localAddress_.sin_addr.s_addr = htonl(INADDR_LOOPBACK);

    int no;

    //
    // Get the bind address
    //
    struct sockaddr_in addr;
    memset(&addr, 0, sizeof(struct sockaddr_in));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port_);
    addr.sin_addr.s_addr = htonl(INADDR_ANY);

    //
    // Create socket
    //
    if((fd_ = CreateSocket()) == -1)
    {
	_OB_setRef(0);
	throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    }

    //
    // Set TCP_NODELAY option
    //
    no = 1;
    if(setsockopt(fd_, IPPROTO_TCP, TCP_NODELAY, (char*)&no, sizeof(int))
       == -1)
    {
	int err = GetError();
	close();
	_OB_setRef(0);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }

    //
    // Bind to all network interfaces
    //
#ifndef WIN32
    no = 1;
    if(setsockopt(fd_, SOL_SOCKET, SO_REUSEADDR, (char*)&no,
		  sizeof(int)) == -1)
    {
	int err = GetError();
	close();
	_OB_setRef(0);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }
#endif
	
    if(bind(fd_, OB_REINTERPRET_CAST(struct sockaddr*, &addr),
	    sizeof(addr)) == -1)
    {
	int err = GetError();
	close();
	_OB_setRef(0);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorBind, COMPLETED_NO);
    }

    socklen_type len = sizeof(addr);
    int result = getsockname(fd_,
			     OB_REINTERPRET_CAST(struct sockaddr*, &addr),
			     &len);
    assert(result != -1);

    //
    // Read back the port. This is needed if the port was selected by
    // the operating system.
    //
    port_ = ntohs(addr.sin_port);
    localAddress_.sin_port = addr.sin_port;
}

OCI::IIOP::Acceptor_impl::Acceptor_impl(const char* address,
                                        const StringSeq& hosts,
                                        bool multiProfile,
                                        UShort port, UShort backlog)
    : hosts_(hosts), multiProfile_(multiProfile), port_(port),
      backlog_(backlog), fd_(-1)
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "OCI::IIOP::Acceptor_impl" << OB_ENDL;
#endif

    if(backlog == 0)
        backlog_ = 5;
    else
        backlog_ = backlog;

    //
    // Get the local address for use by connect_self
    //
    memset(&localAddress_, 0, sizeof(localAddress_));
    localAddress_.sin_family = AF_INET;
    localAddress_.sin_port = htons(port_);
    GetInAddr(address, localAddress_);

    int no;

    //
    // Create socket
    //
    if((fd_ = CreateSocket()) == -1)
    {
	_OB_setRef(0);
	throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    }

    //
    // Set TCP_NODELAY option
    //
    no = 1;
    if(setsockopt(fd_, IPPROTO_TCP, TCP_NODELAY, (char*)&no, sizeof(int))
       == -1)
    {
	int err = GetError();
	close();
	_OB_setRef(0);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }

    //
    // Bind to requested network interface
    //
#ifndef WIN32
    no = 1;
    if(setsockopt(fd_, SOL_SOCKET, SO_REUSEADDR, (char*)&no,
		  sizeof(int)) == -1)
    {
	int err = GetError();
	close();
	_OB_setRef(0);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }
#endif
	
    if(bind(fd_, OB_REINTERPRET_CAST(struct sockaddr*, &localAddress_),
	    sizeof(localAddress_)) == -1)
    {
	int err = GetError();
	close();
	_OB_setRef(0);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorBind, COMPLETED_NO);
    }

    socklen_type len = sizeof(localAddress_);
    int result =
        getsockname(fd_, OB_REINTERPRET_CAST(struct sockaddr*, &localAddress_),
                    &len);
    assert(result != -1);

    //
    // Read back the port. This is needed if the port was selected by
    // the operating system.
    //
    port_ = ntohs(localAddress_.sin_port);
}

OCI::IIOP::Acceptor_impl::~Acceptor_impl()
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "~Acceptor_impl" << OB_ENDL;
#endif

    if(fd_ >= 0)
	close();
}

// ----------------------------------------------------------------------
// Acceptor_impl public member implementation
// ----------------------------------------------------------------------

void
OCI::IIOP::Acceptor_impl::close()
    throw(SystemException)
{
    //
    // Close the socket
    //
    int saveFd = fd_;
    fd_ = -1;
    CloseSocket(saveFd);
}

void
OCI::IIOP::Acceptor_impl::shutdown()
    throw(SystemException)
{
    ::shutdown(fd_, SHUT_RD); // Shutdown receive side only
}

void
OCI::IIOP::Acceptor_impl::listen()
    throw(SystemException)
{
    //
    // Listen
    //
    while(true)
    {
	if(::listen(fd_, backlog_) == -1)
	{
            if(IsInterrupted())
		continue;
	    int err = GetError();
	    close();
	    SetError(err);
	    throw COMM_FAILURE(LastError(), MinorListen, COMPLETED_NO);
	}
	break;
    }
}

OCI::Transport_ptr
OCI::IIOP::Acceptor_impl::accept(Boolean block)
    throw(SystemException)
{
    if(block)
	SetBlock(fd_);
    else
	SetNoBlock(fd_);

    //
    // Accept
    //
    struct sockaddr_in newAddr;
    socklen_type len = sizeof(newAddr);
    int newFd;
    while(true)
    {
	if((newFd = ::accept(fd_,
			     OB_REINTERPRET_CAST(struct sockaddr*, &newAddr),
			     &len)) == -1)
        {
	    //
	    // Here we have special handling for this case, since
	    // accept can generate other loop-and-retry errors on some
	    // platforms.
	    //
            if(AcceptRetry())
                continue;
	    
	    if(WouldBlock())
	    {
		if(!block)
		    return Transport::_nil();
		
		SetBlock(fd_);
		continue;
	    }

	    //
	    // Accept failed, throw an exception
	    //
            throw COMM_FAILURE(LastError(), MinorAccept, COMPLETED_NO);
        }
	
        break;
    }
	
    //
    // Set TCP_NODELAY option
    //
    int no = 1;
    if(setsockopt(newFd, IPPROTO_TCP, TCP_NODELAY, (char*)&no, sizeof(int))
       == -1)
    {
	int err = GetError();
	CloseSocket(newFd);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }
    
    //
    // Create new transport
    //
    OCI::Transport_var tr = new Transport_impl(newFd);

    //
    // Return new transport
    //
    return tr._retn();
}

OCI::Transport_ptr
OCI::IIOP::Acceptor_impl::connect_self()
    throw(SystemException)
{
    //
    // Create socket
    //
    int fd;
    if((fd = CreateSocket()) == -1)
    {
	throw COMM_FAILURE(LastError(), MinorSocket, COMPLETED_NO);
    }

    //
    // Set TCP_NODELAY option
    //
    int no = 1;
    if(setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (char*)&no, sizeof(int))
       == -1)
    {
	int err = GetError();
	CloseSocket(fd);
	SetError(err);
	throw COMM_FAILURE(LastError(), MinorSetsockopt, COMPLETED_NO);
    }
   
    //
    // Connect to local address
    //
    while(true)
    {
        if(::connect(fd, OB_REINTERPRET_CAST(struct sockaddr*, &localAddress_),
		     sizeof(localAddress_)) == -1)
        {
            if(IsInterrupted())
                continue;
	    
	    int err = GetError();
	    CloseSocket(fd);
	    SetError(err);
	    if(ConnectFailed())
		throw TRANSIENT(LastError(), MinorConnectFailed, COMPLETED_NO);
	    else
		throw COMM_FAILURE(LastError(), MinorConnect, COMPLETED_NO);
        }

        break;
    }
    
    //
    // Create and return new transport
    //
    return new Transport_impl(fd);
}

void
OCI::IIOP::Acceptor_impl::add_profiles(const OCI::ProfileInfo& profileInfo,
                                       OCI::IOR& ior)
    throw(SystemException)
{
    assert(port_ > 0);
    
    for(ULong i = 0 ; i < hosts_.length() ; i++)
    {
	::IIOP::ProfileBody_1_0 body;
	body.iiop_version.major = profileInfo.major;
	body.iiop_version.minor = profileInfo.minor;
	body.host = hosts_[i];
	body.port = port_;
	body.object_key.length(profileInfo.key.length());
	memcpy(body.object_key.get_buffer(), profileInfo.key.get_buffer(),
	       profileInfo.key.length());
	
	ULong len = ior.profiles.length() + 1;
	ior.profiles.length(len);
	ior.profiles[len - 1].tag = tag();
	
	OutputStreamImpl out;
	out._OB_writeEndian();
	body._OB_marshal(&out);
	ULong count = out._OB_pos();
	ior.profiles[len - 1].profile_data.length(count);
	Octet* oct = ior.profiles[len - 1].profile_data.get_buffer();
	memcpy(oct, out._OB_bufferImpl() -> data(), count);
    }
}

OCI::ProfileInfoSeq*
OCI::IIOP::Acceptor_impl::get_local_profiles(const OCI::IOR& ior)
    throw(SystemException)
{
    //
    // Get local profiles for all hosts
    //
    OCI::ProfileInfoSeq_var seq = new OCI::ProfileInfoSeq;

    for(ULong i = 0 ; i < hosts_.length() ; i++)
        IIOPExtractAllProfileInfos(ior, seq.inout(), true, hosts_[i], port_,
                                   true);

    return seq._retn();
}

// ----------------------------------------------------------------------
// AccFactory_impl constructor and destructor
// ----------------------------------------------------------------------

OCI::IIOP::AccFactory_impl::AccFactory_impl()
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "AccFactory_impl" << OB_ENDL;
#endif
}

OCI::IIOP::AccFactory_impl::~AccFactory_impl()
{
#ifdef OB_OCI_DEBUG
    OB_COUT << "~AccFactory_impl" << OB_ENDL;
#endif
}

// ----------------------------------------------------------------------
// AccFactory_impl public member implementation
// ----------------------------------------------------------------------

OCI::Acceptor_ptr
OCI::IIOP::AccFactory_impl::create_acceptor(const OCI::ParamSeq& params)
    throw(OCI::InvalidParam, SystemException)
{
    String_var bind;
    StringSeq hosts;
    bool multiProfile = false;
    UShort port = 0;
    UShort backlog = 0;
    bool numeric = false;

    for(ULong i = 0 ; i < params.length() ; i++)
    {
	if(strcmp(params[i].name, "bind") == 0)
	{
	    if(strlen(params[i].value) == 0)
		throw InvalidParam(params[i], "address cannot be empty");

	    bind = params[i].value;
	}
	else if(strcmp(params[i].name, "host") == 0)
	{
	    //
	    // Parse hosts, allowing a space or comma as the delimiters
	    //
	    const char* cp = params[i].value;
	    while(*cp != '\0')
	    {
		while(*cp != '\0' && (*cp == ' ' || *cp == ','))
		    cp++;
		const char* start = cp;
		while(*cp != '\0' && *cp != ' ' && *cp != ',')
		    cp++;
		ULong len = cp - start;
		if(len > 0)
		{
		    hosts.length(hosts.length() + 1);
		    hosts[hosts.length() - 1] = string_alloc(len);
		    strncpy((hosts[hosts.length() - 1].inout()), start, len);
		    hosts[hosts.length() - 1][len] = '\0';
		}
	    }

	    if(hosts.length() == 0)
		throw InvalidParam(params[i], "no hosts specified");
	}
	else if(strcmp(params[i].name, "multi_profile") == 0)
	{
	    multiProfile = (strcmp(params[i].value, "true") == 0);
	}
	else if(strcmp(params[i].name, "port") == 0)
	{
	    int p = atoi(params[i].value);
	    if(p < 1 || p > 65535)
		throw InvalidParam(params[i],
				   "port must be between 1 and 65535");

	    port = (UShort)p;
	}
	else if(strcmp(params[i].name, "backlog") == 0)
	{
	    int b = atoi(params[i].value);
	    if(b < 1 || b > 65535)
		throw InvalidParam(params[i],
				   "backlog must be between 1 and 65535");

	    backlog = (UShort)b;
	}
	else if(strcmp(params[i].name, "numeric") == 0)
	{
	    numeric = (strcmp(params[i].value, "true") == 0);
	}
	else
	    throw InvalidParam(params[i], "unknown parameter");
    }

    if(hosts.length() == 0)
    {
        hosts.length(1);
        hosts[0] = GetCanonicalHostname(numeric);
    }

    if(bind.in() == 0)
        return new Acceptor_impl(hosts, multiProfile, port, backlog);
    else
        return new Acceptor_impl(bind, hosts, multiProfile, port, backlog);
}

// ----------------------------------------------------------------------
// Initialization functions
// ----------------------------------------------------------------------

void
OCI::IIOP::initServer(OCI::AccFactoryRegistry_ptr registry)
{
    try
    {
        OCI::AccFactory_var factory = new AccFactory_impl;
        registry -> add_factory(factory);
    }
    catch(const OCI::FactoryAlreadyExists&)
    {
        // Ignore
    }
}
