// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/ORBControl.h>
#include <OBE/ORBInstance.h>
#include <OBE/InitialServiceManager.h>
#include <OBE/Properties.h>
#include <OBE/Logger.h>
#include <OBE/OBPOAManager.h>

#include <ORBControlS.h>
#include <ORBInstanceS.h>

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

using namespace OB;
using namespace CORBA;

// ----------------------------------------------------------------------
// ORBControlS private and protected member implementation
// ----------------------------------------------------------------------

void
ORBControlS::completeServerShutdown()
    throw()
{
    JTCSynchronized synchronized(*this);

    //
    // Get the server ORBInstance
    //
    ORBInstanceS* orbInstance = dynamic_cast<ORBInstanceS*>(orbInstance_.in());
    assert(orbInstance);

    //
    // If the shutdown_ is false, or the server side has already
    // shutdown then do nothing
    //
    if(!shutdown_ || state_ == StateServerShutdown)
	return;

    assert(state_ != StateClientShutdown && state_ != StateDestroyed);

    //
    // If run was called then only the main thread may complete the
    // shutdown
    //
    assert(state_ == StateNotRunning || mainThread_ == JTCThreadId::self());

    //
    // Deactivate the POAManager
    //
    OBPortableServer::POAManager_var manager = orbInstance -> getPOAManager();

    if(!CORBA::is_nil(manager))
	manager -> _OB_deactivate();
    
    //
    // Wait for all the threads in the server worker group to
    // terminate
    //
    JTCThreadGroupHandle group = orbInstance -> getServerWorkerGroup();
    if(group)
    {
    	JTCSynchronized groupSync(*group);
	while(group -> activeCount() > 0)
	{
	    try
	    {
		group -> wait();
	    }
	    catch(const JTCInterruptedException&)
	    {
	    }
	}
    }

    //
    // Mark the server side state as shutdown and notify any waiting
    // threads
    //
    state_ = StateServerShutdown;

    JTCSynchronized sync(monitor_);
    monitor_.notifyAll();
}

void
ORBControlS::validateState()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(state_ == StateDestroyed)
	throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    if(state_ == StateServerShutdown || state_ == StateClientShutdown)
	throw BAD_INV_ORDER("", MinorShutdownCalled, COMPLETED_NO);
	
    if(state_ == StateNotRunning)
    {
	//
	// Remember the main thread id
	//
	mainThread_ = JTCThreadId::self();

	//
	// Set the state to StateRunning
	//
	state_ = StateRunning;
    }
}

bool
ORBControlS::insideMethodInvocation()
    throw()
{
    bool inInvocation = false;
/* TODO: How to do this without a POACurrent?
    try
    {
	InitialServiceManager_var initialServiceManager =
	    orbInstance_ -> getInitialServiceManager();
	Object_var o = initialServiceManager ->
	    resolveInitialReferences("POACurrent");
	PortableServer::Current_impl* current =
	    dynamic_cast<PortableServer::Current_impl*>(o.in());
	inInvocation = (current != 0 && current -> _OB_inUpcall());
    }
    catch(const ORB::InvalidName&)
    {
    }
*/

    return inInvocation;
}

void
ORBControlS::blockServerShutdownComplete()
    throw()
{
    //
    // Wait for the server side to shutdown. Note that the client side
    // shutting down or the state being destroyed also implies that
    // the server side has shutdown
    //
    while(true)
    {
        {
	    JTCSynchronized synchronized(*this);
	    if(state_ != StateRunning)
	        return;
	}

	try
	{
	    JTCSynchronized sync(monitor_);
	    monitor_.wait();
	}
	catch(const JTCInterruptedException&)
	{
	}
    }
}

// ----------------------------------------------------------------------
// ORBControl constructor and destructor
// ----------------------------------------------------------------------

ORBControlS::ORBControlS(ORBInstance_ptr instance)
    : ORBControl(instance)
{
}

ORBControlS::~ORBControlS()
{
}

// ----------------------------------------------------------------------
// ORBControlS public member implementation
// ----------------------------------------------------------------------

void
ORBControlS::initializeRootPOA()
{
    //
    // Get the server ORBInstance
    //
    ORBInstanceS* orbInstance = dynamic_cast<ORBInstanceS*>(orbInstance_.in());
    assert(orbInstance);

    //
    // Get the POAManager
    //
    OBPortableServer::POAManager_var manager = orbInstance -> getPOAManager();

    //
    // Create the root POA
    //
    manager -> _OB_createRootPOA();
}

void
ORBControlS::run()
    throw(SystemException)
{
    validateState();
	
    if(mainThread_ != JTCThreadId::self())
    {
	//
	// If this is not the main thread then block until the ORB is
	// shutdown
	//
	blockServerShutdownComplete();
	return;
    }

    //
    // Run the event loop until the ORB server side has shutdown. Note
    // that the client side shutting down or the state being destroyed
    // also implies that the server side has shutdown
    //
    do
    {
	//
	// TODO: This is only a temporary solution.
	//
#if defined(WIN32)
        Sleep(1000);
#else
        sleep(1);
#endif

	//
	// After this call state is either ShutdownClient, or Running
	//
	completeServerShutdown();
    }
    while(state_ == StateRunning);
}

//
// ASYNC SAFE if waitForCompletion == false
//
void
ORBControlS::shutdownServer(bool waitForCompletion)
    throw(SystemException)
{
    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(state_ == StateDestroyed)
	throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    if(state_ == StateServerShutdown || state_ == StateClientShutdown)
	throw BAD_INV_ORDER("", MinorShutdownCalled, COMPLETED_NO);

    //
    // If waitForCompletion is true then find out whether we're inside
    // a method invocation -- if so throw a BAD_INV_ORDER exception
    //
    if(waitForCompletion && insideMethodInvocation())
	throw BAD_INV_ORDER(MinorDestroyWouldBlock, COMPLETED_NO);

    shutdown_ = true;

    //
    // waitForCompletion false? We're done.
    //
    if(!waitForCompletion)
	return;

    //
    // If run was called and this is not the main thread and
    // waitForCompletion is true then wait for the shutdown to
    // complete.
    //
    if(state_ == StateRunning && mainThread_ != JTCThreadId::self())
    {
	blockServerShutdownComplete();
	return;
    }

    //
    // This is the main thread -- complete the shutdown process
    //
    completeServerShutdown();
}

void
ORBControlS::shutdownServerClient()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(state_ == StateDestroyed)
        throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    //
    // If the ORB client side is already shutdown, then we're done
    //
    if(state_ == StateClientShutdown)
        return;

    //
    // First shutdown the server side, if necessary.
    //
    if(state_ != StateServerShutdown)
        shutdownServer(true);

    //
    // The server shutdown must have completed
    //
    assert(state_ == StateServerShutdown);

    //
    // Shutdown the client side. Continue to dispatch events until all
    // client type event handlers have unregistered.
    //
    ClientManager_var clientManager = orbInstance_ -> getClientManager();
    if(!is_nil(clientManager))
        clientManager -> destroy();

    //
    // Mark the ORB is client side as shutdown and notify any waiters
    //
    state_ = StateClientShutdown;

    JTCSynchronized sync(monitor_);
    monitor_.notifyAll();
}
