// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/ORBInstance.h>
#include <OBE/InitialServiceManager.h>
#include <OBE/POACommon.h>
#include <OBE/Properties.h>
#include <OBE/Logger.h>
#ifndef HAVE_OBE_NO_TRACE
#  include <OBE/CoreTraceLevels.h>
#endif // !HAVE_OBE_NO_TRACE
#include <OBE/Client.h>
#include <OBE/Net.h>
#include <OBE/OCIClient_IIOP_init.h>

#include <OCIClient_impl.h>

using namespace CORBA;
using namespace OB;

JTCMutex* ORBInstance::globalStateMutex_ = new JTCMutex;

namespace OB
{

class GlobalStateMutexDestroyer
{
public:

    ~GlobalStateMutexDestroyer()
    {
	delete ORBInstance::globalStateMutex_;
	ORBInstance::globalStateMutex_ = 0;
    }
};

static GlobalStateMutexDestroyer destroyer;

} // End of namespace OB

int ORBInstance::globalStateCounter_ = 0;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
ORBInstance::createGlobalState()
{
    if(globalStateMutex_ != 0)
	globalStateMutex_ -> lock();

    //
    // Only create on first call
    //
    if(globalStateCounter_++ == 0)
    {
#ifdef HAVE_WINSOCK_H
	//
	// Winsock startup
	//
	WORD wVersionRequested = MAKEWORD(1, 1);
	WSADATA wsaData;
	if(WSAStartup(wVersionRequested, &wsaData) != 0)
	{
	    if(globalStateMutex_ != 0)
		globalStateMutex_ -> unlock();

	    throw COMM_FAILURE(LastError(), MinorWSAStartup, COMPLETED_NO);
	}
#endif
	
#ifndef WIN32
	//
	// Set SIGPIPE signal handler
	//
	struct sigaction actPIPE;
	actPIPE.sa_handler = SIG_IGN;
	sigemptyset(&actPIPE.sa_mask);
	actPIPE.sa_flags = 0;
	sigaction(SIGPIPE, &actPIPE, 0);
#endif
    }
    
    if(globalStateMutex_ != 0)
	globalStateMutex_ -> unlock();
}

void
ORBInstance::destroyGlobalState()
{
    if(globalStateMutex_ != 0)
	globalStateMutex_ -> lock();

    //
    // Only clean up on last call
    //
    assert(globalStateCounter_ > 0);
    if(--globalStateCounter_ == 0)
    {
#ifdef HAVE_WINSOCK_H
	//
	// Winsock shutdown
	//
	WSACleanup();
#endif
	
#ifndef WIN32
	//
	// Reset SIGPIPE signal handler
	//
	struct sigaction actPIPE;
	actPIPE.sa_handler = SIG_DFL;
	sigemptyset(&actPIPE.sa_mask);
	actPIPE.sa_flags = 0;
	sigaction(SIGPIPE, &actPIPE, 0);
#endif
    }
    
    if(globalStateMutex_ != 0)
	globalStateMutex_ -> unlock();
}

void
OBDuplicate(OB::ORBInstance_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::ORBInstance_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// ORBInstance private member implementation
// ----------------------------------------------------------------------

void
ORBInstance::initialize(Properties_ptr properties,
			Logger_ptr logger
#ifndef HAVE_OBE_NO_TRACE
			, CoreTraceLevels_ptr coreTraceLevels
#endif // !HAVE_OBE_NO_TRACE
			)
{
    properties_ = Properties::_duplicate(properties);
    logger_ = Logger::_duplicate(logger);
#ifndef HAVE_OBE_NO_TRACE
    coreTraceLevels_ = CoreTraceLevels::_duplicate(coreTraceLevels);
#endif // !HAVE_OBE_NO_TRACE

    //
    // Create the InitialServiceManager
    //
    initServiceManager_ = new InitialServiceManager(this);

    //
    // Create the ClientManager
    //
    clientManager_ = createClientManager(this);

    //
    // Create the ObjectFactory
    //
    objectFactory_ = new ObjectFactory(this);

    //
    // Create the ConFactoryRegistry
    //
    conFactoryRegistry_ = new OCI::ConFactoryRegistry_impl();

    //
    // Add initial services to list
    //
    initServiceManager_ -> addInitialReference("RootPOA",
					       Object::_nil()); // Dummy
    initServiceManager_ -> addInitialReference("OCIConFactoryRegistry",
					       conFactoryRegistry_);

    //
    // Initialize the IIOP OCI plug-in, if necessary
    //
    const char* initIIOP = properties -> getProperty("ooc.orb.init_iiop");
    if(initIIOP == 0 ||
       strcmp(initIIOP, "both") == 0 ||
       strcmp(initIIOP, "client") == 0)
    {
	OCI::IIOP::initClient(conFactoryRegistry_);
    }

    //
    // Initialize server members
    //
    initializeHook();
}

void
ORBInstance::destroy()
{
    assert(!destroy_); // May only be destroyed once
    destroy_ = true;

    //
    // Destroy the server members
    //
    destroyHook();

    //
    // Destroy the Initial Service manager
    //
    if(!is_nil(initServiceManager_))
	initServiceManager_ -> destroy();
    initServiceManager_ = InitialServiceManager::_nil();

    //
    // Destroy the Object factory
    //
    if(!is_nil(objectFactory_))
	objectFactory_ -> destroy();
    objectFactory_ = ObjectFactory::_nil();

    //
    // Destroy the ClientManager
    //
    // ORBControl destroys the ClientManager
    //clientManager_ -> destroy();
    clientManager_ = ClientManager::_nil();

    //
    // Properties are not destroyed - they are indestructible
    //
    //properties_ -> destroy(); // No destroy operation defined
    //properties_ = Properties::_nil();

    //
    // Logger is not destroyed -- it is indestructable
    //
    //logger_ -> destroy();
    //logger_ = Logger::_nil();

#ifndef HAVE_OBE_NO_TRACE
    //
    // CoreTraceLevels is not destroyed -- it is indestructible
    //
    //coreTraceLevels_ -> destroy();
    //coreTraceLevels_ = CoreTraceLevels::_nil();
#endif // !HAVE_OBE_NO_TRACE

    //
    // Destroy the ConFactoryRegistry
    //
    //conFactoryRegistry_ -> destroy(); // No destroy operation defined
    conFactoryRegistry_ = OCI::ConFactoryRegistry::_nil();
}

// ----------------------------------------------------------------------
// ORBInstance constructor and destructor
// ----------------------------------------------------------------------

ORBInstance::ORBInstance()
    : destroy_(false)
{
    createGlobalState();
}

ORBInstance::~ORBInstance()
{
    assert(destroy_);

    destroyGlobalState();
}

// ----------------------------------------------------------------------
// ORBInstance public member implementation
// ----------------------------------------------------------------------

ObjectFactory_ptr
ORBInstance::getObjectFactory()
{
    return ObjectFactory::_duplicate(objectFactory_);
}

ClientManager_ptr
ORBInstance::getClientManager()
{
    return ClientManager::_duplicate(clientManager_);
}

InitialServiceManager_ptr
ORBInstance::getInitialServiceManager()
{
    return InitialServiceManager::_duplicate(initServiceManager_);
}

Properties_ptr
ORBInstance::getProperties()
{
    return Properties::_duplicate(properties_);
}

Logger_ptr
ORBInstance::getLogger()
{
    return Logger::_duplicate(logger_);
}

#ifndef HAVE_OBE_NO_TRACE
CoreTraceLevels_ptr
ORBInstance::getCoreTraceLevels()
{
    return CoreTraceLevels::_duplicate(coreTraceLevels_);
}
#endif // !HAVE_OBE_NO_TRACE

OCI::ConFactoryRegistry_ptr
ORBInstance::getConFactoryRegistry()
{
    return OCI::ConFactoryRegistry::_duplicate(conFactoryRegistry_);
}

// ----------------------------------------------------------------------
// Get an ORBInstance, given and ORB_ptr
// ----------------------------------------------------------------------

/* In OBORB.cpp */
