// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/Object.h>
#include <OBE/Logger.h>
#include <OBE/Properties.h>
#ifndef HAVE_OBE_NO_TRACE
#  include <OBE/CoreTraceLevels.h>
#endif // !HAVE_OBE_NO_TRACE
#include <OBE/ORB_init.h>
#include <OBE/ORBInstance.h>
#include <OBE/OBORB.h>

#include <Logger_impl.h>

#include <stdlib.h> // for strtoul
#include <ctype.h>
#include <errno.h>

#ifdef WIN32
# include <direct.h>
#endif

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// ORB Initialization
// ----------------------------------------------------------------------

void
OB::ParseArgs(int& ac, char** av, Properties_ptr properties,
	      Logger_ptr logger)
{
    assert(!is_nil(properties));

    OB::Logger_var dummy;
    if(is_nil(logger))
    {
	logger = new Logger_impl();
	dummy = logger;
    }

    String_var configFile;

    //
    // Scan through the argument list looking for the config file
    // variable. We need to do this since command line arguments have
    // precendence over the environment variables.
    //
    int i = 0;
    while(i < ac)
    {
        if(strcmp(av[i], "-ORBconfig") == 0)
	{
	    if(i + 1 >= ac)
	    {
#ifndef HAVE_OBE_NO_TRACE
		logger -> error("ORB_init: argument expected for -ORBconfig");
#endif
		throw INITIALIZE();
	    }
            configFile = string_dup(av[i + 1]);
	    for(int j = i ; j + 2 < ac ; j++)
		av[j] = av[j + 2];

	    ac -= 2;
	}
        else
            i++;
    }

    //
    // Load the contents of the configuration file if present
    //
    if(configFile && configFile[0] != '\0')
    {
	if(!properties -> load(configFile))
	{
#ifndef HAVE_OBE_NO_TRACE
	    String_var msg = string_dup(
		"ORB_init: could not load configuration file ");
	    msg += configFile;
	    logger -> warning(msg);
#endif
	}
    }

    //
    // Process the remaining -ORB arguments. Turn each argument into
    // an appropriate property.
    //
    i = 0;
    while(i < ac)
    {
        const char* option = av[i];
        if(strcmp(option, "-ORBInitRef") == 0)
        {
            if(i + 1 >= ac)
            {
#ifndef HAVE_OBE_NO_TRACE
                logger -> error("ORB_init: argument expected for "
				"-ORBInitRef");
#endif
                throw INITIALIZE();
            }

            const char* cp = strchr(av[i + 1], '=');
            if(cp != 0 && av[i + 1][0] != '=' && strlen(cp + 1) > 0)
            {
                String_var id = string_alloc(cp - av[i + 1]);
                strncpy(id.inout(), av[i + 1], cp - av[i + 1]);
                id[cp - av[i + 1]] = '\0';
                String_var prop = string_dup("ooc.orb.service.");
                prop += id;
                properties -> setProperty(prop, cp + 1);
            }
            else
            {
#ifndef HAVE_OBE_NO_TRACE
                logger -> error("ORB_init: invalid argument for "
                                "-ORBInitRef");
#endif
                throw INITIALIZE();
            }

            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
        }
        else if(strcmp(option, "-ORBDefaultInitRef") == 0)
        {
            if(i + 1 >= ac)
            {
#ifndef HAVE_OBE_NO_TRACE
                logger -> error("ORB_init: argument expected for "
                                "-ORBDefaultInitRef");
#endif
                throw INITIALIZE();
            }

            if(strlen(av[i + 1]) == 0)
            {
#ifndef HAVE_OBE_NO_TRACE
                logger -> error("ORB_init: invalid argument for "
                                "-ORBDefaultInitRef");
#endif
                throw INITIALIZE();
            }

            properties -> setProperty("ooc.orb.default_init_ref",
				      av[i + 1]);

            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];

            ac -= 2;
        }
        else if(strcmp(option, "-ORBtrace_connections") == 0 ||
		strcmp(option, "-ORBtrace_retry") == 0)
        {
            if(i + 1 >= ac)
            {
#ifndef HAVE_OBE_NO_TRACE
		String_var msg =
		(const char*)"ORB_init: argument expected for ";
		msg += option;
                logger -> error(msg);
#endif
                throw INITIALIZE();
            }

	    String_var prop = (const char*)"ooc.orb.trace.";
	    prop += option + 10;
            properties -> setProperty(prop, av[i + 1]);
            
            for(int j = i ; j + 2 < ac ; j++)
                av[j] = av[j + 2];
            
            ac -= 2;
        }        
        else if(strcmp(option, "-ORBversion") == 0)
        {
            logger -> info(Version);
            
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
            
            ac -= 1;
        }
        else if(strcmp(option, "-OAnumeric") == 0)
	{
            properties -> setProperty("ooc.iiop.numeric", "true");

            for(int j = i ; j + 1 < ac ; j++)
		av[j] = av[j + 1];

	    ac -= 1;
	}
	else if(strcmp(option, "-OAhost") == 0)
	{
	    if(i + 1 >= ac)
	    {
#ifndef HAVE_OBE_NO_TRACE
		logger -> error("ORB_init: argument expected for -OAhost");
#endif
		throw INITIALIZE();
	    }

            properties -> setProperty("ooc.iiop.host", av[i + 1]);

	    for(int j = i ; j + 2 < ac ; j++)
		av[j] = av[j + 2];

	    ac -= 2;
	}
	else if(strcmp(option, "-OAport") == 0)
	{
	    if(i + 1 >= ac)
	    {
#ifndef HAVE_OBE_NO_TRACE
                logger -> error("ORB_init: argument expected for -OAport");
#endif
		throw INITIALIZE();
            }

            properties -> setProperty("ooc.iiop.port", av[i + 1]);

	    for(int j = i ; j + 2 < ac ; j++)
		av[j] = av[j + 2];

	    ac -= 2;
	}
	else if(strcmp(option, "-OAbind") == 0)
	{
	    if(i + 1 >= ac)
	    {
#ifndef HAVE_OBE_NO_TRACE
                logger -> error("ORB_init: argument expected for -OAbind");
#endif
		throw INITIALIZE();
            }

            properties -> setProperty("ooc.iiop.bind", av[i + 1]);

	    for(int j = i ; j + 2 < ac ; j++)
		av[j] = av[j + 2];

	    ac -= 2;
	}
	else if(strcmp(option, "-OAbacklog") == 0)
	{
	    if(i + 1 >= ac)
	    {
#ifndef HAVE_OBE_NO_TRACE
                logger -> error("ORB_init: argument expected for -OAbacklog");
#endif
		throw INITIALIZE();
            }

            properties -> setProperty("ooc.iiop.backlog", av[i + 1]);

	    for(int j = i ; j + 2 < ac ; j++)
		av[j] = av[j + 2];

	    ac -= 2;
	}
        else if(strncmp(option, "-OA", 3) == 0 ||
		strncmp(option, "-ORB", 4) == 0)
        {
#ifndef HAVE_OBE_NO_TRACE
            String_var err =
	    string_dup("ORB_init: unknown option: ");
            err += option;
            logger -> warning(err);
#endif
	    
            for(int j = i ; j + 1 < ac ; j++)
                av[j] = av[j + 1];
	    
            ac -= 1;
        }
	else
	    i++;
    }
}

OBCORBA::ORB_ptr
OBCORBA::ORB_init(int& ac, char** av, Properties_ptr properties,
		  Logger_ptr logger,
		  const char* orb_identifier, const char* version)
{
    assert_nca(orb_identifier != 0, NCANullString);

    //
    // Scan through the argument list looking for -ORBid.  This is
    // needed to initialize the ORB itself. This variable can take the
    // form "-ORBidfoo" or "-ORBid foo".
    //
    int i = 0;
    while(i < ac)
    {
        if(strncmp(av[i], "-ORBid", 6) == 0)
        {
	    int n;

	    //
	    // Is the entire argument -ORBid?
	    //
	    if(strlen(av[i]) == 6)
	    {
		if(i + 1 >= ac)
		    throw INITIALIZE();
		orb_identifier = av[i+1];
		n = 2;
	    }
	    else
	    {
		orb_identifier = av[i] + 6;
		n = 1;
	    }
	    for(int j = i ; j + n < ac ; j++)
		av[j] = av[j + n];
	    ac -= n;
	    break;
        }
	else
	    i++;
    }

    //
    // Initialize the Logger
    //
    OB::Logger_var loggerDummy;
    if(is_nil(logger))
    {
	loggerDummy = new Logger_impl();
	logger = loggerDummy.in();
    }

    //
    // Check the version
    //
    if(strcmp(version, OB::Version) != 0)
    {
#ifndef HAVE_OBE_NO_TRACE
        String_var err = string_dup("ORB_init: version mismatch:");
        err += "\nlibrary version is ";
        err += Version;
        err += "\nheader file version is ";
        err += version;
        logger -> error(err);
#endif
        throw INITIALIZE();
    }

    //
    // Need to sync on the orbImplMap_ here to ensure that the
    // properties and what-not are initialized atomically
    //
    JTCSynchronized sync(CORBA::ORB::orbImplMapMutex_);

    //
    // Initialize the properties
    //
    Properties_var propertyDummy;
    if(is_nil(properties))
    {
	Properties_var defaultProps = Properties::getDefaultProperties();
	propertyDummy = new Properties(defaultProps);
	properties = propertyDummy.in();
    }

    ParseArgs(ac, av, properties, logger);

    //
    // Process each property
    //
    String_var id = orb_identifier;

    Properties::KeySeq keys = properties -> getKeys("ooc.orb.");
    ULong j;
    for(j = 0; j < keys.length(); ++j)
    {
        const char* key = keys[j];
        const char* value = properties -> getProperty(key);
        assert(value != 0);

	if(strncmp(key, "ooc.orb.trace.", 14) == 0)
        {
	    // Ignore -- handled in CoreTraceLevels
        }
        else if(strcmp(key, "ooc.orb.id") == 0)
        {
            id = string_dup(value);
        }
        else if(strcmp(key, "ooc.orb.init_iiop") == 0)
        {
            if(strcmp(value, "none") != 0 && strcmp(value, "client") != 0 &&
               strcmp(value, "server") != 0 && strcmp(value, "both") != 0)
            {
#ifndef HAVE_OBE_NO_TRACE
                String_var err =
                    string_dup("ORB_init: illegal value for "
                               "ooc.orb.init_iiop: ");
                err += value;
                logger -> error(err);
#endif
                throw INITIALIZE();
            }
        }
        else if(strcmp(key, "ooc.orb.default_init_ref") == 0)
        {
#ifndef HAVE_OBE_NO_TRACE
            if(strlen(value) == 0)
                logger -> warning("ORB_init: invalid value for "
                                  "ooc.orb.default_init_ref");
#endif
        }
	else if(strcmp(key, "ooc.orb.server_shutdown_timeout") == 0 )
	{
	    // Used by GIOPServerWorker
	}
	else if(strcmp(key, "ooc.orb.client_shutdown_timeout") == 0)
	{
	    // Used by GIOPClientWorker
	}
	else if(strncmp(key, "ooc.orb.service.", 16) == 0)
	{
	    // Ignore
	}
	else if(strncmp(key, "ooc.orb.oa.", 11) == 0)
	{
	    // Ignore
	}
        else
        {
#ifndef HAVE_OBE_NO_TRACE
            String_var err = string_dup("ORB_init: unknown property: ");
            err += key;
            logger -> warning(err);
#endif
        }
    }

#ifndef HAVE_OBE_NO_TRACE
    //
    // Parse the tracing levels from the properties
    //
    CoreTraceLevels_var coreTraceLevels =
	new CoreTraceLevels(logger, properties);
#endif // !HAVE_OBE_NO_TRACE

    //
    // Is there an existing ORB implementation for this orb id?
    //
    // TODO: Is this correct? Why did we bother to parse all of those
    // command line options and stuff if we're only going to return
    // the existing ORB?
    //
    CORBA::ORB_var orb;
    OBCORBA::ORB_var oborb;
    if(CORBA::ORB::orbImplMap_.get(id, orb))
    {
	oborb = ORB::_narrow(orb);
	return oborb._retn();
    }

    //
    // Create new ORB
    //
    oborb = new ORB(ac, av, id,
#ifndef HAVE_OBE_NO_TRACE
		    coreTraceLevels,
#endif // !HAVE_OBE_NO_TRACE
		    properties, logger);

    return oborb._retn();
}

OBCORBA::ORB_ptr
OBCORBA::ORB_init(int& ac, char** av, Properties_ptr properties,
		  const char* orb_identifier, const char* version)
{
    return OBCORBA::ORB_init(ac, av, properties, Logger::_nil(),
			     orb_identifier, version);
			     
}

CORBA::ORB_ptr
CORBA::ORB_init(int& ac, char** av, const char* orb_identifier,
		const char* version)
{
    return OBCORBA::ORB_init(ac, av, Properties::_nil(),
			     Logger::_nil(), orb_identifier, version);
}
