// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/Object.h>
#include <OBE/NativeTypes.h>
#include <OBE/MarshalStubImpl.h>
#include <OBE/Policy.h>
#include <OBE/OBORB.h>
#include <OBE/OBPolicies.h>
#include <OBE/Stream.h>
#include <OBE/Client.h>
#include <OBE/ORBInstance.h>
#include <OBE/RefCounted.h>
#include <OBE/Logger.h>
#ifndef HAVE_OBE_NO_TRACE
#  include <OBE/CoreTraceLevels.h>
#endif // !HAVE_OBE_NO_TRACE

#include <CorbaURL.h>

using namespace CORBA;
using namespace OB;

//
// The ids_ for Object
//
const char* Object::ids_[] =
{
    //
    // Don't put "IDL:omg.org/CORBA/Object:1.0" here, as this ID is
    // always implictly there.
    //
    0
};

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(Object_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(Object_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

void
OBDuplicate(ObjectFactory_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(ObjectFactory_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Object private and protected members
// ----------------------------------------------------------------------

void
Object::checkRetry(bool retry, const SystemException& ex)
{
    //
    // We remove the StubImpl, whether we retry or not
    //
    stubImpl_ = MarshalStubImpl::_nil();

    //
    // Reset to the original IOR
    //
    IOR_ = origIOR_;

#ifndef HAVE_OBE_NO_TRACE
    //
    // Get the core trace levels
    //
    CoreTraceLevels_var coreTraceLevels = orbInstance_ -> getCoreTraceLevels();
#endif // !HAVE_OBE_NO_TRACE

    //
    // We only retry upon COMM_FAILURE, TRANSIENT, and NO_RESPONSE
    //
    if(dynamic_cast<const COMM_FAILURE*>(&ex) == 0 &&
       dynamic_cast<const TRANSIENT*>(&ex) == 0 &&
       dynamic_cast<const NO_RESPONSE*>(&ex) == 0)
    {
#ifndef HAVE_OBE_NO_TRACE
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"retry only upon COMM_FAILURE, TRANSIENT and "
		"NO_RESPONSE exceptions\n");
	    String_var exMsg = ex._rep_id();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
#endif // !HAVE_OBE_NO_TRACE
	ex._raise();
    }

    //
    // Only retry once
    //
    if(!retry)
    {
#ifndef HAVE_OBE_NO_TRACE
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"can't try again, because I tried again once already\n");
	    String_var exMsg = ex._rep_id();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
#endif // !HAVE_OBE_NO_TRACE
	ex._raise();
    }
    
    //
    // We can't retry if RETRY_NEVER is set
    //
    if(policies_ -> retry == RETRY_NEVER)
    {
#ifndef HAVE_OBE_NO_TRACE
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"can't try again because the RETRY_NEVER policy is set\n");
	    String_var exMsg = ex._rep_id();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
#endif // !HAVE_OBE_NO_TRACE
	ex._raise();
    }
    
    //
    // We can't retry if RETRY_STRICT is set and the completion status
    // is not COMPLETED_NO
    //
    if(
       policies_ -> retry == RETRY_STRICT &&
       ex.completed() != COMPLETED_NO)
    {
#ifndef HAVE_OBE_NO_TRACE
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"can't try again, because the RETRY_STRICT policy is set\n"
		"and completion status is not COMPLETED_NO\n");
	    String_var exMsg = ex._rep_id();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
#endif // !HAVE_OBE_NO_TRACE
	ex._raise();
    }
}

MarshalStubImpl_ptr
Object::_OB_getStubImpl_nodup()
{
    if(is_nil(stubImpl_))
    {
	MarshalStubImpl_var marshalStubImpl = _OB_createMarshalStubImpl();
	marshalStubImpl -> _OB_setup(orbInstance_, IOR_, origIOR_, policies_);
	stubImpl_ = MarshalStubImpl::_duplicate(marshalStubImpl);
	assert(!is_nil(stubImpl_));
    }
	    
    return stubImpl_.in();
}

MarshalStubImpl_ptr
Object::_OB_createMarshalStubImpl()
{
    return new MarshalStubImpl();
}

// ----------------------------------------------------------------------
// Object constructor and destructor
// ----------------------------------------------------------------------

Object::Object()
{
}

Object::~Object()
{
}

// ----------------------------------------------------------------------
// Object public member implementation
// ----------------------------------------------------------------------

Boolean
Object::_is_a(const char* _ob_id)
{
    assert_nca(_ob_id != 0, NCANullString);

    JTCSynchronized sync(_ob_mutex_);

    //
    // Check IDL:omg.org/CORBA/Object:1.0
    //
    if(strcmp(_ob_id, "IDL:omg.org/CORBA/Object:1.0") == 0)
        return true;
    
    //
    // Check all other ids
    //
    const char** ids = _OB_ids();
    assert(ids != 0);
    for(ULong i = 0; ids[i] != 0; i++)
        if(strcmp(_ob_id, ids[i]) == 0)
            return true;
    
    //
    // Check the type_id in the IOR and the original IOR
    //
    if(strcmp(_ob_id, IOR_ -> value.type_id) == 0 ||
       strcmp(_ob_id, origIOR_ -> value.type_id) == 0)
	return true;

    //
    // If this is an l-c object, return false now
    //
    if(_OB_isLC())
        return false;

    //
    // Execute _is_a remotely
    //
    bool _ob_retry = true;
    while(true)
    {
        MarshalStubImpl_ptr _ob_stubImpl = _OB_getStubImpl_nodup();
        try
        {
            return _ob_stubImpl -> _is_a(_ob_id);
        }
        catch(const ExceptionBase& _ob_ex)
        {
            _OB_handleException(_ob_ex, _ob_retry);
        }
    }
}

Boolean
Object::_non_existent()
{
    JTCSynchronized sync(_ob_mutex_);

    if(_OB_isLC())
        throw NO_IMPLEMENT();

    bool _ob_retry = true;
    while(true)
    {
        MarshalStubImpl_ptr _ob_stubImpl = _OB_getStubImpl_nodup();
        try
        {
            return _ob_stubImpl -> _non_existent();
        }
        catch(const ExceptionBase& _ob_ex)
        {
            _OB_handleException(_ob_ex, _ob_retry);
        }
    }
}

Object_ptr
Object::_set_policy_overrides(const PolicyList& np, SetOverrideType set_add)
{
    if(_OB_isLC())
        throw BAD_OPERATION();
    
    //
    // Create the new policy list
    //
    RefCountPolicyList_var newPolicies;

    if(set_add == SET_OVERRIDE)
    {
	newPolicies = new RefCountPolicyList(np);
    }
    else // ADD_OVERRIDE
    {
	PolicyList policies = policies_ -> value;
	
	for(ULong i = 0 ; i < np.length() ; i++)
	{
	    ULong j;
	    for(j = 0 ; j < policies.length() ; j++)
		if(policies[j] -> policy_type() == np[i] -> policy_type())
		    break;
	    
	    if(j < policies.length()) {
		policies[j] = np[i];
	    } else {
		CORBA::ULong policiesLength = policies.length();
		policies.length(policiesLength + 1);
		policies[policiesLength] = np[i];
            }
	}
	
	newPolicies = new RefCountPolicyList(policies);
    }

    //
    // Create a new reference
    //
    Object_var p = new Object;

    //
    // Copy the data from this object reference
    //
    p -> _OB_copyFrom(this);

    //
    // Setup the policies
    //
    p -> policies_ = newPolicies;

    return p._retn();
}

const char**
Object::_OB_ids() const
{
    return ids_;
}

const char**
Object::_OB_staticIds()
{
    return ids_;
}

void
Object::_OB_setup(ORBInstance_ptr instance, RefCountIOR_ptr ior,
		  RefCountPolicyList_ptr policies)
{
    if(_OB_isLC())
        throw BAD_OPERATION();

    orbInstance_ = ORBInstance::_duplicate(instance);
    IOR_ = RefCountIOR::_duplicate(ior);
    origIOR_ = RefCountIOR::_duplicate(ior);
    policies_ = RefCountPolicyList::_duplicate(policies);
}

void
Object::_OB_copyFrom(Object_ptr p)
{
    if(_OB_isLC())
        throw BAD_OPERATION();

    //
    // Synchronization on this -> _ob_mutex_ is not necessary, since
    // _OB_copyFrom is called right after reference creation, before
    // any other thread gets a chance to use it. We must, however,
    // synchronize on p -> _ob_mutex_, since the reference we're
    // copying from might be used by other threads.
    //

    JTCSynchronized sync(p -> _ob_mutex_);

    orbInstance_ = p -> orbInstance_;
    IOR_ = p -> IOR_;
    origIOR_ = p -> origIOR_;
    policies_ = p -> policies_;

    //
    // Don't copy stubImpl_, because it might be a StubImpl for a base
    // type, but not for the most derived type.
    //
}

ORBInstance_ptr
Object::_OB_ORBInstance() const
{
    return ORBInstance::_duplicate(orbInstance_);
}

RefCountIOR_ptr
Object::_OB_IOR() const
{
    JTCSynchronized sync(_ob_mutex_);

    return RefCountIOR::_duplicate(IOR_);
}

RefCountIOR_ptr
Object::_OB_origIOR() const
{
    //
    // ORBacus/E implements IIOP 1.0, which doesn't support
    // LOCATION_FORWARD_PERM, so origIOR_ is immutable, and no
    // synchronization is necessary here.
    //
    //JTCSynchronized sync(_ob_mutex_);

    return RefCountIOR::_duplicate(origIOR_);
}

void
Object::_OB_marshalOrigIOR(OutputStreamImpl* out) const
{
    //
    // ORBacus/E implements IIOP 1.0, which doesn't support
    // LOCATION_FORWARD_PERM, so origIOR_ is immutable, and no
    // synchronization is necessary here.
    //
    //JTCSynchronized sync(_ob_mutex_);

    origIOR_ -> value._OB_marshal(out);
}

void
Object::_OB_handleException(const ExceptionBase& ex, bool& retry)
{
    try
    {
        ex._raise();
    }
    catch(const LocationForward& ex)
    {
	//
	// Can't location forward l-c objects
	//
	assert(!_OB_isLC());
	
	//
	// Change the IOR
	//
	IOR_ = RefCountIOR::_duplicate(ex.ior);
	
	//
	// We need to re-get the stub implementation
	//
        stubImpl_ = MarshalStubImpl::_nil();

#ifndef HAVE_OBE_NO_TRACE
	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceRetry())
	{
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry",
			    "trying again because of location forward");
	}
#endif // !HAVE_OBE_NO_TRACE
    }
    catch(const FailureException& ex)
    {
	//
	// If it's not safe to retry, throw the exception
	//
	checkRetry(retry, *ex.exception);

	//
	// Only retry once
	//
	retry = false;

#ifndef HAVE_OBE_NO_TRACE
	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"trying again because of failure\n");
	    String_var exMsg = ex.exception -> _rep_id();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
#endif // !HAVE_OBE_NO_TRACE
    }
    //
    // We don't retry on non-locally raised exceptions
    //
/*
    catch(const TRANSIENT& ex)
    {
	//
	// If it's not safe to retry, throw the exception
	//
	checkRetry(retry, ex);

	//
	// Only retry once
	//
	retry = false;

#ifndef HAVE_OBE_NO_TRACE
	CoreTraceLevels_var coreTraceLevels =
	    orbInstance_ -> getCoreTraceLevels();
	if(coreTraceLevels -> traceRetry())
	{
	    String_var msg = string_dup(
		"trying again because server sent a TRANSIENT exception\n");
	    String_var exMsg = ex._rep_id();
	    msg += exMsg;
	    Logger_var logger = orbInstance_ -> getLogger();
	    logger -> trace("retry", msg);
	}
#endif // !HAVE_OBE_NO_TRACE
    }
*/
}

// ----------------------------------------------------------------------
// ObjectFactory constructor and destructor
// ----------------------------------------------------------------------

ObjectFactory::ObjectFactory(ORBInstance_ptr instance)
    : destroy_(false),
      orbInstance_(ORBInstance::_duplicate(instance))
{
}

ObjectFactory::~ObjectFactory()
{
    assert(destroy_);
}

// ----------------------------------------------------------------------
// ObjectFactory public member implementation
// ----------------------------------------------------------------------

void
ObjectFactory::destroy()
{
    assert(!destroy_); // May only be destroyed once
    destroy_ = true;

    //
    // Set the ORBInstance object to nil
    //
    orbInstance_ = ORBInstance::_nil();

    //
    // Set the PolicyManager object to nil
    //
    policyManager_ = PolicyManager::_nil();
}

void
ObjectFactory::setPolicyManager(PolicyManager_ptr policyManager)
{
    policyManager_ = PolicyManager::_duplicate(policyManager);
}

Object_ptr
ObjectFactory::createObject(RefCountIOR_ptr ior)
{
    //
    // The ORB destroys this object, so it's an initialization error
    // if the this operation is called after ORB destruction
    //
    if(destroy_)
	throw INITIALIZE(MinorORBDestroyed, COMPLETED_NO);

    //
    // Check for nil object reference
    //
    if(strlen(ior -> value.type_id) == 0 &&
       ior -> value.profiles.length() == 0)
	return Object::_nil();

    //
    // Create a new CORBA::Object
    //
    Object_var p = new Object();
    RefCountPolicyList_var policyList = policies();
    p -> _OB_setup(orbInstance_, ior, policyList);

    return p._retn();
}

Object_ptr
ObjectFactory::stringToObject(const char* s)
{
    CorbaURL url(s);
    return url.toObject(orbInstance_);
}

RefCountPolicyList_ptr
ObjectFactory::policies()
{
    PolicyList_var policies = policyManager_ -> get_all_policy_overrides();
    return new RefCountPolicyList(policies);
}
