// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/NativeTypes.h>
#include <OBE/POAServantBase.h>
#include <OBE/Object.h> // For Logger
#include <OBE/Logger.h>
#include <OBE/Stream.h>
#include <OBE/ORBInstance.h>
#include <OBE/InitialServiceManager.h>
#include <OBE/RefCounted.h>
#include <OBE/Upcall.h>
#include <OBE/OBPOAManager.h>
#include <OBE/ObjectKey.h>
#include <OBE/POA.h>

#include <ORBInstanceS.h>

using namespace CORBA;
using namespace OB;

// ----------------------------------------------------------------------
// ActiveObjectEntry constructor
// ----------------------------------------------------------------------

PortableServer::ActiveObjectEntry::ActiveObjectEntry()
    : iter(0)
{
}

// ----------------------------------------------------------------------
// External, non-inline functions
// ----------------------------------------------------------------------

void
OBDuplicate(PortableServer::POA_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(PortableServer::POA_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// ObjectAlreadyActive public members
// ----------------------------------------------------------------------

const char*
PortableServer::POA::ObjectAlreadyActive::_rep_id() const
{
    return "IDL:omg.org/PortableServer/POA/ObjectAlreadyActive:1.0";
}

CORBA::Exception*
PortableServer::POA::ObjectAlreadyActive::_OB_clone() const
{
    return new ObjectAlreadyActive(*this);
}

// ----------------------------------------------------------------------
// ObjectNotActive public members
// ----------------------------------------------------------------------

const char*
PortableServer::POA::ObjectNotActive::_rep_id() const
{
    return "IDL:omg.org/PortableServer/POA/ObjectNotActive:1.0";
}

CORBA::Exception*
PortableServer::POA::ObjectNotActive::_OB_clone() const
{
    return new ObjectNotActive(*this);
}

// ----------------------------------------------------------------------
// ServantNotActive public members
// ----------------------------------------------------------------------

const char*
PortableServer::POA::ServantNotActive::_rep_id() const
{
    return "IDL:omg.org/PortableServer/POA/ServantNotActive:1.0";
}

CORBA::Exception*
PortableServer::POA::ServantNotActive::_OB_clone() const
{
    return new ServantNotActive(*this);
}

// ----------------------------------------------------------------------
// WrongAdapter public members
// ----------------------------------------------------------------------

const char*
PortableServer::POA::WrongAdapter::_rep_id() const
{
    return "IDL:omg.org/PortableServer/POA/WrongAdapter:1.0";
}

CORBA::Exception*
PortableServer::POA::WrongAdapter::_OB_clone() const
{
    return new WrongAdapter(*this);
}

// ----------------------------------------------------------------------
// POA_impl private member implementation
// ----------------------------------------------------------------------

void
PortableServer::POA::handleUnknownException(UpcallImpl* upcall)
{
#ifndef HAVE_OBE_NO_TRACE
    String_var msg = string_dup("Servant method raised a non-CORBA exception");
            
    if(upcall -> responseExpected())
        msg += "\nClient receives this exception as CORBA::UNKNOWN";
            
    msg += "\noperation name: \"";
    msg += upcall -> operation();
    msg += '"';
 
    Logger_var logger = orbInstance_ -> getLogger();
    logger -> warning(msg);
#endif
            
    upcall -> setSystemException(UNKNOWN());
}

PortableServer::ServantBase*
PortableServer::POA::idToServant(const PortableServer::ObjectId& oid)
{
    ObjectKeyData keydata;
    if(!ObjectIdToObjectKey(oid, keydata))
        return 0;

    JTCSynchronized sync(activeObjectArray_[keydata.slot].mutex);

    if(activeObjectArray_[keydata.slot].iter == keydata.iter &&
       activeObjectArray_[keydata.slot].servant.in() != 0)
    {
        activeObjectArray_[keydata.slot].servant -> _add_ref();
	return activeObjectArray_[keydata.slot].servant;
    }

    return 0;
}

PortableServer::ServantBase*
PortableServer::POA::locate(const ObjectKeyData& key)
    throw(SystemException, LocationForward)
{
    if(key.slot >= MAX_ACTIVE_OBJECTS)
	throw OBJECT_NOT_EXIST();

    JTCSynchronized sync(activeObjectArray_[key.slot].mutex);

    if(key.slot >= MAX_ACTIVE_OBJECTS ||
       activeObjectArray_[key.slot].iter != key.iter ||
       activeObjectArray_[key.slot].servant.in() == 0)
        throw OBJECT_NOT_EXIST();

    activeObjectArray_[key.slot].servant -> _add_ref();
    return activeObjectArray_[key.slot].servant;
}

RefCountIOR_ptr
PortableServer::POA::createIOR(const OCI::ObjectKey& key, const char* intf)
{
    IOP::IOR ior;
    ior.type_id = intf;
    OCI::AcceptorSeq_var acceptors = manager_ -> _OB_getAcceptors();
    OCI::ProfileInfo profileInfo;
    profileInfo.major = 1;
    profileInfo.minor = 0;
    profileInfo.key = key;

    for(ULong i = 0 ; i < acceptors -> length() ; i++)
        acceptors[i] -> add_profiles(profileInfo, ior);

    return new RefCountIOR(ior);
}

bool 
PortableServer::POA::ObjectIdToObjectKey(const PortableServer::ObjectId& oid,
					 ObjectKeyData& key)
{
    //
    // If the ObjectId is not 8 bytes, it is not valid
    //
    if(oid.length() != 8)
	return false;

    key.iter = *(ULong*)oid.get_buffer();
    key.slot = *(ULong*)(oid.get_buffer() + 4);

    //
    // If the slot is greater than the maximum number the ObjectId
    // is not valid
    //
    if(key.slot >= MAX_ACTIVE_OBJECTS)
	return false;

    return true;
}

PortableServer::ObjectId*
PortableServer::POA::ObjectKeyToObjectId(const OB::ObjectKeyData& key)
{
    PortableServer::ObjectId_var oid = new PortableServer::ObjectId(8);
    oid -> length(8);

    memcpy(oid -> get_buffer(), &key.iter, 4);
    memcpy(oid -> get_buffer() + 4, &key.slot, 4);

    return oid._retn();
}

PortableServer::ObjectId*
PortableServer::POA::activate(PortableServer::Servant servant)
{
    //
    // Get a slot in the AOM
    //
    ULong index;
    if(freeSlots_)
    {
        for(index = 0 ; index < nextSlot_ ; ++index)
        {
            if(activeObjectArray_[index].servant.in() == 0)
    	    {
    	        --freeSlots_;
    	        break;
    	    }
        }
    }
    else
    {
        if(nextSlot_ >= MAX_ACTIVE_OBJECTS)
            throw IMP_LIMIT();
        index = nextSlot_++;
    }

    //
    // Set the array entry
    //
    JTCSynchronized sync2(activeObjectArray_[index].mutex);

    activeObjectArray_[index].servant = servant;
    activeObjectArray_[index].servant -> _add_ref();

    //
    // Create the ObjectId
    //
    ObjectKeyData key(activeObjectArray_[index].iter, index);
    return ObjectKeyToObjectId(key);
}

// ----------------------------------------------------------------------
// POA constructor and destructor
// ----------------------------------------------------------------------

//
// Constructor for the POA
//
PortableServer::POA::POA(ORBInstance_ptr orbInstance,
			 PortableServer::POAManager_ptr manager)
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      manager_(PortableServer::POAManager::_duplicate(manager)),
      nextSlot_(0),
      freeSlots_(0)
{
}

PortableServer::POA::~POA()
{
}

// ----------------------------------------------------------------------
// POA public member implementation
// ----------------------------------------------------------------------

PortableServer::POA_ptr
PortableServer::POA::_narrow(CORBA::Object_ptr p)
{
    if(!CORBA::is_nil(p))
    {
        POA_ptr v = dynamic_cast<POA_ptr>(p);
        if(v)
            return _duplicate(v);
    }

    return _nil();
}

void
PortableServer::POA::_OB_destroy()
{
    //
    // Clear POA Manager
    //
    manager_ = PortableServer::POAManager::_nil();

    //
    // Clear the Active Object Array
    //
    //
    // TODO: should the table be locked?
    //
    for(ULong i = 0; i < nextSlot_; ++i)
	activeObjectArray_[i].servant = 0;
}

PortableServer::POAManager_ptr
PortableServer::POA::the_POAManager()
    throw(SystemException)
{
    return PortableServer::POAManager::_duplicate(manager_);
}

PortableServer::ObjectId*
PortableServer::POA::activate_object(PortableServer::Servant servant)
    throw(SystemException)
{
    assert_nca(servant != 0, NCANullValue);

    JTCSynchronized sync(AOMMutex_);

    return activate(servant);
}

void
PortableServer::POA::deactivate_object(const PortableServer::ObjectId& oid)
    throw(PortableServer::POA::ObjectNotActive, SystemException)
{
    //
    // If no object in the active object table associated with this
    // key then raise an ObjectNotActive exception.
    //
    ObjectKeyData keydata;
    if(!ObjectIdToObjectKey(oid, keydata))
        throw PortableServer::POA::ObjectNotActive();

    JTCSynchronized sync(activeObjectArray_[keydata.slot].mutex);

    if(activeObjectArray_[keydata.slot].iter == keydata.iter &&
       activeObjectArray_[keydata.slot].servant.in() != 0)
    {
	//
	// This is safe to do since activate_object() will not lock
	// a slot that is used and deactivate_object() will not lock 
	// the AOM if it has a lock on a slot that is not used. Thus
	// no deadlocks should occur.
	//
	JTCSynchronized sync2(AOMMutex_);

        activeObjectArray_[keydata.slot].servant = 0;
	++activeObjectArray_[keydata.slot].iter;

	++freeSlots_;
    }
    else
	throw PortableServer::POA::ObjectNotActive();
}

PortableServer::ObjectId*
PortableServer::POA::servant_to_id(PortableServer::Servant servant)
    throw(PortableServer::POA::ServantNotActive, SystemException)
{
    assert_nca(servant != 0, NCANullValue);

    //
    // Lock the entire map so no object will be activated/deactivated
    //
    JTCSynchronized sync(AOMMutex_);

    for(ULong i = 0 ; i < nextSlot_ ; ++i)
        if(activeObjectArray_[i].servant.in() == servant)
            return ObjectKeyToObjectId(
	        ObjectKeyData(activeObjectArray_[i].iter, i));

    return activate(servant);
}

Object_ptr
PortableServer::POA::servant_to_reference(PortableServer::Servant servant)
    throw(PortableServer::POA::ServantNotActive, SystemException)
{
    assert_nca(servant != 0, NCANullValue);

    PortableServer::ObjectId_var oid = servant_to_id(servant);
    String_var intf = servant -> _primary_interface(oid, this);
    return _OB_createReference(oid, intf);
}

PortableServer::Servant
PortableServer::POA::reference_to_servant(Object_ptr reference)
    throw(PortableServer::POA::ObjectNotActive,
	  PortableServer::POA::WrongAdapter, SystemException)
{
    assert_nca(reference != 0, NCANullValue);

    PortableServer::ObjectId_var oid = reference_to_id(reference);
    PortableServer::ServantBase_var servant = idToServant(oid);

    if(!servant)
	throw POA::ObjectNotActive();

    return servant._retn();
}

PortableServer::ObjectId*
PortableServer::POA::reference_to_id(Object_ptr reference)
    throw(PortableServer::POA::WrongAdapter, SystemException)
{
    assert_nca(reference != 0, NCANullValue);

    RefCountIOR_var ior = reference -> _OB_IOR();

    //
    // Extract the object key from the IOR of the object reference.
    //
    OCI::AcceptorSeq_var acceptors = manager_ -> _OB_getAcceptors();    
    bool local = false;
    ULong i;

    OCI::ProfileInfoSeq_var profileInfos;
    for(i = 0 ; i < acceptors -> length() ; i++)
    {
	profileInfos = acceptors[i] -> get_local_profiles(ior -> value);
	if(profileInfos -> length() > 0)
	    local = true;
    }

    //
    // Is the key local?
    //
    if(local)
    {
	ObjectKeyData keyData;
	if(ParseObjectKey(profileInfos[0].key, keyData))
	    return ObjectKeyToObjectId(keyData);

    }

    throw POA::WrongAdapter();
    return 0; // Some compilers need this
}

PortableServer::Servant
PortableServer::POA::id_to_servant(const PortableServer::ObjectId& oid)
    throw(PortableServer::POA::ObjectNotActive, SystemException)
{
    PortableServer::ServantBase_var servant = idToServant(oid);

    if(!servant)
        throw POA::ObjectNotActive();

    return servant._retn();
}

Object_ptr
PortableServer::POA::id_to_reference(const PortableServer::ObjectId& oid)
    throw(PortableServer::POA::ObjectNotActive, SystemException)
{
    PortableServer::ServantBase_var servant = idToServant(oid);
    if(!servant)
        throw POA::ObjectNotActive();

    String_var intf = servant -> _primary_interface(oid, this);
    return _OB_createReference(oid, intf);
}

bool
PortableServer::POA::_OB_locateServant(const ObjectKeyData& key)
    throw(SystemException, LocationForward)
{
    PortableServer::ServantBase_var servant = locate(key);

    return servant.in() != 0;
}

void
PortableServer::POA::_OB_dispatch(const ObjectKeyData& key, UpcallImpl* upcall)
    throw()
{
    try
    {
        //
        // Locating a servant can throw LocationForward and
        // SystemException errors
        //
        PortableServer::ServantBase_var servant = locate(key);
    
        //
        // If there is a servant then dispatch the request
        //
        if(servant.in() != 0)
        {
            //
            // Dispatch the request
            //
            servant -> _OB_dispatch(upcall);
        }
        else
        {
            upcall -> setSystemException(OBJECT_NOT_EXIST());
        }
    }
    catch(const LocationForward& ex)
    {
	upcall -> setLocationForward(ex.ior);
    }
    catch(const UserException& ex)
    {
	upcall -> setUserException(ex);
    }
    catch(const SystemException& ex)
    {
	upcall -> setSystemException(ex);
    }
    catch(...)
    {
        handleUnknownException(upcall);
    }
}

Object_ptr
PortableServer::POA::_OB_createReference(
    const PortableServer::ObjectId& oid, const char* intf)
{
    //
    // Create the object key for this object id
    //
    ObjectKeyData keydata;
    ObjectIdToObjectKey(oid, keydata);
    OCI::ObjectKey_var key = CreateObjectKey(keydata);

    //
    // Create local IOR
    //
    RefCountIOR_var ior = createIOR(key, intf);

    //
    // Create and return reference
    //
    ObjectFactory_var objectFactory = orbInstance_ -> getObjectFactory();
    return objectFactory -> createObject(ior);
}

