// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/MarshalStubImpl.h>
#include <OBE/Upcall.h>
#include <OBE/Server.h>
#include <OBE/ORBInstance.h>
#include <OBE/InitialServiceManager.h>
#include <OBE/Logger.h>
#include <OBE/Properties.h>
#include <OBE/RefCounted.h>
#include <OBE/IIOPConfig.h>
#include <OBE/OAInterface.h>
#include <OBE/POA.h>
#include <OBE/OBPOAManager.h>

#include <OCI_impl.h>
#include <OCI_IIOP_impl.h>
#include <POAOAInterface_impl.h>
#include <ORBInstanceS.h>

using namespace CORBA;
using namespace OB;

void
OBDuplicate(PortableServer::POAManager_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(PortableServer::POAManager_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

PortableServer::POAManager_ptr
PortableServer::POAManager::_narrow(CORBA::Object_ptr p)
{
    if(!CORBA::is_nil(p))
    {
        POAManager_ptr v = dynamic_cast<POAManager_ptr>(p);
        if(v)
            return _duplicate(v);
    }

    return _nil();
}

const char* OBPortableServer::POAManager::ids_[] =
{
    "IDL:ooc.com/OBPortableServer/POAManager:1.0",
    "IDL:omg.org/PortableServer/POAManager:2.3",
    0
};

void
OBDuplicate(OBPortableServer::POAManager_ptr p)
{
    if(p)
        p -> _OB_incRef();
}

void
OBRelease(OBPortableServer::POAManager_ptr p)
{
    if(p)
        p -> _OB_decRef();
}

OBPortableServer::POAManager_ptr
OBPortableServer::POAManager::_narrow(CORBA::Object_ptr p)
{
    if(!CORBA::is_nil(p))
    {
        POAManager_ptr v = dynamic_cast<POAManager_ptr>(p);
        if(v)
            return _duplicate(v);
    }

    return _nil();
}

// ----------------------------------------------------------------------
// POAManager constructor and destructor
// ----------------------------------------------------------------------

PortableServer::POAManager::POAManager(ORBInstance_ptr orbInstance)
    throw(SystemException)
    : orbInstance_(ORBInstance::_duplicate(orbInstance)),
      state_(HOLDING)
{
    try
    {
	//
	// Get the Logger
	//
	Logger_var logger = orbInstance_ -> getLogger();

        //
        // Get the ORB properties
        //
        Properties_var properties = orbInstance -> getProperties();

	//
	// Get the IIOP acceptor configuration parameters and
	// create an IIOP acceptor
	//
	IIOPConfig cfg(properties, logger);
	cfg.parseAll();
	OCI::ParamSeq_var params = cfg.getParams();
	OCI::Acceptor_var acceptor;
	try
	{
	    ORBInstanceS* orbInstance =
		dynamic_cast<ORBInstanceS*>(orbInstance_.in());
	    assert(orbInstance);

	    OCI::AccFactoryRegistry_var registry =
		orbInstance -> getAccFactoryRegistry();
	    OCI::AccFactory_var factory =
		registry -> get_factory(OCI::IIOP::TAG_IIOP);
	    acceptor = factory -> create_acceptor(params);
	}
	catch(const OCI::NoSuchFactory&)
	{
#ifndef HAVE_OBE_NO_TRACE
	    logger -> error("unable to create IIOP acceptor");
#endif
	    throw INITIALIZE();
	}
	catch(const OCI::InvalidParam& ex)
	{     
#ifndef HAVE_OBE_NO_TRACE
	    String_var err = string_dup("invalid configuration parameter `");
	    err += ex.p.name;
	    err += "' for RootPOAManager: ";
	    err += ex.reason;
	    logger -> error(err);
#endif
	    throw INITIALIZE(); 
	}

	acceptors_.length(1);
	acceptors_[0] = acceptor;

        //
        // Create the OAInterface
        //
        oaInterface_ = new POAOAInterface_impl(this, orbInstance_);

        //
        // Create the server manager
        //
        serverManager_ = new ServerManager(orbInstance_,
					   acceptors_, 
					   oaInterface_);
    }
    catch(...)
    {
        _OB_setRef(0);
        throw;
    }
}

PortableServer::POAManager::~POAManager()
{
}

// ----------------------------------------------------------------------
// POAManager public member implementation
// ----------------------------------------------------------------------

//
// Mapping for PortableServer::POAManager
//
void
PortableServer::POAManager::activate()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);;

    if(state_ != ACTIVE)
    {
        state_ = ACTIVE;

	//
	// Activate the server manager
	//
	serverManager_ -> activate();

	//
	// Tell the OAInterface to accept requests
	//
	oaInterface_ -> activate();
    }
}

//
// Internal methods.
//

//
// Create the root POA
//
void
PortableServer::POAManager::_OB_createRootPOA()
    throw()
{
    if(is_nil(poa_))
    {
	//
	// Create the Root POA
	//
	poa_ = new PortableServer::POA(orbInstance_, this);

	//
	// Add the root POA to the list of initial services
	//
	InitialServiceManager_var initServiceManager =
	    orbInstance_ -> getInitialServiceManager();
	initServiceManager -> addInitialReference("RootPOA", poa_, true);
    }
}

PortableServer::POA_ptr
PortableServer::POAManager::_OB_POA_nodup()
    throw(SystemException)
{
    //
    // If the GIOP engine sends a request while the POAManager is in
    // INACTIVE state, then something is wrong.
    //
    assert(state_ != INACTIVE);

    return poa_;
}

#ifndef HAVE_OBE_NO_COLLOC
CollocatedServer_ptr
PortableServer::POAManager::_OB_getCollocatedServer()
    throw()
{
    return serverManager_ -> getCollocatedServer();
}
#endif // !HAVE_OBE_NO_COLLOC

//
// TODO: waitCompletion?
//
void
PortableServer::POAManager::_OB_deactivate()
    throw(SystemException)
{
    JTCSynchronized synchronized(*this);
    
    //
    // If the POA manager is in inactive state then return
    //
    if(state_ == INACTIVE)
	return;

    //
    // Destroy the server manager
    //
    serverManager_ -> destroy();

    //
    // Set the state to INACTIVE *after* the serverManager_ has
    // been destroyed, to avoid a race condition.
    //
    state_ = INACTIVE;
}

void
PortableServer::POAManager::_OB_destroy()
    throw(SystemException)
{
    //
    // Remove the references to the orbInstance_
    //
    orbInstance_ = ORBInstance::_nil();

    //
    // Destroy all the POAs
    //
    if(!is_nil(poa_))
        poa_ -> _OB_destroy();

    //
    // Shutdown, if necessary
    //
    _OB_deactivate();
}

OCI::AcceptorSeq*
PortableServer::POAManager::_OB_getAcceptors()
     throw(SystemException)
{
    JTCSynchronized synchronized(*this);

    return new OCI::AcceptorSeq(acceptors_);
}

// ----------------------------------------------------------------------
// OBPortableServer::POAManager constructor and destructor
// ----------------------------------------------------------------------

OBPortableServer::POAManager::POAManager(ORBInstance_ptr orbInstance)
    throw(SystemException)
    : PortableServer::POAManager(orbInstance)
{
}

OBPortableServer::POAManager::~POAManager()
{
}

//
// Mapping for OBPortableServer::POAManager
//

OCI::AcceptorSeq*
OBPortableServer::POAManager::get_acceptors()
     throw(SystemException)
{
  return _OB_getAcceptors();
}

