// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/Except.h>
#include <OBE/Properties.h>

#include <stdlib.h>

using namespace CORBA;
using namespace OB;

JTCMutex Properties::propertiesMutex_;
Properties_var Properties::defaultProperties_;

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(OB::Properties_ptr p)
{
    if(p)
	p -> _OB_incRef();
}

void
OBRelease(OB::Properties_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

// ----------------------------------------------------------------------
// Properties private members
// ----------------------------------------------------------------------

//
// Merge of two string sequences
//
void
Properties::merge(KeySeq& k1, const KeySeq& k2) const
{
    //
    // Pretty inefficient
    //
    for(ULong i = 0 ; i < k2.length() ; i++)
    {
	ULong j;
	for(j = 0 ; j < k1.length() ; j++)
	    if(strcmp(k2[i], k1[j]) == 0)
		break;
	if(j >= k1.length()) {
	    CORBA::ULong k1Length = k1.length();
	    k1.length(k1Length + 1);
	    k1[k1Length] = k2[i];
        }
    }
}

// ----------------------------------------------------------------------
// Properties constructor and destructor
// ----------------------------------------------------------------------

Properties::Properties()
{
}

Properties::Properties(OB::Properties_ptr defaults)
    : defaults_(OB::Properties::_duplicate(defaults))
{
}

Properties::~Properties()
{
}

// ----------------------------------------------------------------------
// Properties public member implementation
// ----------------------------------------------------------------------

//
// Return the default properties
//
Properties_ptr
Properties::getDefaultProperties()
{
    JTCSynchronized sync(propertiesMutex_);

    //
    // Have the defaultProperties_ been initialized?
    //
    if(!defaultProperties_)
    {
	//
	// Allocate the default properties
	//
	defaultProperties_ = new Properties();

	//
	// Load properties from the registry. First from
	// HKEY_LOCAL_MACHINE, next from HKEY_CURRENT_USER.
	//
#ifdef WIN32
	const char* base = "Software\\OOC\\Properties";
	defaultProperties_ -> load(HKEY_LOCAL_MACHINE, base, "");
	defaultProperties_ -> load(HKEY_CURRENT_USER, base, "");
#endif

	//
	// Load the contents of the ORBACUS_CONFIG environment
	// variable
	//
	const char* configFile = getenv("ORBACUS_CONFIG");
	if(configFile != 0 && *configFile != '\0')
	    defaultProperties_ -> load(configFile);
    }

    return Properties::_duplicate(defaultProperties_);
}

//
// Set a property value
//
void
Properties::setProperty(const char* key, const char* value)
{
    JTCSynchronized synchronized(*this);

    for(ULong i = 0 ; i < props_.length() ; i += 2)
        if(strcmp(props_[i], key) == 0)
        {
            props_[i+1] = string_dup(value);
            return;
        }

    CORBA::ULong propsLength = props_.length();
    props_.length(propsLength + 1);
    props_[propsLength] = key;

    propsLength = props_.length();
    props_.length(propsLength + 1);
    props_[propsLength] = value;
}

//
// Get a property value, 0 if no such property exists
//
const char*
Properties::getProperty(const char* key) const
{
    JTCSynchronized synchronized(*this);

    for(ULong i = 0 ; i < props_.length() ; i += 2)
        if(strcmp(props_[i], key) == 0)
            return props_[i+1];

    if(!is_nil(defaults_))
	return defaults_ -> getProperty(key);

    return 0;
}

Properties::KeySeq
Properties::getKeys(const char* prefix) const
{
    JTCSynchronized synchronized(*this);

    KeySeq keys;
    int prefLen = strlen(prefix);
    for(ULong i = 0 ; i < props_.length() ; i += 2)
        if(strncmp(props_[i], prefix, prefLen) == 0) {
            CORBA::ULong keysLength = keys.length();
            keys.length(keysLength + 1);
            keys[keysLength] = props_[i];
        }

    if(!is_nil(defaults_))
    {
	KeySeq defaultKeys = defaults_ -> getKeys(prefix);
	merge(keys, defaultKeys);
    }

    return keys;
}

Properties::KeySeq
Properties::getKeys() const
{
    JTCSynchronized synchronized(*this);

    KeySeq keys;
    for(ULong i = 0 ; i < props_.length() ; i += 2) {
        CORBA::ULong keysLength = keys.length();
        keys.length(keysLength + 1);
        keys[keysLength] = props_[i];
    }

    if(!is_nil(defaults_))
    {
	KeySeq defaultKeys = defaults_ -> getKeys();
	merge(keys, defaultKeys);
    }
    
    return keys;
}

bool
Properties::load(const char* name)
{
    JTCSynchronized synchronized(*this);

    assert_nca(name != 0, NCANullString);

    FILE* f = fopen(name, "r");
    if(f == 0)
	return false;

    int ch = fgetc(f);

    String_var key = string_alloc(1024);
    String_var value = string_alloc(1024);

    while(ch != EOF)
    {
        key[0] = '\0';
        value[0] = '\0';

        //
        // Find start of key
        //
        switch(ch)
        {
        case '#':
        case '!':
            do
            {
		ch = fgetc(f);
            }
            while(ch != EOF && ch != '\n' && ch != '\r');
            continue;
            
        case '\n':
        case '\r':
        case ' ':
        case '\t':
	    ch = fgetc(f);
            continue;
        }
        
        //
        // Read the key
        //
        while((ch != EOF) &&
              (ch != '=') && (ch != ':') && 
              (ch != ' ') && (ch != '\t') &&
              (ch != '\n') && (ch != '\r'))
        {
            key += (char)ch;
	    ch = fgetc(f);
        }
        while(ch != EOF && ((ch == ' ') || (ch == '\t')))
            ch = fgetc(f);
        if(ch != EOF && ((ch == '=') || (ch == ':')))
            ch = fgetc(f);
        while(ch != EOF && ((ch == ' ') || (ch == '\t')))
            ch = fgetc(f);

        //
        // Read the value
        //
        while(ch != EOF && ((ch != '\n') && (ch != '\r')))
        {
            int next = 0;
            if(ch == '\\')
            {
                if((ch = fgetc(f)) == EOF)
                    continue;
                switch (ch)
                {
                case '\r':
                    ch = fgetc(f);
                    if(ch != EOF && ((ch == '\n') ||
		       (ch == ' ') || (ch == '\t')))
                    {
                        // fall thru to '\n' case
                    }
                    else
                        continue;
                    
                case '\n':
                    ch = fgetc(f);
                    while(ch != EOF && ((ch == ' ') || (ch == '\t')))
                        ch = fgetc(f);
                    continue;
                    
                case 't':
                    ch = '\t';
                    next = fgetc(f);
                    break;
                    
                case 'n':
                    ch = '\n';
                    next = fgetc(f);
                    break;
                    
                case 'r':
                    ch = '\r';
                    next = fgetc(f);
                    break;
                    
                case 'u':
                    {
                        while(ch != EOF && ch == 'u')
                            ch = fgetc(f);
			int d = 0;
                        bool stop = false;
                        for(int i = 0 ; i < 4 && !stop; i++)
                        {
                            if((ch = fgetc(f)) == EOF)
                                break;
                            switch (ch)
                            {
                            case '0': case '1': case '2': case '3': case '4':
                            case '5': case '6': case '7': case '8': case '9':
                                d = (d << 4) + ch - '0';
                                break;
                            case 'a': case 'b': case 'c': case 'd': case 'e':
                            case 'f':
                                d = (d << 4) + 10 + ch - 'a';
                                break;
                            case 'A': case 'B': case 'C': case 'D': case 'E':
                            case 'F':
                                d = (d << 4) + 10 + ch - 'A';
                                break;
                            default:
                                stop = true;
                            }
                            if(!stop)
                                ch = next;
                        }
			ch = d;
			break;
                    }
                
                default:
                    next = fgetc(f);
                    break;
                }
            }
            else
            {
                next = fgetc(f);
            }
            value += (char)ch;
            ch = next;
        }
        setProperty(key, value);
    }

    fclose(f);

    return true;
}

#ifdef WIN32
//
// Add all values in this registry key, and all subkeys.
//
bool
Properties::load(HKEY base, const char* path, const char* prefix)
{
    JTCSynchronized synchronized(*this);

    HKEY key;

    //
    // Open the registry.
    //
    DWORD rc = RegOpenKeyEx(base, path, 0, KEY_READ, &key);

    //
    // If the key doesn't exist, then return.
    //
    if(rc != ERROR_SUCCESS)
	return false;

    // Buffer for class name.
    char className[MAX_PATH] = "";
    // Length of class string.
    DWORD szClassLen = MAX_PATH;
    // Number of sub keys.
    DWORD numSubKeys;
    // Longest sub key size.
    DWORD numMaxSubKey;
    // Longest class string.
    DWORD numMaxClass;
    // Number of values for this key.
    DWORD numValues;
    // Longest Value name.
    DWORD  maxValueName;
    // Longest Value data.
    DWORD  maxValueData;
    // Security descriptor.
    DWORD  secDesc;
    // Last write time.
    FILETIME lastWriteTime;

    //
    // Query registry information.
    //
    RegQueryInfoKey(key,               // Key handle.
		    className,         // Buffer for class name.
		    &szClassLen,       // Length of class string.
		    NULL,              // Reserved.
		    &numSubKeys,       // Number of sub keys.
		    &numMaxSubKey,     // Longest sub key size.
		    &numMaxClass,      // Longest class string.
		    &numValues,        // Number of values for this key.
		    &maxValueName,     // Longest Value name.
		    &maxValueData,     // Longest Value data.
		    &secDesc,          // Security descriptor.
		    &lastWriteTime);   // Last write time.

    //
    // We need to allocate one more byte for storage.
    //
    ++maxValueName;
    ++maxValueData;

    int i;
    DWORD szValueName;
    char* valueName = new char[maxValueName];
    DWORD szData;
    unsigned char* data = new unsigned char[maxValueData];
    DWORD type;

    //
    // Enumerate each value in this registry key.
    //
    for(i = 0, rc = ERROR_SUCCESS; rc == ERROR_SUCCESS; i++)
    {
	szValueName = maxValueName;
	valueName[0] = '\0';
	szData = maxValueData;
	rc = RegEnumValue(key, i, valueName,
			  &szValueName, 0,
			  &type, data, &szData);

	//
	// Did we get a valid key?
	//
	if(rc == ERROR_SUCCESS)
	{
	    //
	    // The key should be a string, otherwise it's ignored
	    //
	    if(type == REG_SZ)
	    {
		//
		// Construct the property name.
		//
		char name[MAX_PATH];
		strcpy(name, prefix);
		strcat(name, ".");
		strcat(name, valueName);

		//
		// Set the property.
		//
		setProperty(name, (const char*)data);
	    }
	}
    }
    delete[] valueName;
    delete[] data;

    //
    // Recurse down each set of subkeys.
    //
    char name[MAX_PATH];
    for(i = 0, rc = ERROR_SUCCESS; rc == ERROR_SUCCESS; i++)
    {
	rc = RegEnumKey(key, i, name, MAX_PATH);
	if(rc == ERROR_SUCCESS)
	{
	    //
	    // Construct the subkey path name.
	    //
	    char buf[MAX_PATH];
	    strcpy(buf, path);
	    strcat(buf, "\\");
	    strcat(buf, name);
	    
	    //
	    // Construct the subkey property prefix.
	    //
	    char pref[MAX_PATH];
	    if(*prefix != '\0')
	    {
		strcpy(pref, prefix);
		strcat(pref, ".");
	    }
	    else
	    {
		*pref = '\0';
	    }
	    strcat(pref, name);

	    //
	    // Add all values contained in this subkey.
	    //
	    if(!load(base, buf, pref))
		return false;
	}
    }

    //
    // Close the registry key.
    //
    RegCloseKey(key);

    return true;
}
#endif
