// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/TypeCode.h>
#include <OBE/CompletionStatus.h>
#include <OBE/Stream.h>
#include <OBE/list.h>

using namespace CORBA;
using namespace OB;

#ifdef OB_TYPECODE_DEBUG
static ULong TCCount = 0;
#endif

static JTCRecursiveMutex* GCMutex = new JTCRecursiveMutex;

class GCMutexDestroyer
{
public:

    ~GCMutexDestroyer()
    {
	delete GCMutex;
	GCMutex = 0;
    }
};

static GCMutexDestroyer destroyer;

namespace OB
{

//
// The TypeCodeFactory class
//
class TypeCodeFactory
{
public:

    static CORBA::TypeCode_ptr
    createPrimitiveTC(CORBA::TCKind);

    static CORBA::TypeCode_ptr
    createSysExceptionTC(const char*, const char*);
    
    static CORBA::TypeCode_ptr
    createInterfaceTC(const char*, const char*);
    
    static CORBA::TypeCode_ptr
    createStringTC(CORBA::ULong);

    static CORBA::TypeCode_ptr
    createBadKindTC();
};

} // End of namespace OB

// ----------------------------------------------------------------------
// External, non-inline duplicate/release for templates
// ----------------------------------------------------------------------

void
OBDuplicate(TypeCode_ptr p)
{
    if(p)
	p -> _OB_duplicateWithGC();
}

void
OBRelease(TypeCode_ptr p)
{
    if(p)
	p -> _OB_releaseWithGC();
}

// ----------------------------------------------------------------------
// TypeCodeFactory public member implementation
// ----------------------------------------------------------------------

TypeCode_ptr
TypeCodeFactory::createPrimitiveTC(TCKind kind)
{
    //
    // Note: tk_wchar is not supported in OBE
    //
    assert(kind == tk_null ||
	   kind == tk_void ||
	   kind == tk_short ||
	   kind == tk_long ||
	   kind == tk_longlong ||
	   kind == tk_ushort ||
	   kind == tk_ulong ||
	   kind == tk_ulonglong ||
	   kind == tk_float ||
	   kind == tk_double ||
	   kind == tk_longdouble ||
	   kind == tk_boolean ||
	   kind == tk_char ||
	   kind == tk_octet ||
	   kind == tk_any ||
	   kind == tk_TypeCode ||
	   kind == tk_Principal);

    TypeCode_var tc = new TypeCode;
    tc -> kind_ = kind;
    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createSysExceptionTC(const char* id, const char* name)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_except;
    tc -> id_ = id;
    tc -> name_ = name;

    tc -> memberNames_.length(2);
    tc -> memberTypes_.length(2);
    tc -> memberNames_[0] = string_dup("minor");
    tc -> memberTypes_[0] = TypeCode::_duplicate(_tc_ulong);
    tc -> memberNames_[1] = string_dup("completed");
    tc -> memberTypes_[1] = TypeCode::_duplicate(_tc_completion_status);

    TypeCode::_OB_embedRecTC(tc);
    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createInterfaceTC(const char* id, const char* name)
{
    assert_nca(id != 0, NCANullString);
    assert_nca(name != 0, NCANullString);

    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_objref;
    tc -> id_ = id;
    tc -> name_ = name;

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createStringTC(ULong bound)
{
    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_string;
    tc -> length_ = bound;

    return tc._retn();
}

TypeCode_ptr
TypeCodeFactory::createBadKindTC()
{
    TypeCode_var tc = new TypeCode;

    tc -> kind_ = tk_except;
    tc -> id_ = (const char*)"IDL:omg.org/CORBA/TypeCode/BadKind:1.0";
    tc -> name_ = (const char*)"BadKind";

    tc -> memberNames_.length(0);
    tc -> memberTypes_.length(0);

    TypeCode::_OB_embedRecTC(tc);
    return tc._retn();
}

// ----------------------------------------------------------------------
// TypeCode private and protected member implementation
// ----------------------------------------------------------------------

bool
TypeCode::equivalentRecHelper(TypeCode_ptr t,
                              OB::FixSeq< TypeCode_ptr, int >& history,
                              OB::FixSeq< TypeCode_ptr, int >& otherHistory)
                              const
{
    ULong i;

    if(is_nil(t))
        return false;

    if(t == this)
        return true;

    //
    // Avoid infinite loops
    //
    bool foundLoop = false, foundOtherLoop = false;
    for(i = 0 ; i < history.length() && !foundLoop ; i++)
        if(this == history[i])
            foundLoop = true;
    for(i = 0 ; i < otherHistory.length() && !foundOtherLoop ; i++)
        if(t == otherHistory[i])
            foundOtherLoop = true;
    if(foundLoop && foundOtherLoop)
        return true;

    history.length(history.length() + 1);
    history[history.length() - 1] = (TypeCode_ptr)this;
    otherHistory.length(otherHistory.length() + 1);
    otherHistory[otherHistory.length() - 1] = (TypeCode_ptr)t;

    bool result = equivalentRec(t, history, otherHistory);

    history.length(history.length() - 1);
    otherHistory.length(otherHistory.length() - 1);

    return result;
}

bool
TypeCode::equivalentRec(TypeCode_ptr t,
                        OB::FixSeq< TypeCode_ptr, int >& history,
                        OB::FixSeq< TypeCode_ptr, int >& otherHistory) const
{
    ULong i;

    if(recId_.in())
    {
        assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
        return recType_ -> equivalentRecHelper(t, history, otherHistory);
    }

    if(t -> recId_.in())
    {
        assert_nca(!is_nil(t -> recType_), NCAUnembeddedRecTC);
        return equivalentRecHelper(t -> recType_, history, otherHistory);
    }

    TypeCode_ptr tc1 = GetOrigType_nodup((TypeCode_ptr)this);
    TypeCode_ptr tc2 = GetOrigType_nodup(t);

    if(tc1 -> kind_ != tc2 -> kind_)
        return false;

    if(tc1 -> kind_ == tk_objref ||
       tc1 -> kind_ == tk_struct ||
       tc1 -> kind_ == tk_union ||
       tc1 -> kind_ == tk_enum ||
       tc1 -> kind_ == tk_alias ||
       tc1 -> kind_ == tk_native ||
       tc1 -> kind_ == tk_except)
    {
        if(strcmp(tc1 -> id_, "") != 0 && strcmp(tc2 -> id_, "") != 0)
        {
            if(strcmp(tc1 -> id_, tc2 -> id_) == 0)
                return true;
            else
                return false;
        }
    }

    // names_ and memberNames_ must be ignored

    if(tc1 -> kind_ == tk_struct ||
       tc1 -> kind_ == tk_union ||
       tc1 -> kind_ == tk_except)
    {
        if(tc1 -> memberTypes_.length() != tc2 -> memberTypes_.length())
            return false;

        for(i = 0 ; i < tc1 -> memberTypes_.length() ; i++)
            if(!(tc1 -> memberTypes_[i] -> equivalentRecHelper(
                 tc2 -> memberTypes_[i], history, otherHistory)))
                return false;
    }

    if(tc1 -> kind_ == tk_union)
    {
        if(tc1 -> labels_.length() != tc2 -> labels_.length())
            return false;

        for(i = 0 ; i < tc1 -> labels_.length() ; i++)
        {
            TypeCode_ptr ltc1 = tc1 -> labels_[i]._OB_type_nodup();
            TypeCode_ptr ltc2 = tc2 -> labels_[i]._OB_type_nodup();

            //
            // Don't use equivalentRecHelper here
            //
            if(!ltc1 -> equivalent(ltc2))
                return false;

            const void* v1 = tc1 -> labels_[i].value();
            const void* v2 = tc2 -> labels_[i].value();

            ltc1 = GetOrigType_nodup(ltc1);
            switch(ltc1 -> kind())
            {
            case tk_short:
                if(*(Short*)v1 != *(Short*)v2)
                    return false;
                break;

            case tk_ushort:
                if(*(UShort*)v1 != *(UShort*)v2)
                    return false;
                break;

            case tk_long:
                if(*(Long*)v1 != *(Long*)v2)
                    return false;
                break;

            case tk_ulong:
            case tk_enum:
                if(*(ULong*)v1 != *(ULong*)v2)
                    return false;
                break;

            case tk_longlong:
                if(*(LongLong*)v1 != *(LongLong*)v2)
                    return false;
                break;

            case tk_ulonglong:
                if(*(ULongLong*)v1 != *(ULongLong*)v2)
                    return false;
                break;

            case tk_char:
                if(*(Char*)v1 != *(Char*)v2)
                    return false;
                break;

            case tk_boolean:
                if(*(Boolean*)v1 != *(Boolean*)v2)
                    return false;
                break;

            case tk_octet:
                break;

            default:
                assert(false);
                break;
            }
        }

        //
        // Don't use equivalentRecHelper here
        //
        if(!(tc1 -> discriminatorType_ ->
             equivalent(tc2 -> discriminatorType_)))
            return false;
    }

    if(tc1 -> kind_ == tk_string ||
       tc1 -> kind_ == tk_sequence ||
       tc1 -> kind_ == tk_array)
    {
        if(tc1 -> length_ != tc2 -> length_)
            return false;
    }

    if(tc1 -> kind_ == tk_sequence ||
       tc1 -> kind_ == tk_array ||
       tc1 -> kind_ == tk_alias)
    {
        if(!(tc1 -> contentType_ -> equivalentRecHelper(
             tc2 -> contentType_, history, otherHistory)))
            return false;
    }

    return true;
}

TypeCode_ptr
TypeCode::getCompactTypeCodeRec(
    OB::FixSeq< TypeCode_ptr, int >& history,
    OB::FixSeq< TypeCode_ptr, int >& compacted) const
{
    ULong i;

    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> getCompactTypeCodeRec(history, compacted);
    }

    //
    // Avoid infinite loops
    //
    for(i = 0 ; i < history.length(); i++)
        if(this == history[i])
	    return TypeCode::_duplicate(compacted[i]);

    history.length(history.length() + 1);
    history[history.length() - 1] = (TypeCode_ptr)this;

    //
    // Create the new compacted type code (needed for recursive type
    // codes).
    //
    TypeCode_var result = new TypeCode;
    compacted.length(compacted.length() + 1);
    compacted[compacted.length() - 1] = result.in();

    //
    // Compact member names and types.
    //
    OB::StrSeq< int > names;
    names.length(memberNames_.length());
    for(i = 0 ; i < memberNames_.length() ; i++)
	names[i] = string_dup("");

    OB::ObjSeq< TypeCode, int > types;
    types.length(memberTypes_.length());
    for(i = 0 ; i < memberTypes_.length() ; i++)
	types[i] = memberTypes_[i] -> getCompactTypeCodeRec(history, 
							    compacted);

    //
    // Compact content type.
    //
    TypeCode_var content;
    if(contentType_.in())
        content = contentType_ -> getCompactTypeCodeRec(history, 
							compacted);

    //
    // Compact discriminator type.
    //
    TypeCode_var discriminator;
    if(discriminatorType_.in())
	discriminator = discriminatorType_ -> getCompactTypeCodeRec(history, 
								    compacted);

    switch(kind_)
    {
    case tk_null:
    case tk_void:
    case tk_short:
    case tk_long:
    case tk_longlong:
    case tk_ushort:
    case tk_ulong:
    case tk_ulonglong:
    case tk_float:
    case tk_double:
    case tk_longdouble:
    case tk_boolean:
    case tk_char:
    case tk_octet:
    case tk_any:
    case tk_TypeCode:
    case tk_Principal:
        result -> kind_ = kind_;
        break;

    case tk_objref:
    case tk_native:
        result -> kind_ = kind_;
        result -> id_ = id_;
        result -> name_ = string_dup("");
        break;

    case tk_struct:
    case tk_except:
        result -> kind_ = kind_;
        result -> id_ = id_;
        result -> name_ = string_dup("");
        result -> memberNames_ = names;
        result -> memberTypes_ = types;
        break;

    case tk_union:
        result -> kind_ = kind_;
        result -> id_ = id_;
        result -> name_ = string_dup("");
        result -> memberNames_ = names;
        result -> memberTypes_ = types;
        result -> labels_ = labels_;
        result -> discriminatorType_ = discriminator._retn();
        break;

    case tk_enum:
        result -> kind_ = kind_;
        result -> id_ = id_;
        result -> name_ = string_dup("");
        result -> memberNames_ = names;
        break;

    case tk_string:
        result -> kind_ = kind_;
        result -> length_ = length_;
        break;

    case tk_sequence:
    case tk_array:
        result -> kind_ = kind_;
        result -> length_ = length_;
        result -> contentType_ = content._retn();
        break;

    case tk_alias:
        result -> kind_ = kind_;
        result -> id_ = id_;
        result -> name_ = string_dup("");
        result -> contentType_ = content._retn();
        break;

    default:
        assert(false);
    }

    if(is_nil(result))
        assert(false);

    return result._retn();
}

void
TypeCode::resetTryRelease()
{
    if(releaseVisit_)
	return;
    releaseVisit_ = true;

    tryReleaseVisit_ = false;
    tryRefCount_ = _OB_getRef();
    
    if(recId_.in())
    {
	if(!is_nil(recType_))
	    recType_ -> resetTryRelease();
    }
    else
    {
	if(kind_ == tk_sequence ||
	   kind_ == tk_array ||
	   kind_ == tk_alias)
	{
	    if(!is_nil(contentType_))
		contentType_ -> resetTryRelease();
	}
	else if(kind_ == tk_struct ||
		kind_ == tk_union ||
		kind_ == tk_except)
	{
	    for(ULong i = 0 ; i < memberTypes_.length() ; i++)
		if(!is_nil(memberTypes_[i]))
		    memberTypes_[i] -> resetTryRelease();
	}
    }

    releaseVisit_ = false;
}

void
TypeCode::tryRelease()
{
    if(tryReleaseVisit_)
	return;
    tryReleaseVisit_ = true;

    if(recId_.in())
    {
	if(!is_nil(recType_))
	{
	    recType_ -> tryRefCount_--;
	    recType_ -> tryRelease();
	}
    }
    else
    {
	if(kind_ == tk_sequence ||
	   kind_ == tk_array ||
	   kind_ == tk_alias)
	{
	    if(!is_nil(contentType_))
	    {
		contentType_ -> tryRefCount_--;
		contentType_ -> tryRelease();
	    }
	}
	else if(kind_ == tk_struct ||
		kind_ == tk_union ||
		kind_ == tk_except)
	{
	    for(ULong i = 0 ; i < memberTypes_.length() ; i++)
		if(!is_nil(memberTypes_[i]))
		{
		    memberTypes_[i] -> tryRefCount_--;
		    memberTypes_[i] -> tryRelease();
		}
	}
    }
}

bool
TypeCode::checkTryRelease(TypeCode_ptr tc)
{
    if(releaseVisit_)
	return true;
    releaseVisit_ = true;
    
    if(this == tc)
    {
	if(tryRefCount_ > 1)
	{
	    releaseVisit_ = false;
	    return false;
	}
    }
    else
    {
	if(tryRefCount_ > 0)
	{
	    if(isReachable(tc))
	    {
		releaseVisit_ = false;
		return false;
	    }
	}
    }

    if(recId_.in())
    {
	if(!is_nil(recType_))
	{
	    bool res = recType_ -> checkTryRelease(tc);
	    releaseVisit_ = false;
	    return res;
	}
    }
    else
    {
	if(kind_ == tk_sequence ||
	   kind_ == tk_array ||
	   kind_ == tk_alias)
	{
	    if(!is_nil(contentType_))
	    {
		bool res = contentType_ -> checkTryRelease(tc);
		releaseVisit_ = false;
		return res;
	    }
	}
	else if(kind_ == tk_struct ||
		kind_ == tk_union ||
		kind_ == tk_except)
	{
	    for(ULong i = 0 ; i < memberTypes_.length() ; i++)
		if(!is_nil(memberTypes_[i]))
		    if(!memberTypes_[i] -> checkTryRelease(tc))
		    {
			releaseVisit_ = false;
			return false;
		    }
	}
    }

    releaseVisit_ = false;
    return true;
}

bool
TypeCode::isReachable(TypeCode_ptr tc)
{
    if(reachableVisit_)
	return false;
    reachableVisit_ = true;

    if(recId_.in())
    {
	if(!is_nil(recType_))
	{
	    if(recType_.in() == tc)
	    {
		reachableVisit_ = false;
		return true;
	    }
	    else
	    {
		bool res = recType_ -> isReachable(tc);
		reachableVisit_ = false;
		return res;
	    }
	}
    }
    else
    {
	if(kind_ == tk_sequence ||
	   kind_ == tk_array ||
	   kind_ == tk_alias)
	{
	    if(!is_nil(contentType_))
	    {
		if(contentType_.in() == tc)
		{
		    reachableVisit_ = false;
		    return true;
		}
		else
		{
		    bool res = contentType_ -> isReachable(tc);
		    reachableVisit_ = false;
		    return res;
		}
	    }
	}
	else if(kind_ == tk_struct ||
		kind_ == tk_union ||
		kind_ == tk_except)
	{
	    for(ULong i = 0 ; i < memberTypes_.length() ; i++)
		if(!is_nil(memberTypes_[i]))
		{
		    if(memberTypes_[i].in() == tc)
		    {
			reachableVisit_ = false;
			return true;
		    }
		    else
		    {
			if(memberTypes_[i] -> isReachable(tc))
			{
			    reachableVisit_ = false;
			    return true;
			}
		    }
		}
	}
    }

    return false;
}

// ----------------------------------------------------------------------
// TypeCode constructor and destructor
// ----------------------------------------------------------------------

TypeCode::TypeCode()
    : kind_(tk_null), releaseVisit_(false), reachableVisit_(false)
{
#ifdef OB_TYPECODE_DEBUG
    if(GCMutex != 0)
	GCMutex -> lock();

    TCCount++;

    if(GCMutex != 0)
	GCMutex -> unlock();

    cout << " + " << TCCount << endl;
#endif
}

TypeCode::~TypeCode()
{
#ifdef OB_TYPECODE_DEBUG
    if(GCMutex != 0)
	GCMutex -> lock();

    assert(TCCount > 0);
    TCCount--;

    if(GCMutex != 0)
	GCMutex -> unlock();

    cout << " - " << TCCount << endl;
#endif
}

// ----------------------------------------------------------------------
// TypeCode public member implementation
// ----------------------------------------------------------------------

Boolean
TypeCode::equal(TypeCode_ptr p) const
{
    if(is_nil(p))
        return false;

    if(p == this)
	return true;

    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> equal(p);
    }
    
    if(p -> recId_.in())
    {
	assert_nca(!is_nil(p -> recType_), NCAUnembeddedRecTC);
	return equal(p -> recType_);
    }
    
    if(kind_ != p -> kind_)
	return false;

    if(kind_ == tk_objref ||
       kind_ == tk_struct ||
       kind_ == tk_union ||
       kind_ == tk_enum ||
       kind_ == tk_alias ||
       kind_ == tk_native ||
       kind_ == tk_except)
    {
        if(strcmp(id_, "") != 0 || strcmp(p -> id_, "") != 0)
        {
            if(strcmp(id_, p -> id_) == 0)
                return true;
            else
                return false;
        }

        if(strcmp(name_, "") != 0 || strcmp(p -> name_, "") != 0)
        {
            if(strcmp(name_, p -> name_) != 0)
                return false;
        }
    }
    
    if(kind_ == tk_struct ||
       kind_ == tk_union ||
       kind_ == tk_enum ||
       kind_ == tk_except)
    {
 	if(memberNames_.length() != p -> memberNames_.length())
 	    return false;
	
 	for(ULong i = 0 ; i < memberNames_.length() ; i++)
        {
            if(strcmp(memberNames_[i], "") != 0 ||
               strcmp(p -> memberNames_[i], "") != 0)
            {
                if(strcmp(memberNames_[i], p -> memberNames_[i]) != 0)
                    return false;
            }
        }
    }
    
    if(kind_ == tk_struct ||
       kind_ == tk_union ||
       kind_ == tk_except)
    {
	if(memberTypes_.length() != p -> memberTypes_.length())
	    return false;
	
	for(ULong i = 0 ; i < memberTypes_.length() ; i++)
	    if(!memberTypes_[i] -> equal(p -> memberTypes_[i]))
		return false;
    }
    
    if(kind_ == tk_union)
    {
	if(labels_.length() != p -> labels_.length())
	    return false;
	
	for(ULong i = 0 ; i < labels_.length() ; i++)
	{
	    TypeCode_ptr ltc1 = labels_[i]._OB_type_nodup();
	    TypeCode_ptr ltc2 = p -> labels_[i]._OB_type_nodup();
    
	    if(!ltc1 -> equal(ltc2))
		return false;

	    const void* v1 = labels_[i].value();
	    const void* v2 = p -> labels_[i].value();

	    ltc1 = GetOrigType_nodup(ltc1);
	    switch(ltc1 -> kind())
	    {
	    case tk_short:
		if(*(Short*)v1 != *(Short*)v2)
		    return false;
		break;

	    case tk_ushort:
		if(*(UShort*)v1 != *(UShort*)v2)
		    return false;
		break;

	    case tk_long:
		if(*(Long*)v1 != *(Long*)v2)
		    return false;
		break;

	    case tk_ulong:
	    case tk_enum:
		if(*(ULong*)v1 != *(ULong*)v2)
		    return false;
		break;

	    case tk_longlong:
		if(*(LongLong*)v1 != *(LongLong*)v2)
		    return false;
		break;

	    case tk_ulonglong:
		if(*(ULongLong*)v1 != *(ULongLong*)v2)
		    return false;
		break;

	    case tk_char:
		if(*(Char*)v1 != *(Char*)v2)
		    return false;
		break;

	    case tk_boolean:
		if(*(Boolean*)v1 != *(Boolean*)v2)
		    return false;
		break;

	    case tk_octet:
		break;

	    default:
		assert(false);
		break;
	    }
	}

	if(!discriminatorType_ -> equal(p -> discriminatorType_))
	    return false;
    }
    
    if(kind_ == tk_string ||
       kind_ == tk_sequence ||
       kind_ == tk_array)
    {
	if(length_ != p -> length_)
	    return false;
    }
    
    if(kind_ == tk_sequence ||
       kind_ == tk_array ||
       kind_ == tk_alias)
    {
	if(!contentType_ -> equal(p -> contentType_))
	    return false;
    }

    return true;
}

Boolean
TypeCode::equivalent(TypeCode_ptr p) const
{
    OB::FixSeq< TypeCode_ptr, int > history, otherHistory;

    bool result = equivalentRecHelper(p, history, otherHistory);

    assert(history.length() == 0);
    assert(otherHistory.length() == 0);

    return result;
}

TypeCode_ptr
TypeCode::get_compact_typecode() const
{
    OB::FixSeq< TypeCode_ptr, int > history, compacted;
    return  getCompactTypeCodeRec(history, compacted);
}

TCKind
TypeCode::kind() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> kind();
    }

    return kind_;
}

const char*
TypeCode::id() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> id();
    }

    if(!(kind_ == tk_objref ||
	 kind_ == tk_struct ||
	 kind_ == tk_union ||
	 kind_ == tk_enum ||
	 kind_ == tk_alias ||
	 kind_ == tk_native ||
	 kind_ == tk_except))
    {
	throw BadKind();
    }

    return id_;
}

const char*
TypeCode::name() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> name();
    }

    if(!(kind_ == tk_objref ||
	 kind_ == tk_struct ||
	 kind_ == tk_union ||
	 kind_ == tk_enum ||
	 kind_ == tk_alias ||
	 kind_ == tk_native ||
	 kind_ == tk_except))
    {
	throw BadKind();
    }
    
    return name_;
}

bool
TypeCode::_OB_isSystemException() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> _OB_isSystemException();
    }

    if(kind_ != tk_except)
	return false;

    if(strcmp(id_, "IDL:omg.org/CORBA/UNKNOWN:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/BAD_PARAM:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/NO_MEMORY:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/IMP_LIMIT:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/COMM_FAILURE:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INV_OBJREF:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/NO_PERMISSION:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INTERNAL:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/MARSHAL:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INITIALIZE:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/NO_IMPLEMENT:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/BAD_TYPECODE:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/BAD_OPERATION:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/NO_RESOURCES:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/NO_RESPONSE:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/BAD_INV_ORDER:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/TRANSIENT:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/OBJ_ADAPTER:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/OBJECT_NOT_EXIST:1.0") == 0 ||
       strcmp(id_, "IDL:omg.org/CORBA/INV_POLICY:1.0") == 0)
    {
	return true;
    }
    else
    {
	return false;
    }
}

void
TypeCode::_OB_duplicateWithGC()
{
    if(GCMutex != 0)
	GCMutex -> lock();
	
    _OB_incRef();

    if(GCMutex != 0)
	GCMutex -> unlock();
}

void
TypeCode::_OB_releaseWithGC()
{
    if(GCMutex != 0)
	GCMutex -> lock();
	
    if(_OB_getRef() > 1)
    {
	resetTryRelease();
	tryRelease();
	
	if(checkTryRelease(this))
	{
	    recType_ = _nil();
	    contentType_ = _nil();
	    memberTypes_.length(0);
	    assert(_OB_getRef() == 1);
	}
    }

    _OB_decRef();

    if(GCMutex != 0)
	GCMutex -> unlock();
}

void
TypeCode::_OB_embedRecTC(TypeCode_ptr outer)
{
    //
    // Recursive placeholder TypeCodes are illegal as "outer" argument
    //
    assert(!outer -> recId_.in());

    //
    // Check for illegal recursion
    //
    if(!(outer -> kind_ == tk_struct ||
	 outer -> kind_ == tk_except ||
	 outer -> kind_ == tk_union))
    {
	throw BAD_TYPECODE();
    }

    _OB_embedRecTC(outer, outer);
}

void
TypeCode::_OB_embedRecTC(TypeCode_ptr outer, TypeCode_ptr inner)
{
    //
    // Embed recursive placeholder TypeCodes
    //
    if(inner -> recId_.in())
    {
	if(strcmp(inner -> recId_, outer -> id_) == 0)
	{
	    if(inner -> recType_.in())
	    {
		assert_nca(inner -> recType_.in() == outer, NCAEmbeddedRecTC);
	    }
	    else
	    {
		//
		// Embed the recursive placeholder TypeCode
		//
		inner -> recType_ = TypeCode::_duplicate(outer);
	    }
	}
    }
    else
    {
	//
	// Embed content type
	//
	if(inner -> kind_ == tk_sequence ||
	   inner -> kind_ == tk_array ||
	   inner -> kind_ == tk_alias)
	{
	    assert(inner -> contentType_.in() != outer);
	    _OB_embedRecTC(outer, inner -> contentType_);
	}
    
	//
	// Embed member types
	//
	if(inner -> kind_ == tk_struct ||
	   inner -> kind_ == tk_union ||
	   inner -> kind_ == tk_except)
	{
	    for(ULong i = 0 ; i < inner -> memberTypes_.length() ; i++)
	    {
		assert(inner -> memberTypes_[i].in() != outer);
		_OB_embedRecTC(outer, inner -> memberTypes_[i]);
	    }
	}
    }
}

TypeCode_ptr
TypeCode::_OB_contentType_nodup() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> _OB_contentType_nodup();
    }

    if(!(kind_ == tk_sequence ||
	 kind_ == tk_array ||
	 kind_ == tk_alias))
    {
	throw BadKind();
    }

    return contentType_;
}

ULong
TypeCode::_OB_length() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> _OB_length();
    }

    if(!(kind_ == tk_string ||
	 kind_ == tk_sequence ||
	 kind_ == tk_array))
    {
	throw BadKind();
    }

    return length_;
}

Long
TypeCode::_OB_defaultIndex() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> _OB_defaultIndex();
    }

    if(!(kind_ == tk_union))
    {
	throw BadKind();
    }

    for(Long i = 0 ; i < (Long)labels_.length() ; i++)
    {
	TypeCode_ptr tc = labels_[i]._OB_type_nodup();
	if(tc -> kind() == tk_octet)
	    return i;
    }

    return -1;
}

#ifndef NDEBUG

ULong
TypeCode::_OB_memberCount() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> _OB_memberCount();
    }

    if(!(kind_ == tk_struct ||
	 kind_ == tk_union ||
	 kind_ == tk_enum ||
	 kind_ == tk_except))
    {
	throw BadKind();
    }
    
    return memberNames_.length();
}

const char*
TypeCode::_OB_memberName(ULong index) const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> _OB_memberName(index);
    }

    if(!(kind_ == tk_struct ||
	 kind_ == tk_union ||
	 kind_ == tk_enum ||
	 kind_ == tk_except))
    {
	throw BadKind();
    }

    if(index >= memberNames_.length())
    {
	return string_dup("");
    }

    return memberNames_[index];
}

TypeCode_ptr
TypeCode::_OB_memberType_nodup(ULong index) const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> _OB_memberType_nodup(index);
    }

    if(!(kind_ == tk_struct ||
	 kind_ == tk_union ||
	 kind_ == tk_except))
    {
	throw BadKind();
    }

    if(index >= memberNames_.length())
    {
	return TypeCode::_nil();
    }

    return memberTypes_[index];
}

Any*
TypeCode::_OB_memberLabel(ULong index) const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> _OB_memberLabel(index);
    }

    if(!(kind_ == tk_union))
    {
	throw BadKind();
    }

    if(index >= memberNames_.length())
    {
	return new Any;
    }

    return new Any(labels_[index]);
}

TypeCode_ptr
TypeCode::_OB_discriminatorType_nodup() const
{
    if(recId_.in())
    {
	assert_nca(!is_nil(recType_), NCAUnembeddedRecTC);
	return recType_ -> _OB_discriminatorType_nodup();
    }

    if(!(kind_ == tk_union))
    {
	throw BadKind();
    }

    return discriminatorType_;
}

#endif

// ----------------------------------------------------------------------
// TypeCode::BadKind public member implementation
// ----------------------------------------------------------------------

void
TypeCode::BadKind::_OB_insertInAny(Any& any)
{
    any <<= this;
}

void
TypeCode::BadKind::_OB_unmarshal(TypeCode::BadKind&, InputStreamImpl* in)
{
    const char* id = in -> read_string_nodup();
    assert(strcmp(id, "IDL:omg.org/CORBA/TypeCode/BadKind:1.0") == 0);
}

// ----------------------------------------------------------------------
// Insertion and extraction for TypeCode::BadKind
// ----------------------------------------------------------------------

void
operator<<=(Any& any, TypeCode::BadKind* v)
{
    static const Info< TypeCode::BadKind > info;
    any.replace(TypeCode::_tc_BadKind, v, true, &info);
}

void
operator<<=(Any& any, const TypeCode::BadKind& v)
{
    any <<= new TypeCode::BadKind(v);
}

Boolean
operator>>=(const Any& any, TypeCode::BadKind*& v)
{
    if(any.check_type(TypeCode::_tc_BadKind))
    {
        if(!any.info())
        {
            InputStreamImpl* in = any.get_input_stream();
            TypeCode::BadKind* val = new TypeCode::BadKind;
            TypeCode::BadKind::_OB_unmarshal(*val, in);
            (Any&)any <<= val;
        }

        v = (TypeCode::BadKind*)any.value();
        return true;
    }
    else
        return false;
}

// ----------------------------------------------------------------------
// _tc_<type> constants and class members for _tc_<type> constants
// ----------------------------------------------------------------------

TypeCodePrimitiveConst::operator TypeCode_ptr()
{
    JTCSynchronized synchronized(*this);

    if(is_nil(typeCode_))
	typeCode_ = TypeCodeFactory::createPrimitiveTC(kind_);

    return typeCode_; // No _duplicate() here !
}

namespace CORBA
{

TypeCodePrimitiveConst _tc_null(tk_null);
TypeCodePrimitiveConst _tc_void(tk_void);
TypeCodePrimitiveConst _tc_short(tk_short);
TypeCodePrimitiveConst _tc_long(tk_long);
TypeCodePrimitiveConst _tc_longlong(tk_longlong);
TypeCodePrimitiveConst _tc_ushort(tk_ushort);
TypeCodePrimitiveConst _tc_ulong(tk_ulong);
TypeCodePrimitiveConst _tc_ulonglong(tk_ulonglong);
TypeCodePrimitiveConst _tc_float(tk_float);
TypeCodePrimitiveConst _tc_double(tk_double);
TypeCodePrimitiveConst _tc_longdouble(tk_longdouble);
TypeCodePrimitiveConst _tc_boolean(tk_boolean);
TypeCodePrimitiveConst _tc_char(tk_char);
TypeCodePrimitiveConst _tc_octet(tk_octet);
TypeCodePrimitiveConst _tc_any(tk_any);
TypeCodePrimitiveConst _tc_TypeCode(tk_TypeCode);
TypeCodePrimitiveConst _tc_Principal(tk_Principal);

} // End namespace CORBA

TypeCodeStringConst::operator TypeCode_ptr()
{
    JTCSynchronized synchronized(*this);

    if(is_nil(typeCode_))
    {
	typeCode_ = TypeCodeFactory::createStringTC(0);
    }

    return typeCode_; // No _duplicate() here !
}

TypeCode_ptr
TypeCodeStringConst::_OB_bounded(ULong bound)
{
    return TypeCodeFactory::createStringTC(bound);
}

namespace CORBA
{

TypeCodeStringConst _tc_string(false);
TypeCodeStringConst _tc_wstring(true);

}

TypeCodeObjectConst::operator TypeCode_ptr()
{
    JTCSynchronized synchronized(*this);

    if(is_nil(typeCode_))
	typeCode_ =
	    TypeCodeFactory::createInterfaceTC("IDL:omg.org/CORBA/Object:1.0",
					       "Object");

    return typeCode_; // No _duplicate() here !
}

namespace CORBA
{

TypeCodeObjectConst _tc_Object;

}

TypeCodeSysExConst::operator TypeCode_ptr()
{
    JTCSynchronized synchronized(*this);

    if(is_nil(typeCode_))
    {
	String_var id = string_dup("IDL:omg.org/CORBA/");
	id += name_;
	id += ":1.0";
	String_var name = string_dup(name_);
	typeCode_ = TypeCodeFactory::createSysExceptionTC(id, name);
    }

    return typeCode_; // No _duplicate() here !
}

namespace CORBA
{

TypeCodeSysExConst _tc_UNKNOWN("UNKNOWN");
TypeCodeSysExConst _tc_BAD_PARAM("BAD_PARAM");
TypeCodeSysExConst _tc_NO_MEMORY("NO_MEMORY");
TypeCodeSysExConst _tc_IMP_LIMIT("IMP_LIMIT");
TypeCodeSysExConst _tc_COMM_FAILURE("COMM_FAILURE");
TypeCodeSysExConst _tc_INV_OBJREF("INV_OBJREF");
TypeCodeSysExConst _tc_NO_PERMISSION("NO_PERMISSION");
TypeCodeSysExConst _tc_INTERNAL("INTERNAL");
TypeCodeSysExConst _tc_MARSHAL("MARSHAL");
TypeCodeSysExConst _tc_INITIALIZE("INITIALIZE");
TypeCodeSysExConst _tc_NO_IMPLEMENT("NO_IMPLEMENT");
TypeCodeSysExConst _tc_BAD_TYPECODE("BAD_TYPECODE");
TypeCodeSysExConst _tc_BAD_OPERATION("BAD_OPERATION");
TypeCodeSysExConst _tc_NO_RESOURCES("NO_RESOURCES");
TypeCodeSysExConst _tc_NO_RESPONSE("NO_RESPONSE");
TypeCodeSysExConst _tc_BAD_INV_ORDER("BAD_INV_ORDER");
TypeCodeSysExConst _tc_TRANSIENT("TRANSIENT");
TypeCodeSysExConst _tc_OBJ_ADAPTER("OBJ_ADAPTER");
TypeCodeSysExConst _tc_OBJECT_NOT_EXIST("OBJECT_NOT_EXIST");
TypeCodeSysExConst _tc_INV_POLICY("INV_POLICY");

} // End namespace CORBA

TypeCodeConst::operator TypeCode_ptr()
{
    JTCSynchronized synchronized(*this);

    if(is_nil(typeCode_))
    {
	assert(str_ != 0);
	Octet* oct = AsciiToOctets(str_);
        ULong len = strlen(str_);
	InputStreamImpl in(len / 2, 0, false);
	memcpy(in._OB_bufferImpl() -> data(), oct, len / 2);
        delete [] oct;
	in._OB_readEndian();
	typeCode_ = in.read_TypeCode();
    }
    
    return typeCode_; // No _duplicate() here !
}

TypeCode::TypeCodeConstForBadKind::operator TypeCode_ptr()
{
    JTCSynchronized synchronized(*this);

    if(is_nil(typeCode_))
	typeCode_ = TypeCodeFactory::createBadKindTC();
    
    return typeCode_; // No _duplicate() here !
}

TypeCode::TypeCodeConstForBadKind CORBA::TypeCode::_tc_BadKind;

// ----------------------------------------------------------------------
// Get original type, i.e. remove aliases
// ----------------------------------------------------------------------

TypeCode_ptr
OB::GetOrigType_nodup(TypeCode_ptr p)
{
    TypeCode_ptr tc = p;

    while(tc -> kind() == tk_alias)
	tc = tc -> _OB_contentType_nodup();

    return tc;
}

