// **********************************************************************
//
// Copyright (c) 2000
// Object Oriented Concepts, Inc.
// Billerica, MA, USA
//
// All Rights Reserved
//
// **********************************************************************

#include <OBE/CORBA.h>

#include <TestMacro.h>
#include <Test_skel.h>
#include <Test_impl.h>
#include <TestUtil.h>

#include <errno.h>

using namespace CORBA;
using namespace PortableServer;

void
TestToString()
{
    ObjectId_var oid1 = new ObjectId(10);
    oid1 -> length(10);
    memset(oid1 -> get_buffer(), 0, 10);

    //
    // Shouldn't succeed - embedded 0's
    //
    try
    {
	CORBA::String_var s1 = ObjectId_to_string(oid1);
	TEST(false); // ObjectId_to_string should not have succeeded
    }
    catch(const CORBA::BAD_PARAM&)
    {
    }

    oid1 -> length(12);
    memset(oid1 -> get_buffer(), 0, 10);

    //
    // Shouldn't succeed - embedded 0's, wrong length
    //
    try
    {
	CORBA::String_var s1 = ObjectId_to_string(oid1);
	TEST(false); // ObjectId_to_string should not have succeeded
    }
    catch(const CORBA::BAD_PARAM&)
    {
    }

    CORBA::Octet bytes[] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
			     '0', '1' };
    for(CORBA::ULong i = 0; i < sizeof(bytes)/sizeof(bytes[0]); ++i)
	oid1[i] = bytes[i];

    CORBA::String_var s1 = ObjectId_to_string(oid1);
    ObjectId_var oid2 = string_to_ObjectId(s1);
    TEST(oid2 -> length() == oid1 -> length());
    TEST(memcmp(oid1 -> get_buffer(), oid2 -> get_buffer(),
		oid1 -> length()) == 0);
}

void
TestActivateObject(ORB_ptr orb, POA_ptr root)
{
    Test_impl* impl1 = new Test_impl();
    Test_impl* impl2 = new Test_impl();

    //
    // Test: activate_object
    //
    ObjectId_var id1 = root -> activate_object(impl1);
    ObjectId_var id2 = root -> activate_object(impl2);
    TEST(!TestUtil_Compare(id1, id2));

    //
    // Test: calling activate_object again should create a new object
    //
    ObjectId_var tmpid = root -> activate_object(impl1);
    TEST(!TestUtil_Compare(id1, tmpid));

    root -> deactivate_object(id1);
    root -> deactivate_object(id2);
    root -> deactivate_object(tmpid);

    delete impl1;
    delete impl2;
}

void
TestDeactivateObject(ORB_ptr orb, POA_ptr root)
{
    //
    // Test: ObjectNotActive exception
    //
    try
    {
	ObjectId_var tmpid = string_to_ObjectId("bad_id");
        root -> deactivate_object(tmpid);
        TEST(false); // id_to_servant should not have succeeded
    }
    catch(const POA::ObjectNotActive&)
    {
        // expected
    }

    //
    // Create servant
    //
    Test_impl* impl = new Test_impl();

    //
    // Activate the object
    //
    Test_var t = impl -> _this();

    //
    // Call a method on the object
    //
    t -> aMethod();

    //
    // Deactivate the object
    //
    ObjectId_var oid = root -> servant_to_id(impl);
    root -> deactivate_object(oid);

    try
    {
	//
	// Call a method on the object
	//
        t -> aMethod();
        TEST(false); // expected OBJECT_NOT_EXIST
    }
    catch(const OBJECT_NOT_EXIST&)
    {
        // Expected
    }

    //
    // Activate a second object
    //
    oid = root -> activate_object(impl);

    try
    {
	//
	// Call a method on the object
	//
        t -> aMethod();
        TEST(false); // expected OBJECT_NOT_EXIST
    }
    catch(const OBJECT_NOT_EXIST&)
    {
        // Expected
    }

    //
    // Deactivate the second object
    //
    root -> deactivate_object(oid);

    delete impl;
}

void
TestServantToId(ORB_ptr orb, POA_ptr root)
{
    Test_impl* impl1 = new Test_impl();
    Test_impl* impl2 = new Test_impl();

    //
    // Test: servant_to_id
    //
    ObjectId_var id1 = root -> activate_object(impl1);
    ObjectId_var tmpid = root -> servant_to_id(impl1);
    TEST(TestUtil_Compare(id1, tmpid));

    //
    // Test: Implicitly activating impl2 should produce a new ID
    //
    ObjectId_var id2 = root -> servant_to_id(impl2);
    TEST(!TestUtil_Compare(id1, id2));

    root -> deactivate_object(id1);
    root -> deactivate_object(id2);

    delete impl1;
    delete impl2;
}

void
TestServantToReference(ORB_ptr orb, POA_ptr root)
{
    Test_impl* impl1 = new Test_impl();
    Test_impl* impl2 = new Test_impl();

    //
    // Test: servant_to_reference
    //
    ObjectId_var id1 = root -> activate_object(impl1);
    Object_var obj = root -> servant_to_reference(impl1);
    TEST(!is_nil(obj));
    ObjectId_var tmpid1 = root -> reference_to_id(obj);
    TEST(TestUtil_Compare(id1, tmpid1));

    //
    // Test: Implicitly activating impl2 should produce a new ID
    //
    obj = root -> servant_to_reference(impl2);
    TEST(!is_nil(obj));
    ObjectId_var id2 = root -> reference_to_id(obj);
    TEST(!TestUtil_Compare(tmpid1, id2));

    root -> deactivate_object(id1);
    root -> deactivate_object(id2);

    delete impl1;
    delete impl2;
}

void
TestIdToServant(ORB_ptr orb, POA_ptr root)
{
    Test_impl* impl1 = new Test_impl();
    Test_impl* impl2 = new Test_impl();

    //
    // Test: ObjectNotActive exception
    //
    try
    {
        ObjectId_var tmpid = string_to_ObjectId("bad_id");
        root -> id_to_servant(tmpid);
        TEST(false); // id_to_servant should not have succeeded
    }
    catch(const POA::ObjectNotActive&)
    {
        // expected
    }

    ObjectId_var id1 = root -> activate_object(impl1);
    ObjectId_var id2 = root -> activate_object(impl2);

    //
    // Test: id_to_servant
    //
    {
	ServantBase_var tmpservant = root -> id_to_servant(id1);
	TEST(impl1 == tmpservant.in());
    }

    {
	ServantBase_var tmpservant = root -> id_to_servant(id2);
	TEST(impl2 == tmpservant.in());
    }

    root -> deactivate_object(id1);
    root -> deactivate_object(id2);

    delete impl1;
    delete impl2;
}

void
TestIdToReference(ORB_ptr orb, POA_ptr root)
{
    Test_impl* impl1 = new Test_impl();
    Test_impl* impl2 = new Test_impl();

    //
    // Test: ObjectNotActive exception
    //
    try
    {
        ObjectId_var tmpid = string_to_ObjectId("bad_id");
        root -> id_to_reference(tmpid);
        TEST(false); // id_to_servant should not have succeeded
    }
    catch(const POA::ObjectNotActive&)
    {
        // expected
    }

    ObjectId_var id1 = root -> activate_object(impl1);
    ObjectId_var id2 = root -> activate_object(impl2);

    //
    // Test: id_to_reference
    //
    Object_var obj = root -> id_to_reference(id1);
    TEST(!is_nil(obj));
    ObjectId_var tmpid = root -> reference_to_id(obj);
    TEST(TestUtil_Compare(id1, tmpid));
    obj = root -> id_to_reference(id2);
    TEST(!is_nil(obj));
    tmpid = root -> reference_to_id(obj);
    TEST(TestUtil_Compare(id2, tmpid));

    root -> deactivate_object(id1);
    root -> deactivate_object(id2);

    delete impl1;
    delete impl2;
}

void
TestReferenceToServant(ORB_ptr orb, POA_ptr root)
{
    Test_impl* impl1 = new Test_impl();
    Test_impl* impl2 = new Test_impl();

    ObjectId_var id1 = root -> activate_object(impl1);
    ObjectId_var id2 = root -> activate_object(impl2);

    //
    // Test: reference_to_servant
    //
    {
	Object_var obj = root -> _OB_createReference(id1, "IDL:Test:1.0");
	ServantBase_var tmpservant = root -> reference_to_servant(obj);
	TEST(impl1 == tmpservant.in());
    }

    {
	Object_var obj = root -> _OB_createReference(id2, "IDL:Test:1.0");
	ServantBase_var tmpservant = root -> reference_to_servant(obj);
	TEST(impl2 == tmpservant.in());
    }

    root -> deactivate_object(id1);
    root -> deactivate_object(id2);

    delete impl1;
    delete impl2;
}

void
TestReferenceToId(ORB_ptr orb, POA_ptr root)
{
    Test_impl* impl1 = new Test_impl();
    Test_impl* impl2 = new Test_impl();

    ObjectId_var id1 = root -> activate_object(impl1);
    ObjectId_var id2 = root -> activate_object(impl2);

    root -> deactivate_object(id1);
    root -> deactivate_object(id2);

    //
    // Test: reference_to_id
    //
    Object_var obj = root -> _OB_createReference(id1, "IDL:Test:1.0");
    ObjectId_var tmpid = root -> reference_to_id(obj);
    TEST(TestUtil_Compare(tmpid, id1));
    obj = root -> _OB_createReference(id2, "IDL:Test:1.0");
    tmpid = root -> reference_to_id(obj);
    TEST(TestUtil_Compare(tmpid, id2));
}

void
runtests(ORB_ptr orb, POA_ptr root)
{
    TestToString();
    TestActivateObject(orb, root);
    TestDeactivateObject(orb, root);
    TestServantToId(orb, root);
    TestIdToServant(orb, root);
    TestServantToReference(orb, root);
    TestIdToReference(orb, root);
    TestReferenceToServant(orb, root);
    TestReferenceToId(orb, root);
}

int
main(int argc, char* argv[], char*[])
{
    int status = EXIT_SUCCESS;

    ORB_var orb;
    try
    {
        orb = ORB_init(argc, argv);

        POA_var root = TestUtil_GetRootPOA(orb);
	POAManager_var manager = root -> the_POAManager();

        printf("Testing miscellaneous POA functions... ");
	fflush(stdout);

	//
	// Activate the POAManager
	//
	manager -> activate();

        //
        // Run the tests
        //
        runtests(orb, root);

	printf("OK!\n");
	fflush(stdout);
    }
    catch(const Exception& ex)
    {
	fprintf(stderr, "%s\n", ex._rep_id());
        status = EXIT_FAILURE;
    }

    if(!is_nil(orb))
    {
	try
	{
	    orb -> destroy();
	}
	catch(const Exception& ex)
	{
	    fprintf(stderr, "%s\n", ex._rep_id());
	    status = EXIT_FAILURE;
	}
    }

    return status;
}
